#ifndef NMOS6502_H
#define NMOS6502_H

#include "../types.h"

#define NMOS6502CAPS_1BIT			1
#define NMOS6502CAPS_4BIT			2

#define NMOS6502ADDR_A				0x10010
#define NMOS6502ADDR_P				0x10011
#define NMOS6502ADDR_X				0x10012
#define NMOS6502ADDR_Y				0x10013
#define NMOS6502ADDR_S				0x10014
#define NMOS6502ADDR_PCL			0x10015
#define NMOS6502ADDR_PCH			0x10016

#define NMOS6502ADDR_TRGCNT			0x10020
#define NMOS6502ADDR_IRQLINE		0x10021
#define NMOS6502ADDR_TIMECNT		0x10022

#define NMOS6502ADDR_MAPFLAGBASE	0x11000

#define NMOS6502FLAG_C				0x01
#define NMOS6502FLAG_Z				0x02
#define NMOS6502FLAG_I				0x04
#define NMOS6502FLAG_D				0x08
#define NMOS6502FLAG_B				0x10
#define NMOS6502FLAG_A				0x20
#define NMOS6502FLAG_V				0x40
#define NMOS6502FLAG_N				0x80

#define NMOS6502_PLOADFLAGS			(NMOS6502FLAG_A | NMOS6502FLAG_B)

#define NMOS6502TABLE_R8PTRS		0
#define NMOS6502TABLE_W8PTRS		1
#define NMOS6502TABLE_R32PTRS		2
#define NMOS6502TABLE_W32PTRS		3
#define NMOS6502TABLE_STFLGS		5
#define NMOS6502TABLE_OPTYPES		6
#define NMOS6502TABLE_COUTDEV		7

#define NMOS6502TABLE_1COST			8
#define NMOS6502TABLE_2COST			9
#define NMOS6502TABLE_RAMCOST		10

#define NMOS6502_COSTLENGTH			148

#define NMOS6502_TIMEMASK			((1 << 24)-1)

enum E6502_InstTypes{
	BRK=0, RTI, RTS, PHx, PLx, JSR, IMPLIED, IMMEDIATE, ABS_JMP,

	ABS_READ, ABS_MODIFY, ABS_WRITE,
	ZERO_READ, ZERO_MODIFY, ZERO_WRITE,
	ZIND_READ, ZIND_MODIFY, ZIND_WRITE,
	AIND_READ, AIND_MODIFY, AIND_WRITE,
	REL,
	IIND_READ, IIND_MODIFY, IIND_WRITE,
	INDI_READ, INDI_MODIFY, INDI_WRITE,

	ABSOLUTE_IND, IRQ, NMOS6502_NUMTYPES
};

enum E6502_MemoryAccess{ONEMhz_BUS=0, TWOMhz_BUS, RAM, FINISH};

class C_6502CalloutDevice
{
	public :
		virtual bool Keepalive(int counter) = 0;
		virtual void MinimiseTimer(int subvalue) = 0;

		virtual int ReadFrom(BrokenWord addr, unsigned __int8 *value, unsigned __int32 *bvalue, int time) = 0;
		virtual int WriteTo(BrokenWord addr, unsigned __int8 value, unsigned __int32 bvalue, int time) = 0;

		virtual bool TrapOpcode(unsigned __int8 opcode) = 0;
		virtual void UnknownOpcode(unsigned __int8 opcode) = 0;
		virtual void KillOpcode(unsigned __int8 opcode) = 0;
		virtual void VirtualOpcode(unsigned __int8 opcode, unsigned __int8 argument, int time) = 0;
};

class C_6502Base
{
	public :
		virtual void Go(void) = 0;
		virtual void Reset(void) = 0;

		virtual void ReconsiderTiming(void) = 0;
		virtual bool Setup(bool newmem) = 0;
		virtual void CloseDown(void) = 0;

		virtual void SetQuit(bool) = 0;

		virtual void SetAddrPtrs(unsigned __int8 start, unsigned __int8 num, unsigned __int8 *base, unsigned __int32 *bbase, bool read, bool write) = 0;
		virtual void SetAddrFlags(unsigned __int8 start, unsigned __int8 num, bool device, E6502_MemoryAccess btype) = 0;
		virtual void FastSetAddrPtrs(unsigned __int8 start, unsigned __int8 num, unsigned __int8 *read, unsigned __int8 *write) = 0;
		virtual void FastSetAddrFlags(unsigned __int8 start, unsigned __int8 num, bool mapped) = 0;

		virtual void SetContentionTimes(int *ctable) = 0;
		virtual void SetCalloutTime(int states) = 0;
		virtual void SetCalloutDevice(C_6502CalloutDevice *dev) = 0;

		virtual void LeechState(C_6502Base *original) = 0;
		virtual void LeechTables(C_6502Base *original) = 0;
		virtual void *ReturnTable(int tabid) = 0;

		virtual void SetIRQ(void) = 0;
		virtual void ClearIRQ(void) = 0;

		virtual void WriteByte(unsigned __int32 addr, unsigned __int8 value) = 0;
		virtual void WriteDWord(unsigned __int32 addr, unsigned __int32 value) = 0;

		virtual unsigned __int8 ReadByte(unsigned __int32 addr) = 0;
		virtual unsigned __int32 ReadDWord(unsigned __int32 addr) = 0;

		virtual unsigned __int8 *GetByteReadPtr(BrokenWord addr) = 0;
		virtual unsigned __int32 *GetDWordReadPtr(BrokenWord addr) = 0;

		virtual void Push(unsigned __int16 addr) = 0;
};

class C_6502StdBase : public C_6502Base
{
	public :

		C_6502StdBase(void);
		~C_6502StdBase(void);

		void ReconsiderTiming(void);
		void Reset(void);
		void SetQuit(bool);

		virtual bool Setup(bool newmem);
		virtual void CloseDown(void);

		void SetAddrPtrs(unsigned __int8 start, unsigned __int8 num, unsigned __int8 *base, unsigned __int32 *bbase, bool read, bool write);
		void SetAddrFlags(unsigned __int8 start, unsigned __int8 num, bool device, E6502_MemoryAccess btype);
		void FastSetAddrPtrs(unsigned __int8 start, unsigned __int8 num, unsigned __int8 *read, unsigned __int8 *write);
		void FastSetAddrFlags(unsigned __int8 start, unsigned __int8 num, bool mapped);

		void SetContentionTimes(int *ctable);
		void SetCalloutTime(int states);
		void SetCalloutDevice(C_6502CalloutDevice *dev);

		void LeechState(C_6502Base *original);
		void LeechTables(C_6502Base *original);
		void *ReturnTable(int tabid);

		void ClearIRQ(void);

		void WriteByte(unsigned __int32 addr, unsigned __int8 value);
		void WriteDWord(unsigned __int32 addr, unsigned __int32 value);

		unsigned __int8 ReadByte(unsigned __int32 addr);
		unsigned __int32 ReadDWord(unsigned __int32 addr);

		unsigned __int8 *GetByteReadPtr(BrokenWord addr);
		unsigned __int32 *GetDWordReadPtr(BrokenWord addr);

		void Push(unsigned __int16 addr);

	protected :

		static bool quit;
		unsigned __int8 a, x, y, s;
		unsigned __int32 aBig, xBig, yBig;

		unsigned __int32 carryBig;
		unsigned __int8 carry, neg, over, zero, pmisc;
		BrokenWord pc;

		bool mapped_flags[256], allocd;
		unsigned __int8 **o_r_mem;
		unsigned __int8 **o_w_mem;
		unsigned __int32 **f_r_mem;
		unsigned __int32 **f_w_mem;

		C_6502CalloutDevice *device;

		static E6502_MemoryAccess *style_flags;
		static E6502_MemoryAccess **types;

		bool irq, irqe;

		int *Costs[FINISH];
		int *OwnedCosts[FINISH];
		int target, timecounter;
};

#endif
