/*

	ElectrEm (c) 2000 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	gfx/SDL/platgfx.cpp
	===================

	The NULL graphics device

*/
#include "../../targets.h"
#ifdef TRG_GFX_SDL

#include "../gfx.h"
#include "SDL.h"

/*

this is called by the real class constructor in order to achieve the usual
constructor sort of things for platform specific members

*/
void C_gfx_api::PlatformConstructor(void)
{
	screen = NULL;
}

/*

Shutdown is to the platform destructor as Constructor is to the platform
constructor

*/
void C_gfx_api::PlatformShutdown(void)
{
}

/*

This fills a DSPSTRUCT such that its members width and height are equal to
the width and height of the first valid display window at least large enough
to display a bitmap of width w and h.

DSPSTRUCT may have additional platform specific members, and those should be
filled in as required here

*/
DSPSTRUCT C_gfx_api::GetGenuineSize(int w, int h)
{
	SDL_Rect **modes;
	DSPSTRUCT info;
	int i;

	modes = SDL_ListModes(pelformatptr, modeflags);

	if(modes == (SDL_Rect **)-1)
	{
		/* resolution is not restricted */
		info.width = w; info.height = h;
	}
	else
	{
		i = 0;

		/* find last resolution big enough */
		while(modes[i] && modes[i]->w >= w && modes[i]->h >= h)
		{
			i++;
		}

		if(i) i--;

		info.width = modes[i]->w;
		info.height = modes[i]->h;
	}

	return info;
}

/*

SetResolution sets the mode described by DSPSTRUCT, returning true on success
and false on failure

*/
bool C_gfx_api::SetResolution(DSPSTRUCT newsize)
{
	if(screen)
	{
		/* kill screen */
	}

	if((!screen) || (newsize.width != screen->w) || (newsize.height != screen->h))
	{
		screen = SDL_SetVideoMode(newsize.width, newsize.height, 8, modeflags);
	}

	return true;
}

/*

PlatformSetup is called once at the start of the program if program startup
gets as far as requiring a graphics mode, and should do the usual library
initialisation. It is also passed the name of the application.

It should set the class member variable 'CursorRequired' to true if this
graphics driver is not going to draw a mouse cursor while the mouse is
required.

*/
bool C_gfx_api::PlatformSetup(char *appname)
{
	int flaglist[] =
	{
//		SDL_HWSURFACE | SDL_HWPALETTE | SDL_FULLSCREEN,
//		SDL_SWSURFACE | SDL_HWPALETTE | SDL_FULLSCREEN,

		SDL_HWSURFACE | SDL_HWPALETTE,
		SDL_SWSURFACE | SDL_HWPALETTE,

		-1
	};
	int *parser = flaglist;

	pelformat.BitsPerPixel = 8;
	pelformat.BytesPerPixel = 1;
	pelformatptr = &pelformat;

	/* first try with an 8bpp pixel format */
	while((*parser!= -1) && (!SDL_ListModes(&pelformat, *parser)))
		parser++;

	if(*parser == -1)
	{
		/* don't be deterred - try without changing the pixel format */
		pelformatptr = NULL;
		parser = flaglist;
		while((*parser!= -1) && (!SDL_ListModes(NULL, *parser)))
			parser++;
	}

	modeflags = *parser;

	if(*parser == -1)
	{
		errstring = "Required 8bpp pixel format support not found!";
		return false;
	}

	SDL_WM_SetCaption(appname, NULL);

	SDL_ShowCursor(SDL_DISABLE);
	CursorRequired = false;

	return true;
}

/*

This should set the palette pointed to by *cols (256 entries, but not
necessarily all used - see the header file for the definition)

*/
void C_gfx_api::PlatformSetPalette(int pal)
{
	int start = 0, length;

	while(start < 256)
	{
		if(colsets[pal][start].set)
		{
			length = 0;
			while(colsets[pal][start+length].set && start+length < 256)
				length++;

			SDL_SetPalette(screen, SDL_PHYSPAL, &sdlcols[pal][start], start, length);
			start += length;
		}
		else
			start++;
	}
}

/*

This signifies that a palette as defined above has been filled in with the
required r, g and b values, and any preprocessing that is useful for this
target to do on palettes can be done now

*/
void C_gfx_api::PlatformFinishPalette(int pal)
{
	int c = 256;

	while(c--)
	{
		sdlcols[pal][c].r = (int)(colsets[pal][c].red*255.0f);
		sdlcols[pal][c].g = (int)(colsets[pal][c].green*255.0f);
		sdlcols[pal][c].b = (int)(colsets[pal][c].blue*255.0f);
	}
}

/* Mouse stuff from here onwards */

/*

Enable mouse enables the mouse mode. If this target draws a cursor, it should
be enabled (if required) now

*/
void C_gfx_api::EnableMouse(void)
{
	SDL_ShowCursor(SDL_ENABLE);
}

/*

DisableMouse is the opposite of EnableMouse, signifying that the mouse cursor
can be disabled

*/
void C_gfx_api::DisableMouse(void)
{
	SDL_ShowCursor(SDL_DISABLE);
}

/*

GetMouseState should return the current location of the mouse cursor and the
state of its button.

The mouse location is relative to (backupscr.tx, backupscr.ty)

*/

MOUSE_STRUCT C_gfx_api::GetMouseState(void)
{
	MOUSE_STRUCT minfo;

	minfo.lbutton = (SDL_GetMouseState(&minfo.x, &minfo.y)&SDL_BUTTON(1)) ? true : false;
	minfo.x -= backupscr.tx;
	minfo.y -= backupscr.ty;

	return minfo;
}

#endif