#include "gui.h"
#include "guidefs.h"
#include "mainmenu.h"
#include "operror.h"
#include "../gfx/gfx.h"
#include "../ula/ula.h"
#include "../misc/input.h"

#define GENERAL_DEFINES
#include "../targets.h"
#undef GENERAL_DEFINES

#define HIGH_VAL 1.0f
#define MID_VAL 0.7f
#define LOW_VAL 0.1f

C_GUI::C_GUI()
{
	elk = NULL;
	gfx = NULL;
	MainGUI = NULL;
	OpcodeError = NULL;
	HaltOpcode = NULL;
}

void C_GUI::GenerateCols(int colid)
{
	float r1, g1, b1;
	float r2, g2, b2;
	float radd, gadd, badd;
	int c;

	switch(colid&96)
	{
		case FCOL_WHITE :
			r2 = g2 = b2 = HIGH_VAL;
		break;

		case FCOL_GREY :
			r2 = g2 = b2 = MID_VAL;
		break;

		case FCOL_GREEN :
			r2 = b2 = LOW_VAL;
			g2 = HIGH_VAL;
		break;

		case FCOL_RED :
			r2 = HIGH_VAL;
			g2 = b2 = LOW_VAL;
		break;
	}

	switch(colid&24)
	{
		case BCOL_BLUE :
			r1 =
			g1 = LOW_VAL;
			b1 = HIGH_VAL;
		break;

		case BCOL_BLACK :
			r1 =
			g1 =
			b1 = LOW_VAL;
		break;
	}

	radd = (r2 - r1) / 11.0f;
	gadd = (g2 - g1) / 11.0f;
	badd = (b2 - b1) / 11.0f;
	r1 += radd * 4;
	g1 += gadd * 4;
	b1 += badd * 4;

	for(c = colid; c < colid+8; c++)
	{
		gfx->SetPaletteEntry(c, r1, g1, b1, PAL_GUI);
		r1 += radd;
		g1 += gadd;
		b1 += badd;
	}
}

#include <stdio.h>
bool firsttime;

bool C_GUI::Initialise(C_ULA *owner)
{
	elk = owner;
	gfx = elk->GetGFX();

	firsttime = true;

	MainGUI = new C_CardMenu();
	MainGUI->Initialise(this, 0, 0, 640, 512);

	OpcodeError = new C_OpcodeErrorMenu();
	OpcodeError->Initialise(this, 0, 0, 640, 512);

	HaltOpcode = new C_HaltOpcodeMenu();
	HaltOpcode->Initialise(this, 0, 0, 640, 512);

	GenerateCols(FCOL_WHITE | BCOL_BLACK);
	GenerateCols(FCOL_WHITE | BCOL_BLUE);
	GenerateCols(FCOL_RED | BCOL_BLACK);
	GenerateCols(FCOL_GREY | BCOL_BLACK);
	gfx->FinishPalette(PAL_GUI);

	return true;
}

void C_GUI::DeInitialise(void)
{
	if(OpcodeError)
	{
		OpcodeError->DeInitialise();
		delete OpcodeError;
		OpcodeError = NULL;
	}

	if(HaltOpcode)
	{
		HaltOpcode->DeInitialise();
		delete HaltOpcode;
		HaltOpcode = NULL;
	}

	if(MainGUI)
	{
		MainGUI->DeInitialise();
		delete MainGUI;
		MainGUI = NULL;
	}
}

/* this is called to check the preload config file bit */
void C_GUI::EvaluatePreload(void)
{
	MainGUI->EvaluatePreload();
}

/* this passes on an autoload id to the ROM config screen */
bool C_GUI::SetAutoId(unsigned __int8 id)
{
	return MainGUI->SetAutoId(id);
}

void C_GUI::PassROMHint(bool set, int id, char *name)
{
	MainGUI->PassROMHint(set, id, name);
}

/* main part of handling everything! */

void C_GUI::SignalFinished(void)
{
	done = true;
}

void C_GUI::Go(GUI_STARTSCREEN screen)
{
	C_GUIScreen *worker;
	static MOUSE_STRUCT newmouse, oldmouse;
	GUI_MSG msg;
	int shifts;

	switch(screen)
	{
		case GUISCR_DEFAULT :		worker = MainGUI; break;
		case GUISCR_OPCODEERROR :	worker = OpcodeError; break;
		case GUISCR_HALT :			worker = HaltOpcode; break;
	}

	done = false;

	if(firsttime)
	{
		oldmouse = gfx->GetMouseState();
		firsttime = false;
	}

	gfx->EnableGUI();

	elk->GetInput()->ClearBuffer();
	worker->Enter();
	while(!done)
	{
		//check for mouse events
		newmouse = gfx->GetMouseState();
		msg.x = newmouse.x;
		msg.y = newmouse.y;

		if(newmouse.x != oldmouse.x || newmouse.y != oldmouse.y)
		{
			msg.msg = GUIM_MOUSEMOVE;
			worker->SendMsg(msg);
		}

		button = newmouse.lbutton;
		if(button != oldmouse.lbutton)
		{
			msg.msg = button ? GUIM_LBUTTONDOWN : GUIM_LBUTTONUP;
			worker->SendMsg(msg);

			if(!button)
			{
				msg.msg = GUIM_MOUSEMOVE;
				worker->SendMsg(msg);
			}
		}

		oldmouse = newmouse;

		//check for keyboard events
		elk->GetInput()->UpdateKeyboard();
		while(elk->GetInput()->NewKey())
		{
			msg.msg = GUIM_KEYPRESS;
			msg.ch = elk->GetInput()->GetPressedASCII();
			worker->SendMsg(msg);
		}

		//check for exit/enter menu key
		shifts = elk->GetInput()->GetShifts();

		if(shifts&IPT_SHFT_QUIT)
		{
			done = true;
			elk->SetQuit();
		}

		if(shifts&IPT_SHFT_MENU)
			done = true;

		//begin redraw, send keepalive, finish redraw
		gfx->BeginRedraw();
		worker->Draw();
		gfx->EndRedraw();

		check_quit({done = true; elk->SetQuit();});
	}

	/* send button release if button pressed */
	if(button)
	{
		msg.msg = GUIM_LBUTTONUP;
		worker->SendMsg(msg);
	}

	worker->Exit();

	gfx->DisableGUI();
}

C_gfx_api *C_GUI::GetPartialGFX(void)
{
	return gfx;
}

C_ULA *C_GUI::GetElectron(void)
{
	return elk;
}

void C_GUI::SetDefault(GUI_DSCRPTR *desc)
{
	desc->ignoremouse = desc->mouseover = false;
	desc->newclicks = 0;
}

bool C_GUI::PreHandling(GUI_DSCRPTR *desc, GUI_MSG msg)
{
	bool changed = false, over;

	desc->newclicks = 0;

	if(msg.msg == GUIM_KEYPRESS)
	{
		/* handle key input buffer here */
	}
	else
	{
		over =	(msg.x >=	desc->loc.x1) &&
				(msg.x <	desc->loc.x2) &&
				(msg.y >=	desc->loc.y1) &&
				(msg.y <	desc->loc.y2);

		switch(msg.msg)
		{
			case GUIM_MOUSEMOVE :

				if((desc->mouseover != over && !button) || desc->ignoremouse)
				{
					desc->mouseover = over;
					changed = true;
				}

			break;

			case GUIM_LBUTTONDOWN :

				if(over)
				{
					desc->newclicks = 1;
					changed = true;
				}

			break;

			case GUIM_LBUTTONUP :

				if(over || desc->ignoremouse)
				{
					desc->newclicks = -1;
					changed = true;
				}

			break;

			case GUIM_LDCLICK :

				if(over)
				{
					desc->newclicks = 2;
					changed = true;
				}

			break;
		}
	}

	return changed;
}

/* base C_GUIScreen class */

bool C_GUIScreen::Initialise(C_GUI *parent, int x1, int y1, int x2, int y2)
{
	InitialiseList();
	return true;
}

void C_GUIScreen::DeInitialise(void)
{
	Elem *listp, *next;

	listp = list;
	while(listp)
	{
		next = listp->next;
		listp->e->Kill();
		delete listp;
		listp = next;
	}
}

void C_GUIScreen::Enter(void)
{
}

void C_GUIScreen::Exit(void)
{
}

void C_GUIScreen::SendMsg(GUI_MSG msg)
{
	Elem *listp;
	listp = list;
	while(listp)
	{
		listp->e->SendMsg(msg);
		listp = listp->next;
	}
}

void C_GUIScreen::Draw(void)
{
	Elem *listp;
	listp = list;
	while(listp)
	{
		listp->e->Draw();
		listp = listp->next;
	}
}

void C_GUIScreen::InitialiseList(void)
{
	list = NULL;
}

void C_GUIScreen::AddToList(C_GUIElementBase *elem)
{
	Elem *listp;

	if(!list)
	{
		listp = list = new Elem;
	}
	else
	{
		listp = list;
		while(listp->next)
			listp = listp->next;

		listp->next = new Elem;
		listp = listp->next;
	}

	listp->next = NULL;
	listp->e = elem;
}

void C_GUIScreen::KillList(void)
{
	Elem *listp, *next;

	listp = list;
	while(listp)
	{
		next = listp->next;
		delete listp;
		listp = next;
	}
}
