/*

	ElectrEm (c) 2000 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	misc/DirectX/platipt.cpp
	========================

	The input class - DirectX parts.

	Notice : this code falls back onto 'GetKeyboardState' and
	'MapVirtualKey' if the DirectInput init fails. That code is mildly
	slower because it then has to shuffle the key states into the
	DirectInput/scancode ordering, but will make ElectrEm useful for NT
	users whilst I cannot get hold of DirectX 3.0 to compile such that
	'DIRECTINPUT_VERSION' is not too great

*/
#include "../../targets.h"

#ifdef TRG_IPT_DINPUT

#include "../input.h"
#include "../../ula/ula.h"
#include "../../misc/config.h"
#include "../../defines.h"

char aqueue[256];
int lastpos, qpos;

void C_input::PlatformConstructor(void)
{
	lpdiKeyb = NULL;
	lpdi = NULL;
	lastpos = qpos = 0;
}

void C_input::PlatformDestructor(void)
{
	if(lpdiKeyb)
	{
		lpdiKeyb->Unacquire();
		lpdiKeyb->Release();
		lpdiKeyb = NULL;
	}

	if(lpdi)
	{
		lpdi->Release();
		lpdi = NULL;
	}
}

bool C_input::SetupDirectX(void)
{
	if(FAILED(DirectInputCreate(winfo.hInstance, DIRECTINPUT_VERSION, &lpdi, NULL)))
	{
		errstring = "[DInput] Failed at DirectInputCreate";
		return false;
	}

	/* get keyboard device */

	if(FAILED(lpdi->CreateDevice(GUID_SysKeyboard, &lpdiKeyb, NULL)))
	{
		errstring = "[DInput] Failed at CreateDevice for GUID_SysKeyboard";
		return false;
	}

	if(FAILED(lpdiKeyb->SetCooperativeLevel(winfo.hWnd, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE)))
	{
		errstring = "[DInput] Failed at SetCooperativeLevel for lpdiKeyb";
		return false;
	}

	if(FAILED(lpdiKeyb->SetDataFormat(&c_dfDIKeyboard)))
	{
		errstring = "[DInput] Failed at SetDataFormat for lpdiKeyb";
		return false;
	}

	if(FAILED(lpdiKeyb->Acquire()))
	{
		errstring = "[DInput] Failed at Acquire for lpdiKeyb";
		return false;
	}

	return true;
}

bool C_input::PlatformSetup(void)
{
	/* check config file here */
	if(electron->GetConfig()->FindSetting("MSWINDOWS_simpleinput", BOOLEAN_RESULT))
	{
		if(electron->GetConfig()->ReadSetting().result_bool)
			SetupDirectX();
	}
	else
		SetupDirectX();

	SetupJoysticks();

	return true;
}

int C_input::PlatformGetNumKeys(void)
{
	return MAX_KEY;
}

bool C_input::UpdateKeyStates(void)
{
	int c;
	unsigned __int32 *iptr;

	if(lpdiKeyb)
	{
		while(lpdiKeyb->GetDeviceState(256, array) == DIERR_INPUTLOST)
		{
			if(FAILED(lpdiKeyb->Acquire()))
				return false;
		}
	}
	else
	{
		unsigned __int8 temparray[256];

		GetKeyboardState(temparray);

		//now shuffle keyboard bytes for scancode ordering . . .
		c = 256;
		while(c--)
		{
			array[MapVirtualKey(c, 0)] = temparray[c];
		}
	}

	//now : recall that Microsoft only claims the msb is indicative of key state
	c = 64;
	iptr = (unsigned __int32 *)array;
	while(c--)
	{
		iptr[c] &= 0x80808080;
	}

	return true;
}

void C_input::UpdateLEDS(void)
{
	//cannot set the keyboard LEDs under Windows
}

void C_input::ClearBuffer(void)
{
	lastpos = qpos;
}

bool C_input::NewKey(void)
{
	return (qpos != lastpos) ? true : false;
}

unsigned char C_input::GetPressedASCII(void)
{
	unsigned char c;

	c = aqueue[lastpos];
	lastpos = (lastpos+1)&255;
	return c;
}

/* joystick bits */


void C_input::SetupJoysticks(void)
{
	/* to do : check if DirectX is functional, and maybe use that instead */
	JOYINFO jnk;
	JOYCAPS caps;

	int c = numsticks = joyGetNumDevs();
	while(c--)
	{
		if(joyGetPos(c, &jnk) != JOYERR_NOERROR)
			numsticks--;
	}

	sticks = (JOY_INFO *)malloc(sizeof(JOY_INFO)*numsticks);
	int cjoy = numsticks;
	c = joyGetNumDevs();
	while(c--)
	{
		if(joyGetPos(c, &jnk) == JOYERR_NOERROR)
		{
			cjoy--;

			joyGetDevCaps(c, &caps, sizeof(JOYCAPS));

			sticks[cjoy].topx = caps.wXmin;
			sticks[cjoy].xrange = caps.wXmax - caps.wXmin;

			sticks[cjoy].topy = caps.wYmin;
			sticks[cjoy].yrange = caps.wYmax - caps.wYmin;

			sticks[cjoy].id = c;
			sticks[cjoy].analogue = true;
			sticks[cjoy].calibrate = false;
		}
	}
}

JOY_INFO C_input::GetJoystickInfo(int id)
{
	return sticks[id];
}

JOY_STATE C_input::PlatformGetJoystickState(int id)
{
	/* possibly use DirectX here instead */
	JOYINFO info;
	JOY_STATE ret;

	joyGetPos(sticks[id].id, &info);

	ret.button = (info.wButtons & (JOY_BUTTON1 | JOY_BUTTON2 | JOY_BUTTON3 | JOY_BUTTON4)) ? true : false;

	ret.x = 128 - (((info.wXpos - sticks[id].topx) << 8) / sticks[id].xrange);
	ret.y = 128 - (((info.wYpos - sticks[id].topy) << 8) / sticks[id].yrange);

	/* calculate up/down/left/right now */
	ret.up = ret.y < -64;
	ret.down = ret.y > 64;
	ret.right = ret.x < -64;
	ret.left = ret.x > 64;

	return ret;
}

/* neither normal driver nor DirectX driver do calibration */

void C_input::BeginCalibration(int id)
{
}

const char *C_input::GetJoystickInstruction(void)
{
	return NULL;
}

void C_input::FlagJoystickDone(void)
{
}

#endif