/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	plus1/plus1.cpp
	===============

	This file is related to emulation of the Plus 1 Electron peripheral. The
	plus 1 contains :

		an ADC0844 from National Semiconductor (presumably Americans),
		a 6850 ACIA from Motorola &
		a simple printer port probably without a chip of any sort

	Only the ADC is actually emulated in this particular source file - it was
	too simple to bother splitting, but the parts called by the ULA are here,
	and they call the correct parts of the other files.
*/
#include "plus1.h"

#include "../misc/input.h"
#include "../ula/ula.h"
#include "../misc/config.h"

C_Plus1::C_Plus1(void)
{
	PRNout = NULL;
	PRNfname = NULL;
	PRNMode = PRNNextMode = PRNMODE_OFF;
}

C_Plus1::~C_Plus1(void)
{
	PRNClose();
}

bool C_Plus1::Setup(C_ULA *ul)
{
	ula = ul;
	ipt = ula->GetInput();
	fs = ula->GetFS();

	/* read config file */
	C_config *cfg = ula->GetConfig();

	if(cfg->FindSetting("plus1joy1", INTEGER_RESULT))
	{
		joyid[0] = cfg->ReadSetting().result_int;

		if(joyid[0] < 0 || joyid[0] >= ipt->GetNumJoysticks())
			joyid[0] = -1;
	}
	else
		joyid[0] = -1;

	if(cfg->FindSetting("plus1joy2", INTEGER_RESULT))
	{
		joyid[1] = cfg->ReadSetting().result_int;

		if(joyid[1] < 0 || joyid[1] >= ipt->GetNumJoysticks())
			joyid[1] = -1;
	}
	else
		joyid[1] = -1;

	ADCfintime = 0;

	return true;
}

void C_Plus1::Close(void)
{
	PRNClose();

	/* write to config file */
	config_setting s;
	C_config *cfg = ula->GetConfig();

	s.result_type = INTEGER_RESULT;

	s.result_int = joyid[0];
	cfg->FindSetting("plus1joy1", ANY_RESULT);
	cfg->WriteSetting(s);

	s.result_int = joyid[1];
	cfg->FindSetting("plus1joy2", ANY_RESULT);
	cfg->WriteSetting(s);
}

int Plus1_addrs[] = {0xfc70, 0xfc71, 0xfc72, 0x0000}; //add ACIA ports when sure

int *C_Plus1::GetAddressList(void)
{
	return Plus1_addrs;
}

void C_Plus1::WriteTo(BrokenWord addr, unsigned __int8 val8, unsigned __int32 val32, int time)
{
	switch(addr.a)
	{
		/* CONTROL */
		case 0xfc72 :
			/* writing to control register undefined */
		break;


		/* ADC */
		case 0xfc70 :
			/* does ADC ignore or what? */
			if(time < ADCfintime)
				return;

			ADCfintime = time + 80; //conversions take 40us = 80 cycles

			int plus, minus, channel;

			channel = val8&3;
			plus = ADCGetChannel(channel);

			switch(val8&12)
			{
				case 0 :
				case 8 : //differential

					minus = ADCGetChannel(channel^1);

				break;

				case 4 : //single ended

					minus = 0;

				break;

				case 12 : //pseudo-differential

					minus = ADCGetChannel(3);

				break;
			}
			if(minus > plus)
			{
				ADCresult = 0;
			}
			else
			{
				ADCresult = plus - minus;
			}
		break;


		/* printer */
		case 0xfc71 :
			PRNSendChar((char)val8);
		break;
	}
}

void C_Plus1::ReadFrom(BrokenWord addr, unsigned __int8 *val8, unsigned __int32 *val32, int time)
{
	switch(addr.a)
	{
		/* CONTROL */
		case 0xfc72 :

			*val8 = ADCGetStatus(time) | PRNGetStatus(time);

		break;


		/* ADC */
		case 0xfc70 :

			/*

			value is undefined while busy bit is still set, so this might as
			well do

			*/
			*val8 = ADCresult;
		break;

	}
}

void C_Plus1::MinimiseTimer(int sub)
{
	ADCfintime -= sub;
	if(ADCfintime < 0) ADCfintime = 0;
}

/* ADC */

void C_Plus1::SetJoystick(int elkid, int pcid)
{
	joyid[elkid] = pcid;
}

int C_Plus1::GetJoystick(int elkid)
{
	return joyid[elkid];
}

int C_Plus1::ADCGetChannel(int chan)
{
	if(joyid[(chan >> 1)&1] == -1) return 128;

	switch(chan)
	{
		case -1 :

			/* ground value */

		return 0;

		case 0 : return ipt->GetJoystickState(joyid[0]).x;
		case 1 : return ipt->GetJoystickState(joyid[0]).y;
		case 2 : return ipt->GetJoystickState(joyid[1]).x;
		case 3 : return ipt->GetJoystickState(joyid[1]).y;
	}

	return 0;
}

unsigned __int8 C_Plus1::ADCGetStatus(int time)
{
	/* in charge of these bits :

		0 ADC FB2 FB1 x x x x

	*/
	unsigned __int8 res;

	res = 0;

	res |= (joyid[0] != -1) ? (ipt->GetJoystickState(joyid[0]).button ? 0x00 : 0x10) : 0x10;
	res |= (joyid[1] != -1) ? (ipt->GetJoystickState(joyid[1]).button ? 0x00 : 0x20) : 0x20;

	if(time < ADCfintime) res |= 0x40;

	return res;
}

/* ACIA */

void C_Plus1::SetACIAMode(int m)
{
}

int C_Plus1::GetACIAMode(void)
{

	return 0;
}

bool C_Plus1::QueryACIAMode(int m)
{

	return false;
}

bool C_Plus1::SetURL(char *url)
{

	return false;
}

