/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This file at some time additionally worked upon by Paul Boddie

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	screen/screen.cpp
	=================

	Most of the general case screen stuff - when interrupts happen, starting
	the various pixel drawers and palette manipulators and so on.

	Actual pixel drawing is handled elsewhere by polymorphic classes. As in
	palette setting for the various shapes of table.

*/
#include <string.h>
#include <malloc.h>
#include <stdio.h>

#include "screen.h"
#include "screenloc.h"
#include "../defines.h"
#include "../6502/6502.h"
#include "../ula/ula.h"

#include "regular/big.h"
#include "regular/long.h"
#include "regular/small.h"
#include "regular/pal.h"

#include "wide/big1001.h"
#include "wide/long1001.h"
#include "wide/sml1001.h"
#include "wide/pal1001.h"

#define SetTiming() cpu->SetContentionTimes((line < 256) ? GfxTables[Mode].TimeLine : ThinLine);
#define SetPixelTiming() cpu->SetContentionTimes(GfxTables[Mode].TimeLine);
#define SetBorderTiming() cpu->SetContentionTimes(ThinLine);

/* static class members */

C_ScreenPels *C_screen::peldraw;
C_ScreenPalModifier *C_screen::palette;
int C_screen::line;
int C_screen::TimeBasis;
C_ULA *C_screen::ula;
unsigned __int8 C_screen::Mode;
BrokenQuad C_screen::WCols[16];

BrokenWord C_screen::VidAddr;

int C_screen::ModeChangeCount;
int C_screen::ModeChanges[128], *C_screen::ModeChangeCursor;

int C_screen::PalChangeCount;
int C_screen::PalChanges[384], *C_screen::PalChangeCursor;
int C_screen::PalCols[8], C_screen::PalOldCols[8];

C_6502Base *C_screen::cpu;

bool C_screen::blank;
int C_screen::coltab, C_screen::modecols;


/* this lot probably don't all need to be globals, but it'll do for now */

int NoLine[158], WideLine[158], ThinLine[158];

/*ModeTable GfxTables[] =
{
	{ false,	true,	20480,	8|1,	WideLine },
	{ false,	true,	20480,	8|2,	WideLine },
	{ false,	true,	20480,	8|3,	WideLine },
	{ true,		true,	16384,	8|1,	WideLine },

	{ false,	false,	10240,	1,		ThinLine },
	{ false,	false,	10240,	2,		ThinLine },
	{ true,		false,	8192,	1,		ThinLine }
};*/

ModeTable GfxTables[] =
{
	{ false,	true,	80,	8|1,	WideLine },
	{ false,	true,	80,	8|2,	WideLine },
	{ false,	true,	80,	8|3,	WideLine },
	{ true,		true,	64,	8|1,	WideLine },

	{ false,	false,	40,	1,		ThinLine },
	{ false,	false,	40,	2,		ThinLine },
	{ true,		false,	32,	1,		ThinLine }
};

struct MultiMode
{
	C_ScreenPels *low, *wide, *high;
	C_ScreenPels *alow, *awide, *ahigh;
	C_ScreenPalModifier *pal;
	bool hint16;
};

C_ScPelsBig					RBig;
C_ScPelsLong				RLong;
C_ScPelsSmall				RSmall;
C_ScreenPalModifierNormal	NormPal;

C_ScPelsBig1001				WBig;
C_ScPelsLong1001			WLong;
C_ScPelsSmall1001			WSmall;
C_ScreenPalModifier1001		WPal;

/*
	Index for MModes table is as in UEF file multiplexing description chunk
	0006. I.e. in binary

		abcd :		(d lsb)

			a set => 256 colour restriction
			b set => height improvement
			c set => width improvement
			d set => bpp improvement
*/

#define BM_NONE {&RSmall, &RLong, &RBig, NULL, NULL, NULL, &NormPal, true}
#define BM_BPP {&WSmall, &WLong, &WBig, NULL, NULL, NULL, &WPal, false}

#define BM_UNSUPPORTED {NULL}
#define BM_INVALID BM_UNSUPPORTED

MultiMode MModes[] =
{

	BM_NONE,
	BM_UNSUPPORTED, //bpp, but could be 16bpp so not supported. Full stop.
	BM_UNSUPPORTED, //width
	BM_UNSUPPORTED, //width & bpp
	BM_UNSUPPORTED, //height
	BM_UNSUPPORTED, //height & bpp
	BM_UNSUPPORTED, //width & height
	BM_INVALID,

	BM_NONE,
	BM_BPP,
	BM_UNSUPPORTED, //width
	BM_UNSUPPORTED, //width & bpp
	BM_UNSUPPORTED, //height
	BM_UNSUPPORTED, //height & bpp
	BM_UNSUPPORTED, //width & height
	BM_INVALID
};

#undef BM_INVALID
#undef BM_UNSUPPORTED

#undef BM_BPP
#undef BM_NONE

C_screen::C_screen(C_gfx_api *api)
{
	peldraw = NULL;
	palette = NULL;
	gfxptr = api;
	Mode = 0;

	ModeChangeCount = 0;
	ModeChangeCursor = ModeChanges;
	PalChangeCount = 0;
	PalChangeCursor = PalChanges;
}

C_screen::~C_screen(void)
{
	Kill();
}

void C_screen::Kill(void)
{
	if(peldraw)
	{
		peldraw->End();
		peldraw = NULL;
	}

	if(palette)
	{
		palette->Disable();
		palette = NULL;
	}
}

void C_screen::ResetCPU(void)
{
	cpu = ula->GetCPU();
	SetTiming();
}

void C_screen::Reset(void)
{
	float col_values[] =
	{
		1.0f, 0.6f, 0.5f, 0.0f
	};
	float r, g, b;
	int c = 8;

	while(c--)
		gfxptr->SetPaletteEntry(8+c, (c&4)?0.0f:1.0f, (c&2)?0.0f:1.0f, (c&1)?0.0f:1.0f, PAL_ELK);

	c = 64;
	while(c--)
	{
		r = col_values[((c&0x04) >> 2) | ((c&0x20) >> 4)];
		g = col_values[((c&0x02) >> 1) | ((c&0x10) >> 3)];
		b = col_values[((c&0x01) >> 0) | ((c&0x08) >> 2)];
		gfxptr->SetPaletteEntry(64|c, r, g, b, PAL_ELK);
	}

	gfxptr->FinishPalette(PAL_ELK);
}

bool C_screen::ValidMultiMode(unsigned __int8 mode)
{
	if(mode&0xf0) return false;

	return (MModes[mode].low == NULL) ? false : true;
}

bool C_screen::SetMultiMode(bool v, unsigned __int8 mult)
{
	bool ret = true;

	if(multiMode != mult || multiEnable != v)
	{
		multiMode = mult; multiEnable = v;

		multiPresent = multiEnable ? multiMode : 0;

		if(!ValidMultiMode(multiPresent))
		{
			multiPresent = 0;
			ret = false;
		}

		SetSize(wide, GetBitmap()->GetStretchLevel(), GetBitmap()->GetCleanScale());
	}

	return ret;
}

bool C_screen::GetMultiEnabled(void)
{
	return multiEnable;
}

void C_screen::SetSize(bool big, int scalelevel, bool scaletype)
{
	C_ScreenPels *newdraw;
	int nwidth, nheight;

	if(big)
	{
		newdraw = MModes[multiPresent].high;

		wide = true;
		nwidth = 640;
		nheight = 512;

		if(scalelevel>1)
		{
			nheight = 256;
			newdraw = MModes[multiPresent].wide;
		}
	}
	else
	{
		newdraw = MModes[multiPresent].low;

		wide = false;
		nwidth = 320;
		nheight = 256;
	}

	if(MModes[multiPresent].pal != palette)
	{
		if(palette)
			palette->Disable();

		palette = MModes[multiPresent].pal;
		palette->Enable(WCols);
	}

	if(newdraw != peldraw)
	{
		gfxptr->SetNormalResolution(nwidth, nheight);

		if(peldraw)
		{
			peldraw->End();
		}

		peldraw = newdraw;
		peldraw->Setup(ula, big ? ((scalelevel <= 1) ? rbig : rlong) : rsmall, GfxTables);
		peldraw->SetMode(Mode, 0);
	}

	GetBitmap()->SetStretchLevel(scalelevel, scaletype);
}

bool C_screen::GetSize(void)
{
	return wide;
}

char *C_screen::GetError(void)
{
	return errstring;
}

bool C_screen::Setup(C_ULA *electron, bool vidmode, bool cleanscale, int scalelevel, bool multiscr)
{
	int c;

	ula = electron;
	TimeBasis = 0;
	wide = vidmode;

	multiEnable = multiscr;

	if(!gfxptr->Setup(vidmode ? 640 : 320, vidmode && (scalelevel <= 1) ? 512 : 256, title_string, electron))
	{
		errstring = gfxptr->GetError();
		return false;
	}

	if(!gfxptr->LoadData("datapath", "newdata.uef", ula->GetFS()))
	{
		errstring = "invalid or non-existant data.uef";
		return false;
	}

	peldraw = NULL;

	rsmall = gfxptr->GetGenuineSize(320, 256);
	rbig = gfxptr->GetGenuineSize(640, 512);
	rlong = gfxptr->GetGenuineSize(640, 256);

	c = 8;
	while(c--)
	{
		gfxptr->SetPaletteEntry(8+c, (c&4)?0.0f:1.0f, (c&2)?0.0f:1.0f, (c&1)?0.0f:1.0f, PAL_GUI);
	}

	gfxptr->FinishPalette(PAL_ELK);
	gfxptr->SetPalette(PAL_ELK);

	memset(WCols, 8, 4*16);

	//Line is split :

	//		0			-> PEL_START	: left border
	//		PEL_START	-> PEL_START+80	: pixels
	//		PEL_START	-> end			: right border + overrun area

	c = 158;
	while(c--)
	{
		NoLine[c] = 1;
		WideLine[c] =
		ThinLine[c] = 2 | (c&1);
	}

	c = 80;
	while(c--)
	{
		WideLine[c+PEL_START] = 81-c;
	}

	line = 311;
	blank = false;

	//multiplexing screen mode '0' at first - since no data yet provided
	multiPresent =
	multiMode = 0;

	SetSize(vidmode, scalelevel, cleanscale);

	return true;
}

#define ReEvalBlankness()					\
	int c = modecols;						\
	if(blank = !multiPresent)				\
	{										\
		tptr = colptrs[coltab];				\
		while(c--)							\
			if(WCols[tptr[c]].b.b0 != 15)	\
			{								\
				blank = false;				\
				break;						\
			}								\
	}


#define RED 4
#define GREEN 2
#define BLUE 1

#define red0(x)		((x & 0x01) << 2)
#define red1(x)		((x & 0x02) << 1)
#define red2(x)		 (x & 0x04)
#define red3(x)		((x & 0x08) >> 1)

#define green2(x)	((x & 0x04) >> 1)
#define green3(x)	((x & 0x08) >> 2)
#define green4(x)	((x & 0x10) >> 3)
#define green5(x)	((x & 0x20) >> 4)

#define blue4(x)	((x & 0x10) >> 4)
#define blue5(x)	((x & 0x20) >> 5)
#define blue6(x)	((x & 0x40) >> 6)
#define blue7(x)	((x & 0x80) >> 7)

int coltable0[] = {8, 0};
int coltable1[] = {10, 8, 2, 0};
int coltable2[] = {15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0};
int *colptrs[] = {coltable0, coltable1, coltable2}, *tptr;

void C_screen::AffectPalEntry(int *palchange)
{
	//changes palette entry as described in palchange
	BrokenQuad *col0, *col1, *col2, *col3;

	palchange++;

	switch(palchange[0] >> 1)
	{
		case 0 :
			col0 = &WCols[0];	col1 = &WCols[2];	col2 = &WCols[8];	col3 = &WCols[10];
		break;

		case 1 :
			col0 = &WCols[4];	col1 = &WCols[6];	col2 = &WCols[12];	col3 = &WCols[14];
		break;

		case 2 :
			col0 = &WCols[5];	col1 = &WCols[7];	col2 = &WCols[13];	col3 = &WCols[15];
		break;

		case 3 :
			col0 = &WCols[1];	col1 = &WCols[3];	col2 = &WCols[9];	col3 = &WCols[11];
		break;
	}

	if(palchange[0]&1)
	{
		col0->b.b0 = ((col0->b.b0)&~(GREEN | RED)) | green4(palchange[1]) | red0(palchange[1]);
		col1->b.b0 = ((col1->b.b0)&~(GREEN | RED)) | green5(palchange[1]) | red1(palchange[1]);
		col2->b.b0 = ((col2->b.b0)&~RED) | red2(palchange[1]);
		col3->b.b0 = ((col3->b.b0)&~RED) | red3(palchange[1]);
	}
	else
	{
		col0->b.b0 = ((col0->b.b0)&~BLUE) | blue4(palchange[1]);
		col1->b.b0 = ((col1->b.b0)&~BLUE) | blue5(palchange[1]);
		col2->b.b0 = ((col2->b.b0)&~(BLUE | GREEN)) | blue6(palchange[1]) | green2(palchange[1]);
		col3->b.b0 = ((col3->b.b0)&~(BLUE | GREEN)) | blue7(palchange[1]) | green3(palchange[1]);
	}

	col0->b.b1 = col0->b.b2 = col0->b.b3 = col0->b.b0;
	col1->b.b1 = col1->b.b2 = col1->b.b3 = col1->b.b0;
	col2->b.b1 = col2->b.b2 = col2->b.b3 = col2->b.b0;
	col3->b.b1 = col3->b.b2 = col3->b.b3 = col3->b.b0;

	palette->SetPalette(0x8 + palchange[0], palchange[1]);

	ReEvalBlankness();
}

#undef blue7
#undef blue6
#undef blue5
#undef blue4

#undef green5
#undef green4
#undef green3
#undef green2

#undef red3
#undef red2
#undef red1
#undef red0

#undef BLUE
#undef GREEN
#undef RED

void C_screen::SetPalette(unsigned __int8 addr, unsigned __int8 val, int time)
{
	time -= TimeBasis;
	addr -= 0x8;

	if(val != 0xff && val != 0x55)
		PalOldCols[addr] = val;

	if(PalCols[addr] != val)
	{
		PalCols[addr] = val;

		*PalChangeCursor++ = time;
		*PalChangeCursor++ = addr;
		*PalChangeCursor++ = val;

		if(line >= 256 || peldraw->blank)
		{
			AffectPalEntry(PalChangeCursor-3);
			PalChangeCursor = PalChanges;
		}
		else
		{
			PalChangeCount++;
		}
	}
}

void C_screen::ChangeGFXMode(unsigned __int8 newmode, int time)
{
	if(newmode == 7) newmode = 4;

	if(newmode != Mode)
	{
		Mode = newmode;

		if(line >= 256 || peldraw->blank)
		{
			peldraw->SetMode(Mode, line);
		}
		else
		{
			SetPixelTiming();

			ModeChangeCount++;
			*ModeChangeCursor++ = time-TimeBasis;
			*ModeChangeCursor++ = Mode;
		}

		switch(newmode)
		{
			default :
				coltab = 0;
				modecols = 2;
			break;

			case 1 :
			case 5 :
				coltab = 1;
				modecols = 4;
			break;

			case 2 :
				coltab = 2;
				modecols = 16;
			break;
		}

		ReEvalBlankness();
	}
}

void C_screen::SaveImage(char *name)
{
	peldraw->SaveImage(name);
}

C_ScreenPalModifier *C_screen::GetPal(void)
{
	return palette;
}

#define DrawDefinitePelLine()					\
	if(!ModeChangeCount && !PalChangeCount)		\
	{											\
		if(blank)								\
			peldraw->DrawBlankLine(line, 15);	\
		else									\
			peldraw->DrawPelLine(line);			\
	}											\
	else										\
		peldraw->DrawMultiLine(line, ModeChanges, ModeChangeCount, PalChanges, PalChangeCount);

#define DrawLine()							\
	if(peldraw->blank)						\
	{										\
		SetBorderTiming();					\
		peldraw->DrawBlankLine(line, 15);	\
		peldraw->blank--;					\
		if(line == 249)						\
			peldraw->blank = 6;				\
	}										\
	else									\
	{										\
		DrawDefinitePelLine();				\
	}

int C_screen::Keepalive(int time)
{
	switch(line)
	{
		case 0 :
			peldraw->InitDisplay(VidAddr);

		case 1 :
		case 2 :
		case 3 :
		case 4 :
		case 5 :
		case 6 :
		case 7 :
			DrawDefinitePelLine();
		break;

		default :
			DrawLine();
		break;

		case 98 :
			DrawLine();
			ula->AddIRQ(ULAIRQ_RTC);
		break;

		case 255 :
			DrawLine();
			SetBorderTiming();
			peldraw->EndDisplay();
			ula->AddIRQ(ULAIRQ_DISPLAY);
		break;

		case 256 :	case 257 :	case 258 :	case 259 :
		case 260 :	case 261 :	case 262 :	case 263 :	case 264 :
		case 270 :	case 271 :	case 272 :	case 273 :	case 274 :
		case 280 :	case 281 :	case 282 :	case 283 :	case 284 :
		case 290 :	case 291 :	case 292 :	case 293 :	case 294 :
		case 300 :	case 301 :	case 302 :	case 303 :	case 304 :
		case 265 :	case 266 :	case 267 :	case 268 :	case 269 :
		case 275 :	case 276 :	case 277 :	case 278 :	case 279 :
		case 285 :	case 286 :	case 287 :	case 288 :	case 289 :
		case 295 :	case 296 :	case 297 :	case 298 :	case 299 :
		case 305 :	case 306 :	case 307 :	case 308 :	case 309 :
		case 310 :
		break;

		case 311 :
			SetPixelTiming();
			line = -1;
		break;
	}

	ModeChangeCount = 0;
	ModeChangeCursor = ModeChanges;
	PalChangeCount = 0;
	PalChangeCursor = PalChanges;

	line++;
	TimeBasis += 128;
	return line;
/*	if(line == 288)
	{
		peldraw->InitDisplay(VidAddr);
	}

	if(line < 256)
	{
		if(peldraw->blank)
		{
			SetBorderTiming();

			peldraw->DrawBlankLine(line, 15);

			peldraw->blank--;
			if(line == 249)
				peldraw->blank = 6;
		}
		else
		{
			if(!ModeChangeCount && !PalChangeCount)
			{
				if(blank)
					peldraw->DrawBlankLine(line, 15);
				else
					peldraw->DrawPelLine(line);
			}
			else
				peldraw->DrawMultiLine(line, ModeChanges, ModeChangeCount, PalChanges, PalChangeCount);
		}
	}

	ModeChangeCount = 0;
	ModeChangeCursor = ModeChanges;
	PalChangeCount = 0;
	PalChangeCursor = PalChanges;

	line++;
	TimeBasis += 128;

	if(line == 99)
	{
		ula->AddIRQ(ULAIRQ_RTC);
		return line;
	}

	if(line == 256)
	{
		SetBorderTiming();
		peldraw->EndDisplay();
		ula->AddIRQ(ULAIRQ_DISPLAY);
		return line;
	}

	if(line == 312)
	{
		SetPixelTiming();

		line = 0;
	}

	return line;*/
}

void C_screen::MinimiseTimer(int sub)
{
	TimeBasis -= sub;
}

C_bitmap *C_screen::GetBitmap(void)
{
	return peldraw? peldraw->GetBitmap() : NULL;
}

void C_screen::AlertToInterrupt(int time)
{
	/* don't really know the condition here, so err on the side of caution */

	if(
		Mode == 4 &&
		(PalCols[0] == 0xff) &&
		(PalCols[1] == 0xff) &&
		PalOldCols[0] == PalOldCols[1] &&
		(PalOldCols[0] >> 4) == (PalOldCols[0]&0x0f)
	)
	{
		int sum = 0, count = 6;
		while(count--)
			sum += PalCols[count+2];

		if(!sum)
		{
			SetPalette(0x8, PalOldCols[0], time);
			SetPalette(0x9, PalOldCols[1], time);
		}
	}
}

