/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details.

	sound/Allegro/soundcard.cpp
	===========================

	Allegro sound card interface class. Offers both input and output, but with
	the caveat that Allegro secretly halts sound output whenever input occurs,
	an event about which nothing can be done

	Notice that an extra buffer of sound data, OutputMem, is maintained above
	and beyond what (DOS) Allegro offers, since when new data is required it
	usually can't be provided very promptly otherwise

*/

#include "../sounddevs.h"

#ifdef TRG_SND_ALLEGRO

#include <string.h>

#define BUFFER_SIZE 1024

void C_SoundCard::Setup(void)
{
	InputRemaining = InputAllocated =
	InputPause = OutputPause =
	OutputRemaining = OutputAllocated = 0;

	CanOutput = CanInput = false;
	OldMode = 0;

	if(install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL) < 0)
		return;

	CanOutput = true;
	OutputBuffer.length = BUFFER_SIZE;

	OutputMem = (unsigned __int8 *)malloc(BUFFER_SIZE << 1);
	memset(OutputMem, 127, BUFFER_SIZE << 1);
	OutputCursor = 0;

	if(install_sound_input(DIGI_AUTODETECT, MIDI_NONE))
		return;

	CanInput = true;
}

void C_SoundCard::Close(void)
{
	if(outstr)
	{
		stop_audio_stream(outstr);
		outstr = NULL;
	}

	if(InputBuffer.buffer)
	{
		free(InputBuffer.buffer);
		InputBuffer.buffer = NULL;
	}

	if(OutputMem)
	{
		free(OutputMem);
		OutputMem = NULL;
	}
}

void C_SoundCard::Keepalive(int clock)
{
	unsigned __int8 *tmp;

	/* check for new input or output requirement */
	if(outstr)
	{
		if(tmp = (unsigned __int8 *)get_audio_stream_buffer(outstr))
		{
			memcpy(tmp, &OutputMem[OutputCursor], BUFFER_SIZE);
			free_audio_stream_buffer(outstr);

			if(!OutputRemaining)
			{
				OutputCursor = (OutputCursor+BUFFER_SIZE)&((BUFFER_SIZE << 1) - 1);
				OutputBuffer.buffer = &OutputMem[OutputCursor];

				OutputRemaining = OutputAllocated;
				memset(OutputBuffer.buffer, 127, BUFFER_SIZE);
			}
		}
	}

	if(OldMode&1)
		if(read_sound_input(InputBuffer.buffer))
		{
			InputRemaining = InputAllocated;
		}
}

C_SoundCard::C_SoundCard(void)
{
	outstr = NULL;
	InputBuffer.buffer = NULL;
	OutputMem = NULL;
}

bool C_SoundCard::ConsiderReallocation(int newid)
{
	int Mode;

	/* detemine new mode */
	Mode = ((InputAllocated&~1)? 1 : 0) | ((OutputAllocated&~1) ? 2 : 0);

	/* move value of Mode to OldMode, leave Mode^OldMode in Mode */
	Mode ^= OldMode; OldMode ^= Mode;

	/* has input state changed ? */
	if(Mode&1)
	{
		/* is it newly enabled ? */
		if(OldMode&1)
		{
			set_sound_input_source(SOUND_INPUT_LINE);

			int parm = get_sound_input_cap_parm(44100, 8, 0);
			if(parm < 1)
				return false;

			InputBuffer.length = start_sound_input(44100, 8, 0);

			if(InputBuffer.buffer)
			{
				free(InputBuffer.buffer);
			}
			InputBuffer.buffer = (unsigned __int8 *)malloc(InputBuffer.length);
		}
		else
		{
			/* has changed, but isn't enabled => has ended */
			stop_sound_input();
		}
	}
	else
	{
		/* same as above, but for output */
		if(Mode&2)
		{
			if(OldMode&2)
			{
				if(!outstr)
				{
					outstr = play_audio_stream(BUFFER_SIZE, 8, 0, 44100, 255, 128);

					if(!outstr) return false;
				}
				OutputBuffer.buffer = NULL;
			}
			else
			{
				if(outstr)
				{
					stop_audio_stream(outstr);
					outstr = NULL;
				}
			}
		}
	}

	return true;
}

void C_SoundCard::Pause(int time)
{
	if(outstr)
	{
		stop_audio_stream(outstr);
		outstr = NULL;
		OutputBuffer.buffer = NULL;
		OutputRemaining = 0;
	}
}

void C_SoundCard::Unpause(int time)
{
	if((OutputAllocated&~1) && !outstr)
	{
		outstr = play_audio_stream(BUFFER_SIZE, 8, 0, 44100, 255, 128);
	}
}
#endif
