/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details.

	sound/soundfile.cpp
	===================

	This is the implementation of a C_SoundAPI inheritor which reads sound
	files - at this time supporting the following formats :

		Microsoft wave (wav)
		SUN/NeXT au
		Apple Audio Interchange File Format (aiff)
		raw (assumed to be 44100Hz 8bit unsigned bytes)

	It accepts no output, and offers input in 'BUFFER_LENGTH' batches. It only
	works with files that contain linear integer PCM encoding, but is happy
	with 8 or 16bit data of any frequency (it scales to 44.1Khz), signed or
	unsigned with an arbitrary number of channels

	mu-law decoding and/or floating point file support may be added at a later
	date, but that is unlikely
*/

#define BUFFER_LENGTH	1024
#define SHIFTER			8

#include "sounddevs.h"
#include "sound.h"
#include <malloc.h>
#include <math.h>
#include <string.h>

#define GetBEndDWord()	\
		qvalue.b.b3 = gzgetc(File);\
		qvalue.b.b2 = gzgetc(File);\
		qvalue.b.b1 = gzgetc(File);\
		qvalue.b.b0 = gzgetc(File);

#define GetBEndWord()	\
		qvalue.b.b1 = gzgetc(File);\
		qvalue.b.b0 = gzgetc(File);

#define GetLEndDWord()	\
		qvalue.b.b0 = gzgetc(File);\
		qvalue.b.b1 = gzgetc(File);\
		qvalue.b.b2 = gzgetc(File);\
		qvalue.b.b3 = gzgetc(File);

#define GetLEndWord()	\
		qvalue.b.b0 = gzgetc(File);\
		qvalue.b.b1 = gzgetc(File);

unsigned __int32 GetBEndExtended(gzFile File)
{
	/*

		This function assumes a sensible (integer representable) value is
		being read, so skips any NaN or range checks

	*/

	double result;
	BrokenQuad MantissaHigh, MantissaLow;
	BrokenWord Exponential;
	bool sign;

	Exponential.b.h = gzgetc(File);
	Exponential.b.l = gzgetc(File);
	sign = (Exponential.a&0x8000) ? true : false;
	Exponential.a &= 0x7fff;

	MantissaHigh.b.b3 = gzgetc(File);
	MantissaHigh.b.b2 = gzgetc(File);
	MantissaHigh.b.b1 = gzgetc(File);
	MantissaHigh.b.b0 = gzgetc(File);

	MantissaLow.b.b3 = gzgetc(File);
	MantissaLow.b.b2 = gzgetc(File);
	MantissaLow.b.b1 = gzgetc(File);
	MantissaLow.b.b0 = gzgetc(File);

	Exponential.a -= 16383;

	result = ldexp(MantissaHigh.a, Exponential.a-31);
	result += ldexp(MantissaHigh.a, Exponential.a-63);

	if(sign) result = -result;

	return (int)result;
}

void C_SoundFile::Setup(char *path, char *name, C_FSAccess *f)
{
	char *waveextensions[] = {"wav\a", "wave\a", NULL};
	char *auextensions[] = {"au\a", NULL};
	char *aiffextensions[] = {"aif\a", "aiff\a", NULL};
	char *rawextensions[] = {"raw\a", NULL};
	int SampleRate = 0;
	BrokenQuad qvalue;

	/* default state */
	fs = f;
	CanOutput = CanInput = false;
	paused = true;
	InputRemaining = InputAllocated =
	InputPause = OutputPause =
	OutputRemaining = OutputAllocated = 0;

	/* raw audio format */
	if(fs->ExtensionIncluded(name, rawextensions))
	{
		if(!(File = fs->GetGZ(path, NULL, name, "rb")))
			return;

		DataOffset = 0;
		Length = 0xffffffff;
		SampleRate = 44100;
		Pace = 1;

		Sign = 0;

		CanInput = true;
	}

	/* SUN AU format */
	if(fs->ExtensionIncluded(name, auextensions))
	{
		if(!(File = fs->GetGZ(path, NULL, name, "rb")))
			return;

		/* check magic number */
		char buf[5];
		gzread(File, buf, 4);
		buf[4] = '\0';

		if(strcmp(buf, ".snd"))
		{
			fs->ReleaseGZ(File);
			File = NULL;
			return;
		}

		/* read offset in octets - big endian style */
		GetBEndDWord();
		DataOffset = qvalue.a << 3;

		/* read length */
		GetBEndDWord();
		Length = qvalue.a << 3;

		/* format */
		GetBEndDWord();
		switch(qvalue.a)
		{
			default :
				fs->ReleaseGZ(File);
				File = NULL;
			return;

			case 2 : Pace = 1; break; //8bit PCM
			case 3 : Pace = 2; break; //16bit PCM
			case 4 : Pace = 3; break; //24bit PCM
			case 5 : Pace = 4; break; //32bit PCM
		}

		/* sample rate, not in octets */
		GetBEndDWord();
		SampleRate = qvalue.a;

		/* num channels */
		GetBEndDWord();
		Pace *= qvalue.a;

		/* fact */
		Sign = 0x80;
		CanInput = true;
	}

	/* Apple AIFF format */
	if(fs->ExtensionIncluded(name, aiffextensions))
	{
		DataOffset = 0;

		if(!(File = fs->GetGZ(path, NULL, name, "rb")))
			return;

		/* check string id */
		char buf[13];
		gzread(File, buf, 12);
		buf[12] = buf[4] = '\0';

		if(strcmp(buf, "FORM") || strcmp(&buf[8], "AIFF"))
		{
			fs->ReleaseGZ(File);
			File = NULL;
			return;
		}

		/* tiny chunk parser */
		unsigned __int32 clen;
		while(!gzeof(File))
		{
			gzread(File, buf, 4);
			GetBEndDWord(); clen = qvalue.a;

			if(!strcmp(buf, "COMM"))
			{
				/* num channels */
				GetBEndWord();
				Pace = qvalue.w.w0.a;

				/* num samples */
				GetBEndDWord();

				/* sample size */
				GetBEndWord();
				Pace *= qvalue.w.w0.a >> 3;

				/* frequency */
				SampleRate = GetBEndExtended(File);

				/* fact */
				Sign = 0x80;
			}
			else
			if(!strcmp(buf, "SSND"))
			{
				Length = clen - 8;
				DataOffset = 8 + gzseek(File, clen, SEEK_CUR) - clen;
			}
			else
				gzseek(File, clen, SEEK_CUR);
		}

		CanInput = true;
	}

	/* Microsoft wave format */
	if(fs->ExtensionIncluded(name, waveextensions))
	{
		DataOffset = 0;

		if(!(File = fs->GetGZ(path, NULL, name, "rb")))
			return;

		/* check string id */
		char buf[13];
		gzread(File, buf, 12);
		buf[12] = buf[4] = '\0';

		if(strcmp(buf, "RIFF") || strcmp(&buf[8], "WAVE"))
		{
			fs->ReleaseGZ(File);
			File = NULL;
			return;
		}

		/* tiny chunk parser */
		unsigned __int32 clen;
		while(!gzeof(File))
		{
			gzread(File, buf, 4);
			if(gzeof(File)) break;
			GetLEndDWord(); clen = qvalue.a;

			if(!strcmp(buf, "fmt "))
			{
				/* data format */
				GetLEndWord();

				if(qvalue.w.w0.a != 1) /* uncompressed PCM is type 1 */
				{
					fs->ReleaseGZ(File);
					File = NULL;
					return;
				}

				/* num channels */
				GetLEndWord();
				Pace = qvalue.w.w0.a;

				/* sampling frequency */
				GetLEndDWord();
				SampleRate = qvalue.a;

				/* skip : average bytes per second & block alignment count */
				gzseek(File, 6, SEEK_CUR);

				/* bits per sample */
				GetLEndWord();
				Pace *= qvalue.w.w0.a >> 3;

				Sign = (qvalue.w.w0.a == 8) ? 0x00 : 0x80;

				/* little endian, so . . . */
				DataOffset += Pace-1;
			}
			else
			if(!strcmp(buf, "data"))
			{
				Length = clen;
				DataOffset += gzseek(File, clen, SEEK_CUR) - clen;
			}
			else
				gzseek(File, clen, SEEK_CUR);
		}

		CanInput = true;
	}

	/*

	finally calculate the Incrementer by referring to SampleRate, and reset
	Position to 0

	*/

	if(CanInput)
	{
		Position = 0;
		Incrementer = (SampleRate << SHIFTER) / 44100;
		if(Length != 0xffffffff)
			Length /= Pace;

		gzseek(File, DataOffset, SEEK_SET);
		DataLatch = gzgetc(File) ^ Sign;

		InputBuffer.length = BUFFER_LENGTH;
		InputBuffer.buffer = (unsigned __int8 *)malloc(sizeof(unsigned __int8)*BUFFER_LENGTH);
		TimeBase = 0;
	}

}

C_SoundFile::C_SoundFile(void)
{
	File = NULL;
	InputBuffer.buffer = NULL;
}

void C_SoundFile::Close(void)
{
	if(File)
	{
		fs->ReleaseGZ(File);
		File = NULL;
	}

	if(InputBuffer.buffer)
	{
		free(InputBuffer.buffer);
		InputBuffer.buffer = NULL;
	}

}

void C_SoundFile::Keepalive(int clock)
{
	if(!paused && File)
	if(CircularDifference(clock, TimeBase, SOUND_MASK+1) >= BUFFER_LENGTH)
	{
		TimeBase = (TimeBase+BUFFER_LENGTH)&SOUND_MASK;
		InputRemaining = InputAllocated;

		/* fill new buffer */
		int c = BUFFER_LENGTH;
		unsigned __int8 *bptr = InputBuffer.buffer;
		unsigned __int32 oldpos;

		while(c--)
		{
			*bptr++ = DataLatch;

			oldpos = Position;
			Position += Incrementer;

			if((Position >> SHIFTER) != (oldpos >> SHIFTER))
			{
				if(Length == 0xffffffff)
				{
					if(gzeof(File))
						gzseek(File, DataOffset, SEEK_SET);
				}
				else
				{
					if((Position >> SHIFTER) == Length)
						gzseek(File, DataOffset, SEEK_SET);
				}

				DataLatch = gzgetc(File) ^ Sign;
			}
		}
	}
}

void C_SoundFile::MinimiseTimer(int sub)
{
	TimeBase = (TimeBase-sub)&SOUND_MASK;
}

bool C_SoundFile::Control(int command, int value)
{
	switch(command)
	{
		default : return false;

		case SOUNDCONTROL_SETTIME :

			unsigned int offset = DataOffset;

			offset += ((value*Incrementer) >> SHIFTER)*Pace;
			gzseek(File, offset, SEEK_SET);

			DataLatch = gzgetc(File) ^ Sign;
		return true;
	}
}

void C_SoundFile::Pause(int time)
{
	paused = true;
}

void C_SoundFile::Unpause(int time)
{
	TimeBase = time;
	paused = false;
}

