/*

	ElectrEm (c) 2000 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details.

	misc/sound.cpp
	==============

	The sound driver, platform independant parts. All very simple. When in
	buffered mode, 8bit unsigned samples are generated and sent to the sound
	output device as and when it will take them. When in real time mode, any
	requests to the Electron to change frequency or switch on or off are
	simply sent straight to the sound output device. Notice that in the
	second case because of the nature of the timer mechanism that this means
	at the very most change accuracy will be to the nearest scanline, but in
	actual fact with the current target set (Linux console, DOS/Allegro &
	DirectX at time of writing) it is much less just because of the
	combinations of timer accuracy and sound output devices / libraries.

*/
#include "speaker.h"
#include "sound.h"
#include "../ula/ula.h"
#include <math.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>

C_Speaker::C_Speaker(void)
{
	start = current = all;
	bufadd = 128;
	cwave = Silence;
	basenum = 0;
	lfwave = NULL;
	lfwavelen = 0;
	outid = 0;
	card = NULL;
}

C_Speaker::~C_Speaker(void)
{
	if(lfwave)
	{
		free(lfwave);
		lfwave = NULL;
	}
}

bool C_Speaker::Setup(C_SoundProvider *s, bool acc, float vol)
{
	snd = s;

	outid = 0;
	card = snd->GetSoundSource(NULL, SOUNDCARD_NAME, NULL);
	if(card) outid = card->RequestAccess(SoundMode_Output);
	if(!outid) return false;

	accuracy = acc;
	SetVolume(vol);

	return true;
}

void C_Speaker::Close(void)
{
	if(outid)
	{
		card->ReturnAccess(outid);
		outid = 0;
	}

	if(card)
	{
		snd->ReleaseSoundSource(card);
		card = NULL;
	}
}

inline void C_Speaker::UpdatePointer(FreqChange **ptr)
{
	(*ptr)++;
	if((*ptr) > &all[2205])
		(*ptr) = all;
}

void C_Speaker::SetFrequency(int freq, int time)
{
	current->direct = false;
	current->newval = freq;
	current->time = time;
	UpdatePointer(&current);
}

void C_Speaker::DirectSetFrequency(int freq, int time)
{
	current->direct = true;
	current->newval = freq;
	current->time = time;
	UpdatePointer(&current);
}

void C_Speaker::SetVolume(float v)
{
	int c;
	int vol;

	fvol = v;
	vol = (int)(127.0f*v);

	c = 64;
	while(c--)
	{
		SquareWave[c] = ((c&32) ? (-vol) : (vol));
		Silence[c] = - vol;
		HighPitch[0][c] = vol;
		HighPitch[1][c] = -vol;
		SineWave[c] = (signed __int8)( (float)vol*sin( ((float)c*3.141592654f)/32.0f ) );
	}
}

float C_Speaker::GetVolume(void)
{
	return fvol;
}

void C_Speaker::SetMode(int mode, int time)
{
	current->newval = mode;
	current->time = time;
	UpdatePointer(&current);
}

void C_Speaker::Pause(int time)
{
	if(outid)
		card->Pause(outid,time);
}

void C_Speaker::Unpause(int time)
{
	if(outid)
		card->Unpause(outid, time);
}

void C_Speaker::SetAccuracy(bool high)
{
	accuracy = high;
}

bool C_Speaker::GetAccuracy(void)
{
	return accuracy;
}

inline void C_Speaker::ProcessEvent(FreqChange *cpos)
{
	if(cpos->newval < 0)
	{
		switch(cpos->newval)
		{
			case -1 : cwave = Silence; break;
			case -2 : cwave = (bufadd > 8192) ? HighPitch[0] : SquareWave; break;
			case -3 : cwave = SineWave; bufpos = 0; break;
		}
	}
	else
	{
		if(cpos->direct)
		{
			bufadd = (256 * 64 * cpos->newval) / 44100;
		}
		else
		{
			bufadd = 6400000 / ((cpos->newval + 1) * 551);
		}

		if(cwave == HighPitch[0]) cwave = SquareWave;

		if(bufadd > 8192 && cwave == SquareWave)
		{
			cwave = HighPitch[0];
		}
	}
}

void C_Speaker::Resync(int time)
{
	while(start != current)
	{
		ProcessEvent(start);
		UpdatePointer(&start);
	}
	basenum = time;
}

void C_Speaker::Keepalive(int time)
{
	unsigned __int8 *SndBuffer;
	FreqChange *cpos;
	int ctime, bufsize;
	int difference;

	difference = CircularDifference(time, basenum, SOUND_MASK+1);

	if(difference >= 1024)
		cpos = NULL;

	SoundBuffer *buf = card->PollNewBuffer(outid, difference);

	if(buf)
	{
		SndBuffer = buf->buffer;
		bufsize = buf->length;

		if(accuracy)
		{
			cpos = start;
			ctime = basenum;

			//if we are more than two 'bufsizes' behind (assumption : bufsize is
			//relatively constant), flush buffer

			if(difference >= (bufsize*2))
			{
				int aim;

				aim = (time-bufsize)&SOUND_MASK;

				while(ctime != aim)
				{
					while((cpos != current) && (cpos->time == ctime))
					{
						ProcessEvent(cpos);
						UpdatePointer(&cpos);
					}

					ctime = (ctime+1)&SOUND_MASK;
				}
			}

			//now fill the current buffer

			while(bufsize--)
			{
				while((cpos != current) && (cpos->time == ctime) )
				{
					ProcessEvent(cpos);
					UpdatePointer(&cpos);
				}

				bufpos += bufadd;
				*SndBuffer++ += cwave[(bufpos >> 8)&63];

				ctime = (ctime+1)&SOUND_MASK;
			}

			start = cpos;
			basenum = ctime;
		}
		else
		{
			bool refill = false;
			signed __int8 *bptr;

			if(buf->length != lfwavelen)
			{
				lfwavelen = buf->length;
				lfwave = (signed __int8 *)realloc(lfwave, lfwavelen);
				refill = true;
			}

			if(start != current)
			{
				refill = true;

				while(start != current)
				{
					ProcessEvent(start);
					UpdatePointer(&start);
				}
			}

			basenum = time;
			bptr = lfwave;

			if(refill)
				while(bufsize--)
				{
					bufpos += bufadd;
					*bptr++ = 127 + cwave[(bufpos >> 8)&63];
				}

			memcpy(buf->buffer, lfwave, buf->length);
		}

		card->ReturnBuffer(outid);
	}
}

void C_Speaker::MinimiseTimer(int sub)
{
	card->MinimiseTimer(sub);
	basenum = (basenum - sub)&SOUND_MASK;

	FreqChange *tmp = start;
	while(tmp != current)
	{
		tmp->time = (tmp->time-sub)&SOUND_MASK;
		UpdatePointer(&tmp);
	}
}
