/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	tape/fasttape.cpp
	=================

	This is the file that handles fast tape loading via PC trapping and some
	virtual OPs.

	This code makes use of the bitin provider to return chunks of tape data -
	i.e. it is now UEF independant, and works equally well with any filetype
	that can return blocks of bytes.

	THE ROM FILE FORMAT :

		1 byte synchronisation. (&2A *)
		Filename (1 to 10 characters)
		Filename terminator (&00)
		Load address of file, four bytes, low byte first.		[+0]
		Execution address of file, four bytes, low byte first.	[+4]
		Block number, two bytes, low byte first.				[+8]
		Block length, two bytes, low byte first.				[+10]

		Block flag, one byte:									[+11]
			Bit 0 Protection bit.
			Bit 6 Empty block bit.
			Bit 7 Last block bit.

		Four spare bytes (&00)									[+12]

		Header CRC, two bytes, high byte first.					[+16]

		[data, the specified number of bytes]					[+18]

		Two byte data CRC, high byte first
*/

#include <ctype.h>
#include "../ula/ula.h"
#include "../6502/6502.h"
#include "../sound/speaker.h"
#include "../misc/timing.h"
#include <stdio.h>
#include "tape.h"
#include "tapemodes.h"
#include <malloc.h>

/*

BuildTable builds a table of all the available tape file blocks which
conform to the OS ROM tape filing system standard shape and form

*/
void C_tape::BuildTable(void)
{
	int stringlength;
	bool valid = false;
	BrokenWord crc;
	FileDataBlock *tapedata;

	FreeTable();
	if(!bitin->SetFirstBlock()) return;

	while(tapedata = bitin->NextBlock())
	{
		if(tapedata->length > 18)
		{
			valid = false;
			stringlength = StrLen((char *)(tapedata->data+1), 10);

			if(stringlength >= 0)
			{
				crc = GetCRC(tapedata->data+1, stringlength+18);

				if(crc.b.l == tapedata->data[stringlength+1+19] && crc.b.h == tapedata->data[stringlength+1+18])
					valid = true;
			}

			if(valid)
			{
				if(!list)
				{
					current = list = (FileBlock *)malloc(sizeof(FileBlock));
				}
				else
				{
					current->next = (FileBlock *)malloc(sizeof(FileBlock));
					current = current->next;
				}

				current->header = (unsigned __int8 *)malloc(31);
				memcpy(current->header, tapedata->data, 31);

				current->name = (char *)(current->header+1);
				current->header = (unsigned __int8 *)(current->name + StrLen(current->name, 10) + 1);

				current->next = NULL;

				current->blocknum = current->header[8] | (current->header[9] << 8);
				current->last = (current->header[12]&0x80) ? true : false;
				current->empty = (current->header[12]&0x40) && (tapedata->length == strlen(current->name)+21) ? true : false;

				current->locked = (current->header[12]&0x01) ? true : false;
				current->offset = bitin->GetOffset();
			}
		}
	}
}

/*

FreeTable clears memory

*/
void C_tape::FreeTable(void)
{
	FileBlock *nxt;

	current = list;
	while(current)
	{
		nxt = current->next;

		free(current->name-1);
		free(current);
		current = nxt;
	}

	current = list = NULL;
}

/*

GetCRC returns data CRC's, as by the Acorn described algorithm

*/
BrokenWord C_tape::GetCRC(unsigned __int8 *start, int len)
{
	int x;
	BrokenWord crc;

	crc.a = 0;
	while(len--)
	{
		crc.b.h ^= *start++;

		x = 8;
		while(x--)
			crc.a = (crc.b.h&0x80) ? (((crc.a^0x0810) << 1) | 1) : (crc.a << 1);
	}

	return crc;
}

/*

ForgeFootprint adjusts some RAM work areas so that it looks like the ROM
loaded the data, even though it didn't

*/
void C_tape::ForgeFootprint(void)
{
	int addr = 0x3b2;
	int c = 0, c2;

	while(current->name[c])
	{
		elk->GetCPU()->WriteByte(addr, current->name[c]);
		addr++; c++;
	}
	elk->GetCPU()->WriteByte(addr, 0);

	addr = 0x3be;
	c2 = 21;
	c = 0;
	while(c2--)
	{
		elk->GetCPU()->WriteByte(addr, current->header[c]);
		addr++;
		c++;
	}
}

/*

returns true if the current file block is not the last of the current file

*/
bool C_tape::FileContinues(void)
{
	//test 1 - is there even a next block?
	if(!current->next) return false;

	//test 2 - does this block have the last block bit set, and the next not
	//have the very next blocknum?
	if(current->last && (current->next->blocknum != current->blocknum+1)) return false;

	//test 3 (Elite test!) - is the next block not sequentially numbered?
	if(current->next->blocknum != current->blocknum+1) return false;

	return true;
}

/*

finds the start of the next file

*/
void C_tape::NextFile(void)
{
	if(current)
	{
		while(FileContinues())
		{
			current = current->next;
		}

		current = current->next;
	}

	if(!current) current = list;
}

/*

does a partial StrCmp, 'similar' to that done by the ROM

*/
bool C_tape::StrCmp(char *one, char *two)
{
	bool ret = false;
	unsigned int len;
	char *t;

	len = 0;
	t = two;
	while(*t && *t != ' ')
	{
		len++;
		t++;
	}

	if(strlen(one) < len) return true;

	while(len--)
	{
		if(tolower(*one) != tolower(*two))
			ret = true;

		one++;
		two++;
	}

	return ret;
}

/*

like strlen, but never goes above 'maxlen'

*/
int C_tape::StrLen(char *str, int maxlen)
{
	char *str2 = str;

	while(*str2 && (str2-str) <= maxlen)
		str2++;

	if(*str2)
		return -1;
	else
		return str2 - str;
}

#define NOTAPE_ERROR "No tape or not UEF\n\r"

/*

FindFile is used to find the file called 'name', returning true on success

*/
bool C_tape::FindFile(char *name, bool strings)
{
	bool ret = false;
	int c = 0;

	if(strings) AddString("Searching\n\r");
	if(!list)
	{
		AddString(NOTAPE_ERROR);
		return false;
	}

	//now, a quick clause to make Acornsoft games work
	if(name[0] == '|')			//0x7c / [0xb4 -> 0x94] [0x77 -> 0x0d]
		sprintf(name, "");

	//and the actual 'find' bit
	do
	{
		NextFile();
		if(current == list) c++;
	}
	while((c<2) && strlen((char *)name) && StrCmp((char *)current->name, (char *)name));

	if(c < 2)
	{
		if(strings) AddString("\n\rLoading\n\r\n\r");
		return true;
	}
	else
	{
		if(strings) AddString("File not found\n\r\n\r");
	}

	return false;
}

/*

InitialiseOutput prepares the correct buffers for printing the 'Loading',
'Searching', etc messages

*/
void C_tape::InitialiseOutput(void)
{
	outcursor = outputtext;
}

/*

AddString adds a string to the output message queue

*/
void C_tape::AddString(char *output)
{
	if((int)strlen(output) > outend - outcursor)
	{
		int len, offset;

		len = outend - outputtext;
		offset = outcursor - outputtext;

		do
		{
			len += 8192;
		}
		while(len < (outcursor-outputtext)+(int)strlen(output));

		outputtext = (char *)realloc(outputtext, len);
		outend = outputtext + len;
		outcursor = outputtext + offset;
	}

	memcpy(outcursor, output, strlen(output));
	outcursor += strlen(output);
	*outcursor = '\0';
}

/*

EndOutput signals that all of the output that is going to be generated for
now has been, and passes the lot on to the ULA

*/
void C_tape::EndOutput(void)
{
	elk->WriteTapeString(outputtext);
	elk->PageTape();
}

/*

CleansePrint prints filenames from one string to another, replacing control
characters with question marks

*/
void CleansePrint(char *targ, char *source)
{
	while(*source)
	{
		*targ++ = (*source >=32) ? *source : '?';
		source++;
	}
}

/*

LoadNextBlock loads the next file block into RAM

*/
void C_tape::LoadNextBlock(void)
{
	int c;
	char name[22];
	C_UEFChunk *wide = NULL;
	BrokenQuad v;

	FileDataBlock *tapedata;
	unsigned __int8 *tapeptr, *wtapeptr;

	ForgeFootprint();

	//store (latest) execute address
	FileInfo[6] = current->header[4];
	FileInfo[7] = current->header[5];
	FileInfo[8] = current->header[6];
	FileInfo[9] = current->header[7];

	c = current->empty ? 0 : ( current->header[10] | ((int)current->header[11] << 8) );
	FileLength += c;

	tapedata = bitin->SetOffset(current->offset);
	tapeptr = tapedata->data + 21 + strlen(current->name);
	if(wtapeptr = tapedata->wdata)
	{
		wtapeptr += (21 + strlen(current->name)) << 2;
	}

	while(c--)
	{
		elk->GetCPU()->WriteByte(FileAddr.a, *tapeptr); tapeptr++;

		if(wtapeptr)
		{
			v.b.b0 = *wtapeptr; wtapeptr++;
			v.b.b1 = *wtapeptr; wtapeptr++;
			v.b.b2 = *wtapeptr; wtapeptr++;
			v.b.b3 = *wtapeptr; wtapeptr++;

			elk->GetCPU()->WriteDWord(FileAddr.a, v.a);
		}

		FileAddr.a++;
	}

	if(FileContinues())
	{
		sprintf(name, "\r           %02x", current->blocknum&255);
		CleansePrint(&name[1], current->name);
		current = current->next;
	}
	else
	{
		sprintf(name, "\r           %02x %04x\a\n\r", current->blocknum&255, FileLength);
		CleansePrint(&name[1], current->name);

		//store file length
		FileInfo[10] = FileLength & 0xff;
		FileInfo[11] = (FileLength >> 8)&0xff;
		FileInfo[12] = (FileLength >> 16)&0xff;
		FileInfo[13] = (FileLength >> 24)&0xff;

		//set &E2 for 'nothing open / not CAT / no warning'
		elk->GetCPU()->WriteByte(0xe2, 0x00);

		TapeCommand = NextCmd;
	}

	InitialiseOutput();
	AddString(name);
	elk->WriteTapeString(outputtext);

	ForgeFootprint();

}

void C_tape::BeginLoad(unsigned __int8 *info, TapeCommands nxt)
{
	FileInfo = info;
	FileLength = 0;
	NextCmd = nxt;

	//determine load address
	if(info[6])
	{
		FileAddr.b.l = info[2] = current->header[0];
		FileAddr.b.h = info[3] = current->header[1];
	}
	else
	{
		FileAddr.b.l = info[2];
		FileAddr.b.h = info[3];
	}

	//store load address
	info[2] = current->header[0];
	info[3] = current->header[1];
	info[4] = current->header[2];
	info[5] = current->header[3];

	//store 'file attributes' (aren't any on a tape, unless locked does something?)
	info[14] =
	info[15] =
	info[16] =
	info[17] = 0;

	//set &E2 for 'something open / not CAT / no warning'
	elk->GetCPU()->WriteByte(0xe2, 0x01);

	TapeCommand = LOAD;
}

char *C_tape::GetFilename(BrokenWord addr, char *terminators)
{
	char *name;
	unsigned char *s_start;
	int c, qc = 0;
	char *t;
	bool end = false;

	name = (char *)malloc(256);
	c = 0;
	s_start = elk->GetCPU()->GetByteReadPtr(addr)-1;

	do
	{
		s_start++;

		t = terminators;
		while(*t)
		{
			if(*s_start == *t) end = true;
			t++;
		}

		if(*s_start != '"' && !end)
		{
			name[c] = *s_start;
			c++;
		}
		else
			qc++;

	}
	while(!end && *s_start && (qc<2));

	name[c] = '\0';

	return name;
}

void C_tape::ReadOSFileByte(void)
{
	BrokenQuad v;

	/* write newest byte into A, and possibly multiplexed A */
	if(!OSFileData)
	{
		OSFileData = bitin->SetOffset(current->offset);
		OSFileOffset = 21 + strlen(current->name);
	}

	elk->GetCPU()->WriteByte(NMOS6502ADDR_A, OSFileData->data[OSFileOffset]);
	if(OSFileData->wdata)
	{
		v.b.b0 = OSFileData->wdata[(OSFileOffset << 2) + 0];
		v.b.b1 = OSFileData->wdata[(OSFileOffset << 2) + 1];
		v.b.b2 = OSFileData->wdata[(OSFileOffset << 2) + 2];
		v.b.b3 = OSFileData->wdata[(OSFileOffset << 2) + 3];
		elk->GetCPU()->WriteDWord(NMOS6502ADDR_A, v.a);
	}

	OSFileOffset++;
	/* the last two bytes are the data CRC */
	if(OSFileOffset == OSFileData->length-2)
	{
		if(!FileContinues())
			OSFileEOF = true;
		else
		{
			current = current->next;
			OSFileData = NULL;
		}
	}
}

bool C_tape::EnterService(unsigned __int8 command)
{
	char *name;
	BrokenWord addr, blockaddr;
	bool do_load;
	unsigned __int8 a, x, y, p;

	a = elk->GetCPU()->ReadByte(NMOS6502ADDR_A);
	x = elk->GetCPU()->ReadByte(NMOS6502ADDR_X);
	y = elk->GetCPU()->ReadByte(NMOS6502ADDR_Y);

	switch(command)
	{
		case 0 :	//OSFILE
			/*
				Partly implemented: entry with A=0 is save, all other entries
				are considered to be 'load file' (A=&FF).

				ElectrEm : only loading is supported
			*/
			if(a)
			{
				blockaddr.a = (y << 8) | x;
				addr.b.l = elk->GetCPU()->ReadByte(blockaddr.a);
				blockaddr.a++;
				addr.b.h = elk->GetCPU()->ReadByte(blockaddr.a);
				blockaddr.a--;

				name = GetFilename(addr, "\r ");

				InitialiseOutput();

				if(FindFile((char *)name, true))
				{
					BeginLoad(elk->GetCPU()->GetByteReadPtr(blockaddr), LOADEND);
				}
				else
					elk->GetCPU()->WriteByte(NMOS6502ADDR_A, 0);

				EndOutput();
				free(name);

				return true;
			}
			else
				return false;
		break;

		case 1 : //OSARGS

			/*
				Hardly implemented: only filing system identification
				performed (A=0, Y=G).

				ElectrEm : ROM is allowed to deal with this one
			*/

		return false;

		case 2 : //OSBGET

			/*
				fully implemented

				ElectrEm : fully implemented
			*/
			if(OSFileEOF || y != 1)
			{
				elk->GetCPU()->WriteByte(NMOS6502ADDR_A, 0xfe);
				p = elk->GetCPU()->ReadByte(NMOS6502ADDR_P) | NMOS6502FLAG_C;
				elk->GetCPU()->WriteByte(NMOS6502ADDR_P, p);

				return true;
			}

			p = elk->GetCPU()->ReadByte(NMOS6502ADDR_P) & ~NMOS6502FLAG_C;
			elk->GetCPU()->WriteByte(NMOS6502ADDR_P, p);
			ReadOSFileByte();

		return true;

		case 3 : //OSBPUT

			/*
				fully implemented

				ElectrEm : only loading is supported, so the ROM deals with
				this one
			*/

		return false;

		case 4 : //OSGBPB

			/*
				not implemented at all

				ElectrEm : ROM is allowed to deal with this one
			*/

		return false;

		case 5 : //OSFIND

			/*
				Implemented, allowing one file to be open for input, and one
				for output. If an attempt is made to open a file for update,
				it is opened for input.

				The file handles given are constant: 1 is the input file; 2
				is the output file.

				ElectrEm : only opening and closing of input is allowed
			*/
			if(a&0x40)
			{
				/* open file */
				addr.b.h = y;
				addr.b.l = x;
				name = GetFilename(addr, "\r");

				if(FindFile((char *)name, false))
				{
					elk->GetCPU()->WriteByte(NMOS6502ADDR_A, 1);
					OSFileEOF = false;
					OSFileData = NULL;
				}
				else
					elk->GetCPU()->WriteByte(NMOS6502ADDR_A, 0);

				return true;
			}

			if(!a && y == 1)
			{
				/* close input file */
				return true;
			}

		return false;

		case 6 : //OSFSC
			/*
				Calls 0 through 6 are implemented, though no extra commands
				exist (ie. code 3 just gives 'Bad command').

				ElectrEm : calls 2 & 4 (*RUN), 5 (*CAT) are fully implemented,
				and call 1 (EOF) is allowed on the input file
			*/
			switch(a)
			{
				default : return false;

				case 1 :
					if(x == 1)
					{
						elk->GetCPU()->WriteByte(NMOS6502ADDR_X, OSFileEOF ? 0xff : 0x00);
						return true;
					}
				return false;

				case 2 :
				case 4 : //LOAD AND RUN

					blockaddr.b.h = y;
					blockaddr.b.l = x;
					name = GetFilename(blockaddr, "\r");
					do_load = true;

					InitialiseOutput();

					if(FindFile(name, true))
					{
						LoadBuffer[6] = 1;
						BeginLoad(LoadBuffer, RUNEND);
					}

					EndOutput();
					free(name);
				return true;

				case 5 : //*CAT
					char name[50];
					int loadaddr;
					int length;
					FileBlock *i = current;

					InitialiseOutput();

					if(current = list)
					{
						do
						{
							loadaddr = current->header[0] | (((int)current->header[1]) << 8);
							length = current->empty ? 0 : ( current->header[10] | ((int)current->header[11] << 8) );
							while(FileContinues())
							{
								current = current->next;
								length += current->empty ? 0 : ( current->header[10] | ((int)current->header[11] << 8) );
							}

							sprintf(name, "           %04x LOAD:%04x EXEC:%02x%02x\n\r", length, loadaddr, current->header[5], current->header[4]);
							CleansePrint(name, current->name);

							AddString(name);
							NextFile();
						}
						while(current != list);
					}
					else
						AddString(NOTAPE_ERROR);

					current = i;

					EndOutput();
				return true;
			}
		break;
	}

	return false;
}

void C_tape::HandleVOP(unsigned __int8 command, int samptime, int clocktime)
{
	switch(TapeCommand)
	{
		default : break;

		case LOADEND :
			elk->GetCPU()->WriteByte(NMOS6502ADDR_A, 1);

			elk->GetTimer()->Resync();
			elk->GetSpeaker()->Resync(samptime);
		break;

		case RUNEND :
			elk->GetCPU()->Push((FileInfo[6] | (FileInfo[7] << 8))-1);

			if(current->locked)
			{
				//set memory clear on break via OSBYTE &c8
				elk->GetCPU()->WriteByte(NMOS6502ADDR_Y, 0);
				elk->GetCPU()->WriteByte(NMOS6502ADDR_X, 3);
				elk->GetCPU()->WriteByte(NMOS6502ADDR_A, 0xc8);

				//notice that the actual ROM address of OSBYTE
				//is used rather than the memory redirected
				//pointer since some games (about one) load
				//over it
				elk->GetCPU()->Push(0xe544-1);
			}

			elk->GetTimer()->Resync();
			elk->GetSpeaker()->Resync(samptime);
		break;
	}

	TapeCommand = NONE;
	elk->UnPageTape();
}

void C_tape::AtStringEnd(void)
{
	switch(TapeCommand)
	{
		case NONE : break;
		case LOAD : LoadNextBlock(); break;
	}
}

/* Autoload ID */
unsigned __int8 C_tape::GetAutoloadID(void)
{
	unsigned __int8 *tpos;
	int skip, length;
	FileDataBlock *datablock;

	unsigned __int8 bret = 0x00, cret = 0x01;

	if(!list) return 0xff;

	datablock = bitin->SetOffset(list->offset);

	// first examing the first 32 bytes to figure out how long the header is
	// determine number of bytes to skip and length to read
	skip = StrLen((char *)datablock->data+1, 10)+2;
	length = datablock->data[skip+10];
	length |= (int)datablock->data[skip+11] << 8;

	// assume that if locked *RUN rather than CHAIN is the order of the day,
	// even for BASIC
	if(datablock->data[skip+12]&1)
		bret = 0x01;

	// add remainder of fixed header size
	skip += 19;

	// check if it looks like a BASIC program
	tpos = &datablock->data[skip];
	while(length > 4)
	{
		if(tpos[0] != 13) return cret;
		if((tpos[1]&0x7f) == 0x7f) return bret;

		length -= tpos[3];
		tpos += tpos[3];
	}

	return (length > 4)? 0xff : bret;
}
