/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	tape/tape.cpp
	=============

	The genuine tape hardware emulation. Nowadays where the bits come from
	is completely removed into some polymorphic classes.

*/

#include <ctype.h>
#include "../ula/ula.h"
#include "../6502/6502.h"
#include "../misc/timing.h"
#include <stdio.h>
#include "tape.h"
#include "tapemodes.h"
#include <malloc.h>

#include <string.h>
#include "../misc/config.h"
#include "../fs/fs.h"
#include <stdio.h>


/* this exchanges the current input device */
void C_tape::SetBitIn(C_TapeBitProvider *b, int samptime, int clocktime)
{
	/* stop last sound source */
	if(bitin)
	{
		bitin->Enable(false, false, samptime, clocktime);
		if(IOLoop)
		{
			if(UEFOut && bitin != uef) uef->Enable(false, false, samptime, clocktime);
			if(WaveOut && bitin != card) card->Enable(false, false, samptime, clocktime);
		}
	}

	/* establish new bitin */
	bitin = b;

	/* set new flags */
	bitin->Enable(true, false, samptime, clocktime);

	if(IOLoop)
	{
		if(UEFOut && bitin != uef)
			uef->Enable(true, true, samptime, clocktime);

		if(WaveOut && bitin != card)
			card->Enable(true, true, samptime, clocktime);
	}
}

/* intialisation */

C_tape::C_tape(void)
{
	TapeCommand = NONE;
	current = list = NULL;
	bitsubcounter = 0;
	elk = NULL;

	uef = NULL;
	file = NULL;
	card = NULL;
	t2p3 = NULL;

	list = NULL;
	OSFileData = NULL;
}

bool C_tape::Setup(C_ULA *electron)
{
	elk = electron;

	uef = new C_TapeBitProviderUEF;
	uef->Setup(electron);

	t2p3 = new C_TapeBitProviderT2P3;
	t2p3->Setup(electron);

	file = new C_TapeBitProviderWave;
	file->Setup(electron);

	card = new C_TapeBitProviderWave;
	card->Setup(electron);
	card->SetFile(NULL, SOUNDCARD_NAME, 0, 0);

	filein = uef;
	bitin = NULL;
	speaker = elk->GetSpeaker();

	if(elk->GetConfig()->FindSetting("tapeinmode", INTEGER_RESULT))
	{
		switch(elk->GetConfig()->ReadSetting().result_int)
		{
			default :
			case 0 : SetBitIn(uef, 0, 0); break;
			case 1 : SetBitIn(card, 0, 0); break;
		}
	}
	else
		SetBitIn(uef, 0, 0);

	UEFOut = WaveOut = IOLoop = false;

	if(elk->GetConfig()->FindSetting("tapeoutuef", BOOLEAN_RESULT))
		UEFOut = elk->GetConfig()->ReadSetting().result_bool;

	if(elk->GetConfig()->FindSetting("tapeoutcard", BOOLEAN_RESULT))
		WaveOut = elk->GetConfig()->ReadSetting().result_bool;

	if(elk->GetConfig()->FindSetting("tapeioloop", BOOLEAN_RESULT))
		IOLoop = elk->GetConfig()->ReadSetting().result_bool;

	outputtext = (char *)malloc(8192);
	outend = outputtext + 8192;

	return true;
}

/*

signal reset stops the tape interface from doing anything for about 3 seconds
in order to get around some ROM 'deficiencies'

*/
void C_tape::SignalReset(void)
{
}

C_tape::~C_tape(void)
{
}

void C_tape::Closedown(void)
{
	/*

	this top line is a quick way of checking whether Setup was ever actually
	done

	*/
	if(!elk) return;

	Close();

	config_setting s;

	s.result_type = INTEGER_RESULT;
	s.result_int = (bitin == card) ? 1 : 0;
	elk->GetConfig()->FindSetting("tapeinmode", ANY_RESULT);
	elk->GetConfig()->WriteSetting(s);

	s.result_type = BOOLEAN_RESULT;
	s.result_bool = UEFOut;
	elk->GetConfig()->FindSetting("tapeoutuef", ANY_RESULT);
	elk->GetConfig()->WriteSetting(s);

	s.result_bool = WaveOut;
	elk->GetConfig()->FindSetting("tapeoutcard", ANY_RESULT);
	elk->GetConfig()->WriteSetting(s);

	s.result_bool = IOLoop;
	elk->GetConfig()->FindSetting("tapeioloop", ANY_RESULT);
	elk->GetConfig()->WriteSetting(s);

	if(uef)
	{
		uef->CloseFile();
		uef->Close();
		delete uef;
		uef = NULL;
	}

	if(t2p3)
	{
		t2p3->CloseFile();
		t2p3->Close();
		delete t2p3;
		t2p3 = NULL;
	}

	if(card)
	{
		card->CloseFile();
		card->Close();
		delete card;
		card = NULL;
	}

	if(file)
	{
		file->CloseFile();
		file->Close();
		delete file;
		file = NULL;
	}
}

/* this one detects whether a particular UEF has useful tape chunks */
bool C_tape::Useful(C_UEFChunkSelector *selector)
{
	selector->Reset();
	return selector->FindIdMajor(0x01);
}

/* this generates a new UEF tape */
bool C_tape::New(char *path, char *nme, int samptime, int clocktime)
{
	/* create a new tape UEF, write a gap of 0 ms, close it, and call Open */
	C_UEFChunkSelector *tmp;
	char *name = elk->GetFS()->GetCleanFilename(nme, "uef");
	bool res;

	if(!(tmp = elk->GetFS()->RequestUEFChunkSelector(path, NULL, name, "w")))
		return false;

	tmp->EstablishChunk();
	tmp->CurrentChunk()->SetId(0x0112);
	tmp->CurrentChunk()->PutC(0);
	tmp->CurrentChunk()->PutC(0);

	elk->GetFS()->ReleaseUEFChunkSelector(tmp);

	res = Open(path, name, samptime, clocktime);
	free(name);
	return res;
}

/* Opens a new file, and hence possibly changes the bit provider */
bool C_tape::Open(char *path, char *name, int samptime, int clocktime)
{
	char *uefexts[] = {"uef", NULL};
	char *t2p3exts[] = {"t2p\a", "t2p3\a", "t2peg\a", NULL};
	char *tapeexts[] = {"tape\a", "tap\a", NULL};
	bool loaded = false;

	if(elk->GetFS()->ExtensionIncluded(name, uefexts))
	{
		filein = uef;

		uef->CloseFile();
		loaded = uef->SetFile(path, name, samptime, clocktime);
	}

	if(elk->GetFS()->ExtensionIncluded(name, t2p3exts))
	{
		filein = t2p3;

		t2p3->CloseFile();
		loaded = t2p3->SetFile(path, name, samptime, clocktime);
	}

	if(!loaded)
	{
		/* assume to be a wave file of some description */
		filein = file;

		file->CloseFile();
		file->SetFile(path, name, samptime, clocktime);
	}

	if(bitin != card)
	{
		SetBitIn(filein, samptime, clocktime);
	}

	if(bitin == filein)
	{
		bitcount = 0;
		TapeInData = false;
	}

	BuildTable();
	return true;
}

void C_tape::Close(void)
{
	FreeTable();
}

void C_tape::Rewind(int samptime, int clocktime)
{
	bitin->MoveTo(0, samptime, clocktime);
	if(current = list)
	{
		while(current->next)
			current = current->next;
	}
}

void C_tape::Keepalive(int samptime, int clocktime)
{
	if(counter)
	{
		if(tape_read)
		{
			TapeBitBuffer *bit = bitin->PollNewBuffer(samptime, clocktime), newbit;

			if(motor_on && bit)
			{
				TapeData8 = (TapeData8 >> 1) | (bit->data8 << 9);

				if(!bit->data8 && !TapeInData)
				{
					TapeInData = true;
					bitcount = 1;
				}

				bitcount++;
				if(bitcount == 10)
				{
					if(TapeData8&2)
					{
						TapeInData = false;
					}

					if(!bit->gap)
					{
						elk->AddIRQ(TapeInData ? ULAIRQ_SEND : ULAIRQ_HTONE);
					}

					if(IOLoop)
					{
						newbit.data8 = TapeData8 >> 2;
						newbit.data32 = TapeData32 >> 8;
						newbit.use32 = false;
						newbit.hightone = !TapeInData;
						newbit.gap = bit->gap;

						if(UEFOut && (uef != bitin)) uef->WriteNewBuffer(samptime, clocktime, &newbit);
						if(WaveOut && (card != bitin)) card->WriteNewBuffer(samptime, clocktime, &newbit);
					}

					bitcount = 0;
				}
			}
		}
		else if(!bitsubcounter--)
		{
			TapeBitBuffer newbit;

			bitsubcounter = 12;

			if(bitcount < 10)
			{
				if(!bitcount) elk->RemoveIRQ(ULAIRQ_SEND);

				if(bitcount == 9)
				{
					/* push an entire byte */
					newbit.data8 = (unsigned __int8)TapeData8;
					newbit.data32 = TapeData32;
					newbit.hightone = newbit.gap = false;

					if(UEFOut) uef->WriteNewBuffer(samptime, clocktime, &newbit);
					if(WaveOut) card->WriteNewBuffer(samptime, clocktime, &newbit);

					dirty = true;
					elk->AddIRQ(ULAIRQ_RECEIVE);
				}
			}
			else
			{
				if(!(bitcount%10))
				{
					newbit.hightone = true;
					newbit.gap = false;

					if(UEFOut) uef->WriteNewBuffer(samptime, clocktime, &newbit);
					if(WaveOut) card->WriteNewBuffer(samptime, clocktime, &newbit);
				}

				dirty = true;
				elk->AddIRQ(ULAIRQ_RECEIVE);
			}

			bitcount++;
		}
	}

	uef->Keepalive(samptime, clocktime);
	card->Keepalive(samptime, clocktime);
	file->Keepalive(samptime, clocktime);
	t2p3->Keepalive(samptime, clocktime);
}

void C_tape::WriteByte(unsigned __int8 value, unsigned __int32 bv, int samptime, int clocktime)
{
	bitcount = 0;
	elk->RemoveIRQ(ULAIRQ_RECEIVE);
	TapeData8 = value;
	TapeData32 = bv;
}

void C_tape::ReadByte(unsigned __int8 *value, unsigned __int32 *bv, int samptime, int clocktime)
{
	elk->RemoveIRQ(ULAIRQ_SEND);
	if(bitcount == 1)
		*value = TapeData8 >> 1;
	else
		*value = TapeData8 >> 2;
	*bv = TapeData32;
}

void C_tape::SetMode(bool tcounter, bool read, bool motor, int samptime, int clocktime)
{
	/* first consider fixing fast tape abilities */
	if(dirty)
		BuildTable();
	dirty = false;

	counter = tcounter;
	tape_read = read;
	motor_on = motor;

	if(uef) uef->SetMode(counter, read, motor, samptime, clocktime);
	if(card) card->SetMode(counter, read, motor, samptime, clocktime);
	if(file) file->SetMode(counter, read, motor, samptime, clocktime);
	if(t2p3) t2p3->SetMode(counter, read, motor, samptime, clocktime);
}

bool C_tape::GetOperationMode(int mode)
{
	switch(mode)
	{
		case TAPE_INPUTMODE_FILE :
		return (bitin == filein) ? true : false;

		case TAPE_INPUTMODE_SOUND :
		return (bitin == card) ? true : false;

		case TAPE_INPUTMODE_SPEEDUP :
		return list ? true : false;

		case TAPE_OUTPUTMODE_UEF :		return UEFOut;
		case TAPE_OUTPUTMODE_SOUND :	return WaveOut;
		case TAPE_IO_PIPING :			return IOLoop;
	}

	return false;
}

void C_tape::SetOperationMode(int mode, bool s, int samptime, int clocktime)
{
	switch(mode)
	{
		case TAPE_INPUTMODE_FILE :		SetBitIn(filein, samptime, clocktime);	break;
		case TAPE_INPUTMODE_SOUND :		SetBitIn(card, samptime, clocktime);	break;
		case TAPE_OUTPUTMODE_UEF :		UEFOut = s;								break;
		case TAPE_OUTPUTMODE_SOUND :	WaveOut = s;							break;
		case TAPE_IO_PIPING :			IOLoop = s;								break;
	}
}

void C_tape::MinimiseTimer(int samp, int clock)
{
	uef->MinimiseTimer(samp, clock);
	file->MinimiseTimer(samp, clock);
	card->MinimiseTimer(samp, clock);
	t2p3->MinimiseTimer(samp, clock);
}


