#ifndef TAPEMODES_H
#define TAPEMODES_H

#include "../types.h"
#include "../sound/sound.h"
#include <zlib.h>

class C_FSAccess;
class C_ULA;
class C_Speaker;
class C_UEFChunkSelector;
class C_UEFChunk;

struct TapeBitBuffer
{
	unsigned __int8 data8;
	unsigned __int32 data32;
	bool use32;

	bool gap, hightone;
};

struct FileDataBlock
{
	unsigned __int8 *data, *wdata;
	unsigned int length;
};

class C_TapeBitProvider
{
	public :
		/* for normal read/write access */
		virtual TapeBitBuffer *PollNewBuffer(int samptime, int clocktime) = 0;
		virtual void WriteNewBuffer(int samptime, int clocktime, TapeBitBuffer *) = 0;

		/* mode setting */
		virtual void SetMode(bool counter, bool read, bool motor, int samptime, int clocktime) = 0;
		virtual void Enable(bool, bool, int samptime, int clocktime);

		/* virtual media manipulation */
		virtual void MoveTo(int time, int samptime, int clocktime) = 0;

		/* keepalive bits */
		virtual void Keepalive(int samptime, int clocktime);
		virtual void MinimiseTimer(int subsamp, int subclock) = 0;

		/* fast media manipulation */
		virtual bool SetFirstBlock(void);
		virtual FileDataBlock *NextBlock(void);
		virtual int GetOffset(void);
		virtual FileDataBlock *SetOffset(int offs);

	protected :
		bool Inverse, Enabled;
		bool Paused, OpMode, Counter;
};

class C_TapeBitProviderUEF : public C_TapeBitProvider
{
	public :
		/* UEF specific bits */
		bool Setup(C_ULA *);
		void Close(void);
		bool SetFile(char *path, char *name, int samptime, int clocktime);
		void CloseFile(void);

		/* read/write stuff */
		TapeBitBuffer *PollNewBuffer(int samptime, int clocktime);
		void WriteNewBuffer(int samptime, int clocktime, TapeBitBuffer *);

		/* media manipulation */
		void MoveTo(int time, int samptime, int clocktime);

		/* mode manipulation */
		void SetMode(bool counter, bool read, bool motor, int samptime, int clocktime);

		/* timer stuff */
		void MinimiseTimer(int sampsub, int clocksub);

		/* fast load stuff */
		bool SetFirstBlock(void);
		FileDataBlock *NextBlock(void);
		int GetOffset(void);
		FileDataBlock *SetOffset(int offs);

	private :
		C_FSAccess *fs;
		C_ULA *ula;
		C_UEFChunkSelector *file, *fastfile;

		/* these variables handle reading */
		C_UEFChunk *chunk, *chunkmulti;
		void FindNextChunk(int samptime, int clocktime);
		unsigned __int8 idminor;
		int timetarg, bitoffset, targetoffset1, targetoffset2;

		int bitlength, baudrate;

		unsigned __int8 DataLatch8;
		unsigned __int32 DataLatch32;

		static TapeBitBuffer back;
		static FileDataBlock blockdata;

		/* these variables handle writing */
		C_UEFChunk *OutChunk, *OutChunkMulti;
		BrokenWord OutLength;
		unsigned __int8 WriteID;
		void CloseChunk(void);
		void NewChunk(unsigned __int8 idminor);
		FileDataBlock *ReturnBlock(void);
};

class C_TapeBitProviderT2P3 : public C_TapeBitProvider
{
	public :
		/* T2P3 specific bits */
		bool Setup(C_ULA *);
		void Close(void);
		bool SetFile(char *path, char *name, int samptime, int clocktime);
		void CloseFile(void);

		/* read/write stuff */
		TapeBitBuffer *PollNewBuffer(int samptime, int clocktime);
		void WriteNewBuffer(int samptime, int clocktime, TapeBitBuffer *);

		/* media manipulation */
		void MoveTo(int time, int samptime, int clocktime);

		/* mode manipulation */
		void SetMode(bool counter, bool read, bool motor, int samptime, int clocktime);

		/* timer stuff */
		void MinimiseTimer(int sampsub, int clocksub);

		/* fast load stuff */
		bool SetFirstBlock(void);
		FileDataBlock *NextBlock(void);
		int GetOffset(void);
		FileDataBlock *SetOffset(int offs);

	private :
		C_FSAccess *fs;
		gzFile file;

		/* these variables handle reading */
		int BlockStartOffset, BlockEndOffset;
		int BlockEnterTime, BlockNum;
		bool OverRan;

		static TapeBitBuffer back;
		static FileDataBlock blockdata;

		void DetermineBlockStats(void);
		FileDataBlock *ReturnBlock(void);
};

class C_TapeBitProviderWave : public C_TapeBitProvider
{
	public :
		/* wave specific bits */
		bool Setup(C_ULA *);
		void Close(void);

		bool SetFile(char *path, char *name, int samptime, int clocktime);
		void CloseFile(void);

		/* read/write stuff */
		TapeBitBuffer *PollNewBuffer(int samptime, int clocktime);
		void WriteNewBuffer(int samptime, int clocktime, TapeBitBuffer *);

		/* media manipulation */
		void MoveTo(int time, int samptime, int clocktime);

		/* mode manipulation */
		void SetMode(bool counter, bool read, bool motor, int samptime, int clocktime);

		/* keepalive stuff */
		void Keepalive(int samptime, int clocktime);
		void MinimiseTimer(int sampsub, int clocksub);

	private :
		/* general variables */
		C_SoundProvider *snd;
		C_SoundAPIBase *source;
		C_Speaker *speaker;
		SoundID id;

		/* input variables */
		int InputLoops[16];
		int InputLoopCount;
		int InputTimeBase;

		TapeBitBuffer InputValue;

		bool InputBlocked;
		int FindFarExtrema(int, int);
		int FindNextExtrema(int, int);

		unsigned __int8 *BufMem;
		unsigned __int8 *Peak, *Trough;
		int ReadCursor, WriteCursor, ZeroCursor, TroughCursor, PeakCursor;

		/* output variables */
		void WriteBit(int samptime, __int8 bit);
		signed __int8 sinewave[64];
		int sinepos, sineadd;

};
#endif
