/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	tape/tapet2p3.cpp
	=================

	This file handles streams of bits from Slogger T2??? tape files. Writing
	is not allowed.

	It is assumed that every file block appears on the tape in one of these
	ways :

	block 0 :

		1 second of gap
		3 seconds of high tone
		data
		0.6 seconds of high tone

	all other blocks :

		data
		0.6 seconds of high tone

	Notice that these files can only store strictly valid ROM format tapes, so
	these assumptions should be valid.

*/
#include "tapemodes.h"
#include "../ula/ula.h"
#include "../fs/fs.h"
#include "../uef/uef.h"
#include <stdlib.h>

/*

for reasons known to Slogger only, all data bytes in a Slogger T2 file
are excluisve-or'd with &5a/90. This macro clears all that up

*/
#define T2Value(v) (v^0x5a)

#define ConvClockTime(x) ((x * 3) / 5000)
int timetarg;

bool C_TapeBitProviderT2P3::Setup(C_ULA *ul)
{
	fs = ul->GetFS();

	file = NULL;
	blockdata.wdata = NULL;

	Paused = true;
	Counter = false;
	OpMode = false;

	return true;
}

void C_TapeBitProviderT2P3::Close(void)
{
	CloseFile();
}

bool C_TapeBitProviderT2P3::SetFile(char *path, char *name, int samptime, int clocktime)
{
	CloseFile();

	if(!(file = fs->GetGZ(path, NULL, name, "rb")))
	{
		return false;
	}

	MoveTo(0, samptime, clocktime);
	OpMode = true;

	return true;
}

void C_TapeBitProviderT2P3::CloseFile(void)
{
	if(file)
	{
		fs->ReleaseGZ(file);
		file = NULL;
	}

	if(blockdata.data)
	{
		free(blockdata.data);
		blockdata.data = NULL;
	}

	if(blockdata.wdata)
	{
		free(blockdata.wdata);
		blockdata.wdata = NULL;
	}
}

TapeBitBuffer C_TapeBitProviderT2P3::back;

TapeBitBuffer *C_TapeBitProviderT2P3::PollNewBuffer(int samptime, int clocktime)
{
	int TimeBase = BlockEnterTime;
	int bit, byte;

	if(!file || Paused || !OpMode) return NULL;

	clocktime = ConvClockTime(clocktime);
	/* now, TimBase & clocktime are 1200 baud counters */

	if(clocktime > timetarg)
	{
		timetarg = clocktime;
		back.gap = false;

		if(!BlockNum)
		{
			/* 1 second gap */
			if(clocktime < TimeBase+1200)
			{
				back.gap = true;
				return &back;
			}

			TimeBase += 1200;

			/* maybe do lead-in tone of 3 seconds high tone */
			if(clocktime < TimeBase+3600)
			{
				back.data8 = 1;
				return &back;
			}

			TimeBase += 3600;
		}

		bit = clocktime - TimeBase;
		byte = bit / 10;

		if(byte >= BlockEndOffset-BlockStartOffset)
		{
			/* 0.6 seconds high tone after every block */

			if(clocktime > TimeBase+((BlockEndOffset-BlockStartOffset)*10)+720)
			{
				/* prepare the next block if the time is here */
				BlockStartOffset = BlockEndOffset;
				BlockEnterTime = timetarg+1;
				DetermineBlockStats();
			}

			back.data8 = 1;
			return &back;
		}

		bit = bit%10;
		switch(bit)
		{
			case 0 :
				back.data8 = 0;
			break;

			case 9 :
				back.data8 = 1;
			break;

			default :
				gzseek(file, byte+BlockStartOffset, SEEK_SET);
				byte = T2Value(gzgetc(file));

				bit--;
				back.data8 = (byte >> bit)&1;
			break;
		}

		return &back;
	}
	else
		return NULL;
}

void C_TapeBitProviderT2P3::WriteNewBuffer(int samptime, int clocktime, TapeBitBuffer *b)
{
	if(OpMode || Paused || !file) return;

	/* write new data */
}

void C_TapeBitProviderT2P3::MoveTo(int time, int samptime, int clocktime)
{
	/* treated as necessarily rewind for now */
	BlockStartOffset = 5;
	BlockEnterTime = ConvClockTime(clocktime);
	timetarg = BlockEnterTime+1;
	DetermineBlockStats();
}

void C_TapeBitProviderT2P3::MinimiseTimer(int sampsub, int clocksub)
{
	if(!Paused)
	{
		BlockEnterTime -= ConvClockTime(clocksub);
		timetarg -= ConvClockTime(clocksub);
	}
}

void C_TapeBitProviderT2P3::SetMode(bool cntr, bool read, bool motor, int samptime, int clocktime)
{
	bool CurrentEnabled = Enabled;

	if(Inverse)
	{
		if(read = !read)
		{
			/* if more is write, disable */
			CurrentEnabled = false;
		}
	}

	if(!CurrentEnabled)
	{
		/* effect a disabling by switching off the counter */
		cntr = false;
	}

	/* store counter mode */
	Counter = cntr;

	/* store & process motor mode */
	if(motor && Paused)
	{
		BlockEnterTime += ConvClockTime(clocktime);
		timetarg = ConvClockTime(clocktime)+1;
	}

	if(!motor && !Paused)
	{
		BlockEnterTime -= ConvClockTime(clocktime);
	}

	Paused = !motor;

	/* store read/write mode */
	OpMode = read;
}

/* block related thing */
void C_TapeBitProviderT2P3::DetermineBlockStats(void)
{
	int travlen = 0, len, c;
	int BlockEndOffset2;

	if(file)
	{
		/* move to start of block, check EOF */
		gzseek(file, BlockStartOffset, SEEK_SET);

		c = gzgetc(file);
		if(T2Value(c) != 0x2a)
		{
			BlockStartOffset = 5;
			gzseek(file, 6, SEEK_SET);
			OverRan = true;
		}

		travlen++;

		c = 10;
		/* find filename terminator */
		while(c-- && T2Value(gzgetc(file)))
		{
			travlen++;
		}

		/* for NULL byte */
		travlen++;

		/* seek on 8 bytes to get block number */
		gzseek(file, 8, SEEK_CUR);
		travlen += 8;

		BlockNum = T2Value(gzgetc(file));
		BlockNum += T2Value(gzgetc(file)) << 8;
		travlen += 2;

		len = T2Value(gzgetc(file));
		len += T2Value(gzgetc(file)) << 8;
		BlockEndOffset = BlockStartOffset + len + travlen + 11;

		if(!(gzgetc(file)&0x40)) /* zero data bit */
		{
			BlockEndOffset2 = BlockStartOffset + travlen + 11;

			/* ignore the zero data bit if it is a blatant lie */
			gzseek(file, BlockEndOffset2, SEEK_SET);
			if(T2Value(gzgetc(file)) == 0x2a)
			{
				BlockEndOffset = BlockEndOffset2;
			}
		}

		/* move back to start of block */
		gzseek(file, BlockStartOffset, SEEK_SET);
	}
}

/* fast load things */

bool C_TapeBitProviderT2P3::SetFirstBlock(void)
{
	BlockEndOffset = 5;
	OverRan = false;
	return file ? true : false;
}

FileDataBlock C_TapeBitProviderT2P3::blockdata;

FileDataBlock *C_TapeBitProviderT2P3::NextBlock(void)
{
	BlockStartOffset = BlockEndOffset;
	DetermineBlockStats();

	return OverRan ? NULL : ReturnBlock();
}

int C_TapeBitProviderT2P3::GetOffset(void)
{
	return BlockStartOffset;
}

FileDataBlock *C_TapeBitProviderT2P3::SetOffset(int offs)
{
	BlockStartOffset = offs;
	DetermineBlockStats();

	return ReturnBlock();
}

#include <errno.h>

FileDataBlock *C_TapeBitProviderT2P3::ReturnBlock(void)
{
	int c;
	unsigned __int8 *ptr;

	/* this is effectively an EOF check */
	if(BlockEndOffset < BlockStartOffset)
	{
		return NULL;
	}

	/* load data */
	blockdata.length = BlockEndOffset - BlockStartOffset;

	ptr = blockdata.data = (unsigned __int8 *)realloc(blockdata.data, blockdata.length);
	gzread(file, blockdata.data, blockdata.length);

	c = blockdata.length;
	while(c--)
		*ptr++ = T2Value(*ptr);

	/* return data */
	return &blockdata;
}

