/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	tape/tapeuef.cpp
	================

	This file handles streams of bits to and from UEF tape files. When writing
	to UEF, the data is always appended to the end of the tape. This is to be
	'fixed'

	Notice : UEF's are the only thing to offer 'wide' data

*/
#include "tapemodes.h"
#include "../ula/ula.h"
#include "../fs/fs.h"
#include "../uef/uef.h"
#include <stdlib.h>

#define SH 2

#define IMPLICIT_DATA_ID	0x00
#define GENERIC_DATA_ID		0x02

#define HTONE_ID			0x10
#define HTONE_DUMMY_ID		0x11
#define GAP_ID				0x12
#define BAUDRATE_ID			0x13

#define SPARE_ID			0xff

bool C_TapeBitProviderUEF::Setup(C_ULA *ul)
{
	fs = ul->GetFS();
	ula = ul;

	fastfile = file = NULL;
	chunk = NULL;
	chunkmulti = NULL;
	OutChunk = NULL;
	OutChunkMulti = NULL;
	blockdata.data =
	blockdata.wdata = NULL;

	Paused = true;
	Counter = false;
	OpMode = false;

	return true;
}

void C_TapeBitProviderUEF::Close(void)
{
	CloseFile();
}

bool C_TapeBitProviderUEF::SetFile(char *path, char *name, int samptime, int clocktime)
{
	CloseFile();

	if(!(file = fs->RequestUEFChunkSelector(path, NULL, name, "rw")))
	{
		/* try without write capabilities */
		if(!(file = fs->RequestUEFChunkSelector(path, NULL, name, "r")))
			return false;
	}

	/* get fast tape selector also */
	if(!(fastfile = fs->RequestUEFChunkSelector(path, NULL, name, "rw")))
	{
		fastfile = fs->RequestUEFChunkSelector(path, NULL, name, "r");
	}

	ula->SetGlobals(file);

	baudrate = 1200;
	bitlength = (2000000 << SH) / 1200;
	MoveTo(0, samptime, clocktime);
	OpMode = true;

	return true;
}

void C_TapeBitProviderUEF::CloseFile(void)
{
	if(chunk)
	{
		file->ReleaseChunkPtr(chunk);
		chunk = NULL;
	}

	if(chunkmulti)
	{
		file->ReleaseChunkPtr(chunkmulti);
		chunkmulti = NULL;
	}

	if(OutChunk)
	{
		file->ReleaseChunkPtr(OutChunk);
		OutChunk = NULL;
	}

	if(file)
	{
		fs->ReleaseUEFChunkSelector(file);
		file = NULL;
	}

	if(fastfile)
	{
		fs->ReleaseUEFChunkSelector(fastfile);
		fastfile = NULL;
	}

	if(blockdata.data)
	{
		free(blockdata.data);
		blockdata.data = NULL;
	}

	if(blockdata.wdata)
	{
		free(blockdata.wdata);
		blockdata.wdata = NULL;
	}
}

TapeBitBuffer C_TapeBitProviderUEF::back;

TapeBitBuffer *C_TapeBitProviderUEF::PollNewBuffer(int samptime, int clocktime)
{
	if(!file || Paused || !OpMode) return NULL;

	if(clocktime >= (timetarg >> SH))
	{
		switch(idminor)
		{
			case IMPLICIT_DATA_ID :
				switch(bitoffset)
				{
					case 0 : //start bit & read new byte
						back.data8 = 0;
						DataLatch8 = chunk->GetC();
					break;

					default : //all the other bits
						back.data8 = DataLatch8&1;
						DataLatch8 >>= 1;
					break;

					case 9 : //stop bit
						back.data8 = 1;
						bitoffset = -1;
					break;
				}
			break;

			case GENERIC_DATA_ID : //generic
				if(!bitoffset)
					DataLatch8 = chunk->GetC();

				back.data8 = DataLatch8&1;
				DataLatch8 >>= 1;

				if(bitoffset == 7)
					bitoffset = -1;
			break;

			case HTONE_ID : //high tone
				back.data8 = 1;

				if(bitoffset == targetoffset1)
					bitoffset = -1;
			break;

			case HTONE_DUMMY_ID : //high tone with dummy
				if(bitoffset < targetoffset1)
					back.data8 = 1;

				if(bitoffset > (targetoffset1+8) && bitoffset < targetoffset2)
					back.data8 = 1;
				else
					if(bitoffset > targetoffset1)
					{
						switch(bitoffset-targetoffset1)
						{
							case 0 : back.data8 = 0; break;
							case 1 : back.data8 = 0; break;
							case 2 : back.data8 = 0; break;
							case 3 : back.data8 = 1; break;
							case 4 : back.data8 = 1; break;
							case 5 : back.data8 = 1; break;
							case 6 : back.data8 = 0; break;
							case 7 : back.data8 = 1; break;
							case 8 : back.data8 = 1; break;
						}
					}

				if(bitoffset == targetoffset2)
					bitoffset = -1;
			break;

			case GAP_ID : //gap
				if(bitoffset == targetoffset1)
					bitoffset = -1;
			break;
		}

		bitoffset++;

		if(!bitoffset && chunk->EOC())
		{
			FindNextChunk(timetarg, timetarg);
		}
		else
			timetarg += bitlength;//(timetarg+bitlength)&((SOUND_MASK << SH) | 0xff);

		return &back;
	}
	else
		return NULL;
}

void C_TapeBitProviderUEF::CloseChunk(void)
{
	if(!OutChunk) return;

	/* if currently in hightone or gap chunk, write length */
	if(WriteID == GAP_ID || WriteID == HTONE_ID)
	{
		OutLength.a = (OutLength.a*5)/6;
		OutChunk->PutC(OutLength.b.l);
		OutChunk->PutC(OutLength.b.h);
	}

	file->ReleaseChunkPtr(OutChunk);
	OutChunk = NULL;
}

void C_TapeBitProviderUEF::NewChunk(unsigned __int8 idminor)
{
	WriteID = idminor;
	/* new data is added at end of tape */
	int offset = file->GetOffset();
	file->Seek(0, SEEK_END);
	file->EstablishChunk();
	OutChunk = file->GetChunkPtr();
	file->Seek(offset, SEEK_SET);

	OutChunk->SetId(0x100 | WriteID);
	OutLength.a = 0;
}

void C_TapeBitProviderUEF::WriteNewBuffer(int samptime, int clocktime, TapeBitBuffer *b)
{
	if(OpMode || Paused || !file) return;

	/* write new data */
	unsigned __int8 newid;

	newid = IMPLICIT_DATA_ID; // assume 'implicit start/stop chunk'

	if(b->hightone) newid = HTONE_ID;
	if(b->gap) newid = GAP_ID;

	if(newid != WriteID)
	{
		if(WriteID != SPARE_ID) CloseChunk();
		NewChunk(newid);
	}

	switch(WriteID)
	{
		case IMPLICIT_DATA_ID :
			if(OutChunk) OutChunk->PutC(b->data8);
		break;

		case GAP_ID :
		case HTONE_ID :
			OutLength.a += 10;
		break;
	}
}

void C_TapeBitProviderUEF::FindNextChunk(int samptime, int clocktime)
{
	if(chunk)
	{
		file->ReleaseChunkPtr(chunk);
		chunk = NULL;
	}

	if(chunkmulti)
	{
		file->ReleaseChunkPtr(chunkmulti);
		chunkmulti = NULL;
	}

	/*

	move through 0x01 chunks to next recognised as a tape data chunk, and
	alter baud rate if required

	*/
	do
	{
		if(!file->FindIdMajor(0x01))
		{
			idminor = 0xff;
			break;
		}
		idminor = (unsigned __int8)(file->CurrentChunk()->GetId()&0xff);

		/* check for change of baud rate */
		if(idminor == BAUDRATE_ID)
		{
			baudrate = file->CurrentChunk()->GetC();
			baudrate |= file->CurrentChunk()->GetC() << 8;

			bitlength = (2000000 << SH) / baudrate;
		}
	}
	while(! (	idminor == IMPLICIT_DATA_ID ||
				idminor == GENERIC_DATA_ID ||
				idminor == HTONE_ID ||
				idminor == HTONE_DUMMY_ID ||
				idminor == GAP_ID));

	chunk = file->GetChunkPtr();

	/* set up the targetoffset's if applicable */
	switch(idminor)
	{
		case HTONE_ID :
		case GAP_ID :
			targetoffset1 = chunk->GetC();
			targetoffset1 |= chunk->GetC() << 8;

			targetoffset1 = (targetoffset1*6)/5;

			chunk->ReadSeek(0, SEEK_END);
		break;

		case HTONE_DUMMY_ID :
			targetoffset1 = chunk->GetC();
			targetoffset1 |= chunk->GetC() << 8;

			targetoffset2 = chunk->GetC();
			targetoffset2 |= chunk->GetC() << 8;

			targetoffset1 = (targetoffset1*6)/5;
			targetoffset2 = (targetoffset2*6)/5;

			targetoffset2 += targetoffset1;

			chunk->ReadSeek(0, SEEK_END);
		break;
	}

	/* set 'gap' flag */
	if(idminor == GAP_ID || idminor == 0xff)
		back.gap = true;
	else
		back.gap = false;

	/* check for multiplexed version */
	if(idminor != 0xff)
	{
		unsigned __int8 nextid;
		int foffset = file->GetOffset();

		file->Seek(1, SEEK_CUR);
		nextid = (unsigned __int8)(file->CurrentChunk()->GetId()&0xff);
		if(!(nextid&0xf0) && (nextid == idminor+1))
		{
			chunkmulti = file->GetChunkPtr();
		}

		file->Seek(foffset, SEEK_SET);
	}

	/* set up the necessary variables */

	timetarg = clocktime + bitlength;
	bitoffset = 0;
}

void C_TapeBitProviderUEF::MoveTo(int time, int samptime, int clocktime)
{
	/* treated as necessarily rewind for now */
	file->Reset();
	FindNextChunk(samptime << SH, clocktime << SH);
}

void C_TapeBitProviderUEF::MinimiseTimer(int sampsub, int clocksub)
{
	if(!Paused)
		timetarg -= (clocksub << SH);
}

void C_TapeBitProviderUEF::SetMode(bool cntr, bool read, bool motor, int samptime, int clocktime)
{
	bool CurrentEnabled = Enabled;

	if(Inverse)
	{
		if(read = !read)
		{
			/* if more is write, disable */
			CurrentEnabled = false;
		}
	}

	if(!CurrentEnabled)
	{
		/* effect a disabling by switching off the counter */
		cntr = false;
	}

	/* store counter mode */
	Counter = cntr;

	/* store & process motor mode */
	if(motor && Paused)
	{
		timetarg += (clocktime << SH);
	}

	if(!motor && !Paused)
	{
		timetarg -= (clocktime << SH);
	}

	Paused = !motor;

	/* store and process read/write mode */
	if(OpMode != read)
	{
		OpMode = read;

		if(file)
		{
			if(OpMode)
			{
				/* close chunk if one is open */
				if(WriteID != SPARE_ID)
				{
					CloseChunk();
					WriteID = SPARE_ID;
				}
			}
		}
	}
}

/* fast load things */

bool C_TapeBitProviderUEF::SetFirstBlock(void)
{
	if(!fastfile) return false;

	fastfile->Seek(0, SEEK_SET);
	fastfile->ResetOverRan();
	return true;
}

FileDataBlock C_TapeBitProviderUEF::blockdata;

FileDataBlock *C_TapeBitProviderUEF::NextBlock(void)
{
	fastfile->FindId(0x0100);
	if(fastfile->OverRan()) return NULL;

	return ReturnBlock();
}

int C_TapeBitProviderUEF::GetOffset(void)
{
	return fastfile->GetOffset();
}

FileDataBlock *C_TapeBitProviderUEF::SetOffset(int offs)
{
	fastfile->Seek(offs, SEEK_SET);
	return ReturnBlock();
}

FileDataBlock *C_TapeBitProviderUEF::ReturnBlock(void)
{
	int offset = fastfile->GetOffset();

	if(fastfile->CurrentChunk()->GetId() == 0x0100)
	{
		/* load data */
		blockdata.length = fastfile->CurrentChunk()->GetLength();
		blockdata.data = (unsigned __int8 *)realloc(blockdata.data, blockdata.length);
		fastfile->CurrentChunk()->Read(blockdata.data, blockdata.length);

		/* check for multiplexed data */
		fastfile->Seek(1, SEEK_CUR);
		if(	fastfile->CurrentChunk()->GetId() == 0x0101 &&
			fastfile->CurrentChunk()->GetLength() == (blockdata.length << 2))
		{
			blockdata.wdata = (unsigned __int8 *)realloc(blockdata.wdata, blockdata.length << 2);
			fastfile->CurrentChunk()->Read(blockdata.wdata, blockdata.length << 2);
		}
		else
		{
			if(blockdata.wdata)
			{
				free(blockdata.wdata);
				blockdata.wdata = NULL;
			}
		}

		fastfile->Seek(offset, SEEK_SET);

		/* return data */
		return &blockdata;
	}
	else
		return NULL;
}

