/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	tape/tapewave.cpp
	=================

	This file handles streams of bits to and from PCM encoded sound sources,
	at least as much as the PCM target allows

	It has a hard coded bit length based on the assumption of 1200 baud, but
	does employ something of a 'variable zero' based on looking at the height
	of extrema. Further, because it is happy with not too exact values, baud
	rates in the range [980, 1520] should be fine - easily enough to handle
	warped tapes

	NB : both ReadCursor and WriteCursor have 8 fractional bits, hence the
	macros

*/
#include "tapemodes.h"
#include "../sound/speaker.h"
#include "../ula/ula.h"
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define BIT_LENGTH 37
#define QUARTER_BIT_LENGTH 8

#define BUFFER_LENGTH (SOUND_MASK+1)

#define INCCOUNTER(x) x = (x+0x100)&(((BUFFER_LENGTH-1) << 8) | 0xff)
#define DECCOUNTER(x) x = (x-0x100)&(((BUFFER_LENGTH-1) << 8) | 0xff)
#define LDCOUNTER(x, y) x = y << 8
#define REFCOUNTER(x) (x >> 8)
#define ADDCOUNTER(x, y) x = (x+((y) << 8))&(((BUFFER_LENGTH-1) << 8) | 0xff)
#define ADCCOUNTER(x, y) x = (x+y)&(((BUFFER_LENGTH-1) << 8) | 0xff)

#define ZERO_HZ		1200
#define ONE_HZ		2400

#ifndef PI
#define PI 3.141592654f
#endif

#define WRBUF_LEN 2400
struct ByteWrite
{
	__int8 bit;
	int samptime;
} Table[WRBUF_LEN], *TableWritePos, *TableReadPos;

bool C_TapeBitProviderWave::Setup(C_ULA *ul)
{
	snd = ul->GetSound();
	speaker = ul->GetSpeaker();
	source = NULL;
	BufMem = Peak = Trough = NULL;
	TableWritePos = TableReadPos = Table;
	id = 0;
	sinepos = 0;

	int c = 64;
	while(c--)
		sinewave[c] = (signed __int8)(-32.0f*sin( ((float)c *PI) / 32.0f));

	return true;
}

void C_TapeBitProviderWave::Close(void)
{
	CloseFile();

	if(BufMem)
	{
		free(BufMem);
		BufMem = NULL;
	}

	if(Peak)
	{
		free(Peak);
		Peak = NULL;
	}

	if(Trough)
	{
		free(Trough);
		Trough = NULL;
	}

	InputLoopCount = 0;
}

bool C_TapeBitProviderWave::SetFile(char *path, char *name, int samptime, int clocktime)
{
	CloseFile();
	source = snd->GetSoundSource(path, name, NULL);

	if(!source) return false;

	BufMem = (unsigned __int8 *)calloc(BUFFER_LENGTH, sizeof(unsigned __int8));
	Peak = (unsigned __int8 *)calloc(BUFFER_LENGTH, sizeof(unsigned __int8));
	Trough = (unsigned __int8 *)calloc(BUFFER_LENGTH, sizeof(unsigned __int8));
	InputBlocked = true;

	InputLoopCount = 0;
	InputTimeBase = samptime;
	SetMode(Counter, OpMode, !Paused, samptime, clocktime);

	return true;
}

void C_TapeBitProviderWave::CloseFile(void)
{
	if(source)
	{
		snd->ReleaseSoundSource(source);
		source = NULL;
	}
}

#define Sign(x) (x ? ((x < 0) ? -1 : +1) : 0)

int C_TapeBitProviderWave::FindNextExtrema(int start, int direction)
{
	unsigned __int8 value1, value2;

	value1 = BufMem[REFCOUNTER(start)];

	do
	{
		INCCOUNTER(start);
		value2 = value1;
		value1 = BufMem[REFCOUNTER(start)];
	}
	while((!Sign(value1-value2) || Sign(value1 - value2) != Sign(direction)) && (REFCOUNTER(start) != REFCOUNTER(WriteCursor)));

	if(REFCOUNTER(start) != REFCOUNTER(WriteCursor))
		return DECCOUNTER(start);
	else
		return -1;
}

#undef Sign

int C_TapeBitProviderWave::FindFarExtrema(int start, int direction)
{
	int dist1, dist2, ostart;

	dist1 = CircularDifference(REFCOUNTER(WriteCursor), REFCOUNTER(start), BUFFER_LENGTH);

	do
	{
		ostart = start;
		if((start = FindNextExtrema(start, direction)) == -1) break;
		dist2 = dist1;
		dist1 = CircularDifference(REFCOUNTER(WriteCursor), REFCOUNTER(start), BUFFER_LENGTH);
	}
	while(dist2 > dist1);

	return ostart;
}

#define Zero(x) ((Peak[REFCOUNTER(x)] + Trough[REFCOUNTER(x)]) >> 1)

TapeBitBuffer *C_TapeBitProviderWave::PollNewBuffer(int samptime, int clocktime)
{
	SoundBuffer *buf;

	if(!id || Paused || !OpMode)
		return NULL;

	//check if any more sound data need be grabbed

	if(buf = source->PollNewBuffer(id, 0))
	{
		// copy to WriteCursor
		if((REFCOUNTER(WriteCursor) + buf->length) > BUFFER_LENGTH)
		{
			int firstlen = BUFFER_LENGTH-REFCOUNTER(WriteCursor);

			memcpy(&BufMem[REFCOUNTER(WriteCursor)], buf->buffer, firstlen);
			memcpy(BufMem, &buf->buffer[firstlen], buf->length-firstlen);
		}
		else
		{
			memcpy(&BufMem[REFCOUNTER(WriteCursor)], buf->buffer, buf->length);
		}

		// update position of WriteCursor

		int dist1 = CircularDifference(REFCOUNTER(WriteCursor), REFCOUNTER(ReadCursor), BUFFER_LENGTH);
		int OldPosition = WriteCursor;

		ADDCOUNTER(WriteCursor, buf->length);
		source->ReturnBuffer(id);

		//check if new data has started overwriting old data flush all buffers

		int dist2 = CircularDifference(REFCOUNTER(WriteCursor), REFCOUNTER(ReadCursor), BUFFER_LENGTH);

		if(dist2 < dist1 && !InputBlocked)
		{
			ReadCursor = OldPosition;
			InputLoopCount = 0;
			InputTimeBase = samptime;
		}

		// calculate new peak & trough levels using piecewise monotonic attribute

		int Pos1 = TroughCursor, Pos2;
		int height, adjustor;

		// determine trough modifier
		while(1)
		{
			Pos2 = Pos1;
			Pos1 = FindNextExtrema(Pos1, -1);
			if(Pos1 == -1) break;
			Pos1 = FindNextExtrema(Pos1, 1);
			if(Pos1 == -1) break;

			if(Pos2 != Pos1)
			{
				height = BufMem[REFCOUNTER(Pos2)] << 16;
				adjustor = (BufMem[REFCOUNTER(Pos1)] - BufMem[REFCOUNTER(Pos2)]) << 16;
				adjustor /= CircularDifference(REFCOUNTER(Pos1), REFCOUNTER(Pos2), BUFFER_LENGTH);
			}

			while(REFCOUNTER(TroughCursor) != REFCOUNTER(Pos1))
			{
				Trough[REFCOUNTER(TroughCursor)] = height >> 16;
//				height += adjustor;
				INCCOUNTER(TroughCursor);
			}
		}

		// determine peak modifier
		Pos1 = PeakCursor;
		while(1)
		{
			Pos2 = Pos1;
			Pos1 = FindNextExtrema(Pos1, 1);
			if(Pos1 == -1) break;
			Pos1 = FindNextExtrema(Pos1, -1);
			if(Pos1 == -1) break;

			height = BufMem[REFCOUNTER(Pos2)] << 16;
			adjustor = (BufMem[REFCOUNTER(Pos1)] - BufMem[REFCOUNTER(Pos2)]) << 16;
			adjustor /= CircularDifference(REFCOUNTER(Pos1), REFCOUNTER(Pos2), BUFFER_LENGTH);

			while(REFCOUNTER(PeakCursor) != REFCOUNTER(Pos1))
			{
				Peak[REFCOUNTER(PeakCursor)] = height >> 16;
//				height += adjustor;
				INCCOUNTER(PeakCursor);
			}
		}

		if(
			CircularDifference(REFCOUNTER(WriteCursor), REFCOUNTER(TroughCursor), BUFFER_LENGTH) >
			CircularDifference(REFCOUNTER(WriteCursor), REFCOUNTER(PeakCursor), BUFFER_LENGTH)
		)
			ZeroCursor = TroughCursor;
		else
			ZeroCursor = PeakCursor;

		// if new data was being awaited, unblock and fix timebase
		InputBlocked = false;

		if(!InputLoopCount)
			InputTimeBase = samptime;
	}

	// check if any more InputLoops can be processed
	int dist;
	int sign1, sign2, ReadCursorStart;

	if(InputLoopCount < 16 && !InputBlocked)
	{
		bool success;

		sign2 = BufMem[REFCOUNTER(ReadCursor)] <= Zero(ReadCursor);

		do
		{
			ReadCursorStart = ReadCursor;
			success = false;
			InputLoops[InputLoopCount] = 0;

			// search for next zero crossing
			do
			{
				sign1 = sign2;
				sign2 = BufMem[REFCOUNTER(ReadCursor)] <= Zero(ReadCursor);

				if(sign1 != sign2) {success = true; break;}

				InputLoops[InputLoopCount]++;
				INCCOUNTER(ReadCursor);

				dist = CircularDifference(REFCOUNTER(ZeroCursor), REFCOUNTER(ReadCursor), BUFFER_LENGTH);
			}
			while(dist);

			if(success)
			{
				InputLoopCount++;
			}
			else
			{
				InputBlocked = true;
				ReadCursor = ReadCursorStart;
			}
		}
		while((InputLoopCount < 16) && !InputBlocked);
	}

	// return a bit if possible
	if(!InputLoopCount)
		return NULL;

	int len, InputLoopstaken = 0, timerequired = 0;

	if(!timerequired && (InputLoopCount > 3))
	{
		len = InputLoops[0] + InputLoops[1] + InputLoops[2] + InputLoops[3];

		if( abs(len - BIT_LENGTH) < QUARTER_BIT_LENGTH)
		{
			InputValue.data8 = 1;
			InputValue.gap = false;
			InputLoopstaken = 4;
			timerequired = len;
		}
	}

	if(!timerequired && (InputLoopCount > 1))
	{
		len = InputLoops[0] + InputLoops[1];

		if( abs(len - BIT_LENGTH) < QUARTER_BIT_LENGTH)
		{
			InputValue.data8 = 0;
			InputValue.gap = false;
			InputLoopstaken = 2;
			timerequired = len;
		}
	}

	//if enough InputLoops exist that the program would otherwise be able to
	//tell, but all the same they don't look like a 1 or a 0, assume to be a
	//gap.

	//This actually may just be the last loop of 'high tone', but that doesn't
	//matter since the tape interface can survive this confusion
	if(!timerequired && (InputLoopCount > 3))
	{
		InputValue.gap = true;
		InputValue.data8 = 1;
		InputLoopstaken = 1;
		timerequired = InputLoops[0];

		if(timerequired > 16384)
		{
			timerequired = 16384;
			InputLoops[0] -= 16384;
			InputLoopstaken = 0;
		}
	}

	if(!timerequired) return NULL;

	// has the event discovered actually occurred yet?
	if(CircularDifference(samptime, InputTimeBase, SOUND_MASK+1) < timerequired)
		return NULL;

	InputTimeBase = (InputTimeBase+timerequired)&SOUND_MASK;

	if(InputLoopstaken)
	{
		InputLoopCount -= InputLoopstaken;
		for(int c = InputLoopstaken; c < 16; c++)
			InputLoops[c-InputLoopstaken] = InputLoops[c];
	}

	return &InputValue;
}

#undef Zero

void C_TapeBitProviderWave::WriteBit(int samptime, __int8 bit)
{
	TableWritePos->samptime = samptime;
	TableWritePos->bit = bit;
	TableWritePos++;
	if(TableWritePos == &Table[WRBUF_LEN]) TableWritePos = Table;
}

void C_TapeBitProviderWave::WriteNewBuffer(int samptime, int clocktime, TapeBitBuffer *b)
{
	if(OpMode || Paused || !Counter) return;

	samptime <<= 2;

	if(b->hightone)
	{
		int c= 10;
		while(c--)
		{
			WriteBit(samptime, 1);
			samptime = (samptime+147)&((SOUND_MASK << 2) | 3);
		}
	}
	else
	{
		WriteBit(samptime, 0);
		samptime = (samptime+147)&((SOUND_MASK << 2) | 3);

		int c= 8;
		unsigned __int8 val = b->data8;
		while(c--)
		{
			WriteBit(samptime, val&1);
			val >>= 1;
			samptime = (samptime+147)&((SOUND_MASK << 2) | 3);
		}

		WriteBit(samptime, 1);
		samptime = (samptime+147)&((SOUND_MASK << 2) | 3);
	}
}

void C_TapeBitProviderWave::MoveTo(int time, int samptime, int clocktime)
{
	if(source)
		source->Control(SOUNDCONTROL_SETTIME, time);
}

void C_TapeBitProviderWave::Keepalive(int samptime, int clocktime)
{
	/* need to monitor output situation here */

	if(source && id && !OpMode)
	{
		SoundBuffer *buf;
		unsigned __int8 *bptr;
		int count;

		if(buf = source->PollNewBuffer(id, CircularDifference((TableWritePos->samptime >> 2), WriteCursor, SOUND_MASK+1)))
		{
			bptr = buf->buffer;
			count = buf->length;

			while(count--)
			{
				if(TableReadPos != TableWritePos)
				{
					if((TableReadPos->samptime >> 2) == WriteCursor)
					{
						sineadd = TableReadPos->bit ? 228261 : 114131;
						if( ((sinepos >> 16)&63) > 1 && ((sinepos >> 16)&63) < 62)
							sinepos = 0;

						TableReadPos++;
						if(TableReadPos == &Table[WRBUF_LEN]) TableReadPos = Table;
					}
				}

				*bptr++ += sinewave[(sinepos >> 16)&63];
				sinepos += sineadd;
				WriteCursor = (WriteCursor+1)&SOUND_MASK;
			}

			source->ReturnBuffer(id);
		}
	}
}

void C_TapeBitProviderWave::MinimiseTimer(int sampsub, int clocksub)
{
	if(source)
		source->MinimiseTimer(sampsub);

	if(OpMode)
		InputTimeBase = (InputTimeBase-sampsub)&SOUND_MASK;
	else
	{
		WriteCursor = (WriteCursor-sampsub)&SOUND_MASK;

		ByteWrite *b = TableReadPos;
		while(b != TableWritePos)
		{
			b->samptime = (b->samptime - (sampsub << 2))&((SOUND_MASK << 2) | 3);

			b++;
			if(b == &Table[WRBUF_LEN]) b = Table;
		}
	}
}

void C_TapeBitProviderWave::SetMode(bool cntr, bool read, bool motor, int samptime, int clocktime)
{
	bool TEnabled;
	int modeswitch;

	/*

		read mode is entered if :

			the device is enabled		&&
			the device is not inverted	&&
			the counter is enabled		&&
			the motor is enabled

		write mode is entered if :

			(the device is enabled		||
			(the device is inverted		&&
			read mode is requested))	&&
			the counter is enabled		&&
			the motor is enabled

	*/

	TEnabled = Paused;

	if(!source || !id)
	{
		modeswitch = 0;
	}
	else
	{
		modeswitch = source->IsInputId(id) ? 1 : 2;
	}
	modeswitch *= 3;

	if(cntr && motor)
	{
		if(Enabled && read)
		{
			modeswitch += 1;
		}
		else
			if((Enabled && !read) || (Inverse && read))
			{
				modeswitch += 2;
			}
	}
	/* 	determine mode switch necessary, format is a two digit base 3 number,
		high digit is old mode, low digit is new mode, digits mean :

		0 - no activity
		1 - read mode
		2 - write mode

	*/

	if(source)
	switch(modeswitch)
	{
		case (0 + (0*3)) :
		case (1 + (1*3)) :
		case (2 + (2*3)) :
			/* no change */
		break;

		case (0 + (1*3)) :
		case (0 + (2*3)) :
			/* disable activity */
			if(id)
			{
				source->ReturnAccess(id);
			}
			id = 0;
		break;

		case (1 + (0*3)) :
		case (1 + (2*3)) :
			/* switch to read */

			if(id)
			{
				source->ReturnAccess(id);
			}

			id = source->RequestAccess(SoundMode_Input);
			InputTimeBase = samptime;
			TroughCursor = PeakCursor = ZeroCursor = ReadCursor;
		break;

		case (2 + (0*3)) :
		case (2 + (1*3)) :
				/* switch to write */

				if(id)
				{
					source->ReturnAccess(id);
				}

				id = source->RequestAccess(SoundMode_Output);
				WriteCursor = samptime;
			break;
		}

	if(source)
	{
		if(motor)
			source->Unpause(id, samptime);
		else
			source->Pause(id, samptime);
	}

	/* store new mode */
	Counter = cntr;
	Paused = !motor;
	OpMode = read;
}

