/*

	ElectrEm (c) 2000 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	uef/chunk.cpp
	=============

	Chunk level parts of (second version) UEF class

*/
#include "uef.h"
#include <stdarg.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>

C_UEFChunk::C_UEFChunk(gzFile input, char *name, bool wrte)
{
	memory = NULL;
	next = NULL;
	Create(input, name, wrte);
	refcount = 0;
}

C_UEFChunk::~C_UEFChunk(void)
{
}

int C_UEFChunk::GetRefCount(void)
{
	return refcount;
}

int C_UEFChunk::Read(void *buf, int numbytes)
{
	if(memory)
	{
		if((numbytes+readptr) >= length)
			numbytes = length-readptr;

		memcpy(buf, &memory[readptr], numbytes);

		readptr += numbytes;

		return numbytes;
	}
	else
		return 0;
}

int C_UEFChunk::Seek(unsigned __int32 *variable, long offset, int mode)
{
	if(memory)
	{
		switch(mode)
		{
			case SEEK_SET :		(*variable) = offset;			break;
			case SEEK_CUR :		(*variable) += offset;			break;
			case SEEK_END :		(*variable) = length - offset;	break;
		}

		if((*variable) > length) (*variable) = length;
		if((*variable) < 0) (*variable) = 0;

		return readptr;
	}
	else
		return -1;
}

int C_UEFChunk::ReadSeek(long offset, int mode)
{
	return Seek(&readptr, offset, mode);
}

int C_UEFChunk::GetC(void)
{
	if(readptr == length || !memory) return UEF_EOF;
	return memory[readptr++];
}

int C_UEFChunk::Write(const void *buf, int numbytes)
{
	if(!write) return -1;

	if(memory)
	{
		dirty = true;

		if(writeptr+numbytes > memorylen)
		{
			/*

			when realloc'ing memory, the newsize is rounded up to the nearest
			256 bytes. This size is fairly optimal for minimising the required
			number of reallocs with respect to the average sort of data sent
			to a UEF file

			*/
			memorylen = (writeptr+numbytes);
			memorylen += 256 - (memorylen&255);
			memory = (unsigned __int8 *)realloc(memory, memorylen);
		}

		memcpy(&memory[writeptr], buf, numbytes);
		writeptr += numbytes;

		if(writeptr > length) length = writeptr;

		return numbytes;
	}

	return -1;
}

int C_UEFChunk::WriteSeek(long offset, int mode)
{
	return write ? Seek(&writeptr, offset, mode) : -1;
}

void C_UEFChunk::PutC(int value)
{
	/*

	PutC would be a little more complex than GetC, as it has to be checked
	that enough memory is allocated, and if not stuff has to be realloc'd and
	adjusted, so its easier just to call Write and not worry about it here

	*/
	unsigned __int8 val = value;
	Write(&val, 1);
}

unsigned __int16 C_UEFChunk::GetId(void)
{
	return id;
}

unsigned __int32 C_UEFChunk::GetLength(void)
{
	return length;
}

bool C_UEFChunk::EOC(void)
{
	return (readptr == length);
}

void C_UEFChunk::SetId(__int16 newid)
{
	if(write)
	{
		id = newid;
		dirty = true;
	}
}

void C_UEFChunk::SetLength(__int32 newlength)
{
	if(write)
	{
		length = newlength;
		dirty = true;
	}
}

bool C_UEFChunk::HasChanged(void)
{
	return dirty;
}

bool C_UEFChunk::Write(gzFile f, C_UEFFile *owner)
{
	/* output the chunk to the specified file */
	bool active = true;

	if(!memory)
	{
		active = false;
		Enable(owner);
	}

	gzputc(f, id&255);
	gzputc(f, id>>8);

	gzputc(f, (length & 0x000000ff) >> 0);
	gzputc(f, (length & 0x0000ff00) >> 8);
	gzputc(f, (length & 0x00ff0000) >> 16);
	gzputc(f, (length & 0xff000000) >> 24);

	gzwrite(f, memory, length);

	if(!active) Disable();

	return true;
}

bool C_UEFChunk::Create(gzFile input, char *name, bool wrte)
{
	write = wrte;

	if(name)
	{
		/* store and setup necessary parts */
		fname = strdup(name);
		dirty = false;

		/* retrieve vital information from gzFile - offset, chunk id and length */
		id = gzgetc(input);
		id |= ((int)gzgetc(input)) << 8;

		length = gzgetc(input);
		length |= ((int)gzgetc(input)) << 8;
		length |= ((int)gzgetc(input)) << 16;
		length |= ((int)gzgetc(input)) << 24;

		if(gzeof(input)) {id = 0xffff; length = 0; return false;}

		/* advance file ptr to next chunk & return */
		offset = gzseek(input, length, SEEK_CUR) - length;

		return true;
	}
	else
	{
		id = 0; length = 0;
		dirty = true;
		fname = NULL;

		/* allocate 256 bytes initial maximum size */

		memory = (unsigned __int8 *)malloc(256);
		memorylen = 256;

		return true;
	}

	return false;
}

void C_UEFChunk::Kill(void)
{
	if(fname)
	{
		free(fname);
		fname = NULL;
	}
}

int C_UEFChunk::Enable(C_UEFFile *owner)
{
	readptr = writeptr = 0;

	if(!memory)
	{
		if(! (memory = (unsigned __int8 *)malloc(length)) ) return false;
		memorylen = length;

		gzFile tmp = owner->GetFile(fname);

		if(tmp)
		{
			gzseek(tmp, offset, SEEK_SET);
			gzread(tmp, memory, length);
		}
		else
		{
			free(memory);
			memory = NULL;
		}
	}
	refcount++;

	return refcount;
}

int C_UEFChunk::Disable(void)
{
	/*

	if no changes were made, safe to just free memory and forget about it

	*/

	refcount--;

	if(!refcount)
	{
		if(!dirty)
		{
			if(memory)
			{
				free(memory);
				memory = NULL;
			}
		}
	}

	return refcount;
}

