/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This file at some time additionally worked upon by Paul Boddie

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	ula/settings.cpp
	================

	This code handles seeding many of the ULA connected devices with settings
	from the config file.

*/
#include "../screen/screen.h"
#include "../sound/speaker.h"
#include "../misc/input.h"
#include "ula.h"
#include "../misc/config.h"
#include "../misc/timing.h"
#include "../wd1770/drive.h"
#include "../plus1/plus1.h"

#include <malloc.h>
#include <string.h>

int setlevel;

#define SeedSetLevel() setlevel = 0;
#define IncSetLevel() setlevel++;
#define DecSetLevel() if(!setlevel--) return;

bool C_ULA::SeedExternal(void)
{
	/* setup screen stuff */
	bool big = false, scaletype = true;
	int scalelevel = 0;

	SeedSetLevel();

	if(settings->FindSetting("vidmode", INTEGER_RESULT))
		big = settings->ReadSetting().result_int ? true : false;

	if(settings->FindSetting("cleanscale", BOOLEAN_RESULT))
		scaletype = settings->ReadSetting().result_bool;

	if(settings->FindSetting("scalelevel", INTEGER_RESULT))
		scalelevel = settings->ReadSetting().result_int;

	if(!scr->Setup(this, big, scaletype, scalelevel, false))
	{
		errstring = scr->GetError();
		return false;
	}

	IncSetLevel();

	/* setup sound subsystem */
	sound->Setup(fs);

	float volume = 1.0f;
	if(settings->FindSetting("volume", INTEGER_RESULT))
		volume = (float)(settings->ReadSetting().result_int & 255) / 255.0f;

	bool highsound = true;
	if(settings->FindSetting("highsound", BOOLEAN_RESULT))
		highsound = settings->ReadSetting().result_bool;
	speaker->Setup(sound, highsound, volume);
	speaker->SetMode(SM_OFF, 0);
	speaker->Pause(0);

	/* setup input stuff */
	if(!input->Setup())
	{
		errstring = input->GetError();
		return false;
	}
	input->UpdateKeyboard();

	IncSetLevel();

	/* & the Plus 1 */
	if(!plus1->Setup(this))
	{
		errstring = "Error setting up Plus 1";
		return false;
	}
	plus1->PRNSetMode(PRNMODE_OFF);

	IncSetLevel();

	/* load keyboard map */
	char *keymap;
	bool maploaded = false;

	if(settings->FindSetting("keymap", STRING_RESULT))
	{
		if(input->LoadMap(keymap = strdup(settings->ReadSetting().result_string)))
			maploaded = true;

		free(keymap);
	}

	if(!maploaded)
	{
		maploaded = input->LoadMap("physical.kyb");
	}

	if(!maploaded)
	{
		/* create a new map */
		input->CreateMap("newmap.kyb");
		forcegui = true;
	}

	/* and joystick deadzones */
	int jc = input->GetNumJoysticks();
	char jtext[13];

	while(jc--)
	{
		sprintf(jtext, "joy%ddeadzone", jc);

		if(settings->FindSetting(jtext, INTEGER_RESULT))
		{
			input->SetDeadZone(jc, (float)settings->ReadSetting().result_int / 65536.0f);
		}
		else
			input->SetDeadZone(jc, 0.04f);
	}

	if(settings->FindSetting("fbjoy", INTEGER_RESULT))
	{
		SetJoystickId(settings->ReadSetting().result_int);
	}
	else
		SetJoystickId(-1);

	IncSetLevel();

	/* init timer */
	slowdown->SetBeatsPerSecond(50);
	if(settings->FindSetting("speed", INTEGER_RESULT))
	{
		slowdown->SetMultiplier((float)settings->ReadSetting().result_int / 65536.0f);
	}

	/* disc drives */
	if(settings->FindSetting("discprotect1", BOOLEAN_RESULT))
		drives[0]->SetWriteProtect(settings->ReadSetting().result_bool);
	if(settings->FindSetting("discprotect2", BOOLEAN_RESULT))
		drives[1]->SetWriteProtect(settings->ReadSetting().result_bool);

	return true;
}

void C_ULA::EndExternal(void)
{
	config_setting s;

	/* write screen settings */
	s.result_type = INTEGER_RESULT;

	s.result_int = scr->GetSize() ? 1 : 0;
	settings->FindSetting("vidmode", ANY_RESULT);
	settings->WriteSetting(s);

	if(scr->GetBitmap())
	{
		s.result_int = scr->GetBitmap()->GetStretchLevel();
		settings->FindSetting("scalelevel", ANY_RESULT);
		settings->WriteSetting(s);

		s.result_type = BOOLEAN_RESULT;
		s.result_bool = scr->GetBitmap()->GetCleanScale();
		settings->FindSetting("cleanscale", ANY_RESULT);
		settings->WriteSetting(s);
	}

	DecSetLevel();

	/* write sound settings */
	s.result_type = INTEGER_RESULT;
	s.result_int = (int)(255.0f*speaker->GetVolume());
	settings->FindSetting("volume", ANY_RESULT);
	settings->WriteSetting(s);

	s.result_type = BOOLEAN_RESULT;
	s.result_bool = speaker->GetAccuracy();
	settings->FindSetting("highsound", ANY_RESULT);
	settings->WriteSetting(s);
	speaker->Close();
	sound->Close();

	DecSetLevel();

	/* write input settings */
	s.result_type = STRING_RESULT;
	s.result_string = input->GetMapName();

	if(s.result_string)
	{
		settings->FindSetting("keymap", ANY_RESULT);
		settings->WriteSetting(s);
	}
	input->CloseMap();

	s.result_type = INTEGER_RESULT;
	s.result_int = GetJoystickId();
	settings->FindSetting("fbjoy", ANY_RESULT);
	settings->WriteSetting(s);

	s.result_type = INTEGER_RESULT;
	int jc = input->GetNumJoysticks();
	char jtext[13];
	while(jc--)
	{
		s.result_int = (int)(input->GetDeadZone(jc)*65536.0f);
		sprintf(jtext, "joy%ddeadzone", jc);
		settings->FindSetting(jtext, ANY_RESULT);
		settings->WriteSetting(s);
	}

	DecSetLevel();

	/* plus 1 */

	plus1->Close();
	DecSetLevel();

	/* write timer settings */
	s.result_type = INTEGER_RESULT;
	s.result_int = (int)(slowdown->GetMultiplier()*65536.0f);
	settings->FindSetting("speed", ANY_RESULT);
	settings->WriteSetting(s);

	/* write disc drive settings */
	s.result_type = BOOLEAN_RESULT;
	s.result_bool = drives[0]->GetWriteProtect() ? true : false;
	settings->FindSetting("discprotect1", ANY_RESULT);
	settings->WriteSetting(s);
	s.result_bool = drives[1]->GetWriteProtect() ? true : false;
	settings->FindSetting("discprotect2", ANY_RESULT);
	settings->WriteSetting(s);
}
