/*

	ElectrEm (c) 2000 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	snapshot/snapshot.cpp
	=====================

	Reads and writes state snapshots (using the internal extended
	addressing map for reading and writing registers), but without
	checking the snapshot chunk ids yet, so won't work properly with
	many strictly valid files
*/

#include "../gfx/gfx.h"
#include "ula.h"
#include "../fs/fs.h"
#include "../uef/uef.h"
#include "../sound/speaker.h"
#include "../misc/timing.h"

#include "../gui/gui.h"
#include "../gui/romclasses.h"

bool C_ULA::Multiplexable(C_UEFChunkSelector *cnk)
{
	return cnk->FindId(0x0006);
}

bool C_ULA::SnapshotAvailable(C_UEFChunkSelector *in, unsigned __int8 id)
{
	bool found;

	/* check if a suitable 6502 state chunk exists */
	found = false;
	in->Reset();
	in->ResetOverRan();
	while(!in->OverRan() && !found)
	{
		if(in->FindId(0x0400))
		{
			if(in->CurrentChunk()->GetC() == id) found = true;
		}
	}

	if(!found) return false;

	/* check if a suitable ULA state chunk exists */
	found = false;
	in->Reset();
	in->ResetOverRan();
	while(!in->OverRan() && !found)
	{
		if(in->FindId(0x0401))
		{
			if(in->CurrentChunk()->GetC() == id) found = true;
		}
	}

	if(!found) return false;

	/* check if a suitable chunk of memory exists */
	found = false;
	in->Reset();
	in->ResetOverRan();
	while(!in->OverRan() && !found)
	{
		if(in->FindId(0x0410))
		{
			found = true;
			if(in->CurrentChunk()->GetC() != id) found = false;
			if(in->CurrentChunk()->GetC()) found = false;
		}
	}

	if(!found) return false;

	return true;
}

void C_ULA::SetGlobals(C_UEFChunkSelector *in)
{
	MultiState nstate;
	unsigned char temp;

	/* determine multiplexing information */
	if(in->FindId(0x0006))
	{
		nstate.multicpu = true;
		temp = in->CurrentChunk()->GetC();

		if(temp == 1)
		{
			nstate.gstyle = in->CurrentChunk()->GetC();
			nstate.multiscr = true;

			ProposeMultiState(nstate);
		}
		else
		{
			nstate.multicpu = false;
			nstate.multiscr = false;
			ProposeMultiState(nstate);
		}
	}

	/* determine multiplexing palette if relevant */
	if(in->FindId(0x0007))
	{
		unsigned char r, g, b;
		int c;

		in->CurrentChunk()->ReadSeek(48, SEEK_SET);
		for(c = 16; c < 256; c++)
		{
			r = in->CurrentChunk()->GetC();
			g = in->CurrentChunk()->GetC();
			b = in->CurrentChunk()->GetC();

			gfx->SetPaletteEntry(c, (float)r / 255.0f, (float)g / 255.0f, (float)b / 255.0f, PAL_ELK);

			if(in->CurrentChunk()->EOC()) break;
		}
		gfx->FinishPalette(PAL_ELK);
	}

	/* parse ROM hints */
	unsigned char presence, type;
	char *namep, name[256];

	in->ResetOverRan();
	while(1)
	{
		if(!in->FindId(0x0008)) break;
		if(in->OverRan()) break;

		presence = in->CurrentChunk()->GetC();
		temp = in->CurrentChunk()->GetC();

		if(!temp)
		{
			/* get a ROM name - limit to 256 chars for now */
			namep = name-1;
			do
			{
				namep++;
				*namep = in->CurrentChunk()->GetC();
			}
			while(*namep && namep < &name[256]);
			if(namep == &name[256]) name[255] = '\0';

			type = ROMCLASS_SPECIFIC;
		}
		else
		{
			/* get a ROM type */
			type = in->CurrentChunk()->GetC();

			switch(type)
			{
				case 0 : type = ROMCLASS_ALL; break;
				case 1 : type = ROMCLASS_PAGED; break;
				case 2 : type = ROMCLASS_DFS; break;
				case 3 : type = ROMCLASS_ADFS; break;
				case 4 : type = ROMCLASS_FS; break;
				case 5 : type = ROMCLASS_LANG; break;
			}
		}

		gui->PassROMHint((presence&1) ? true : false, type, name);
	}
}

bool C_ULA::LoadSnapshot(char *path, char *name, unsigned __int8 id)
{
	C_UEFChunkSelector *in;
	unsigned char temp;
	BrokenWord addr;
	BrokenQuad quad;
	int c;
	bool loaded = false;

	if(in = fs->RequestUEFChunkSelector(path, NULL, name, "r"))
	{
		if(SnapshotAvailable(in, id))
		{
			loaded = true;
			speaker->Resync(ccpu->ReadDWord(NMOS6502ADDR_TIMECNT));

			if(in->FindId(0x0400))
			{
				in->CurrentChunk()->ReadSeek(1, SEEK_SET);

				ccpu->WriteByte(NMOS6502ADDR_A, in->CurrentChunk()->GetC());
				ccpu->WriteByte(NMOS6502ADDR_P, in->CurrentChunk()->GetC());
				ccpu->WriteByte(NMOS6502ADDR_X, in->CurrentChunk()->GetC());
				ccpu->WriteByte(NMOS6502ADDR_Y, in->CurrentChunk()->GetC());
				ccpu->WriteByte(NMOS6502ADDR_S, in->CurrentChunk()->GetC());
				ccpu->WriteByte(NMOS6502ADDR_PCL, in->CurrentChunk()->GetC());
				ccpu->WriteByte(NMOS6502ADDR_PCH, in->CurrentChunk()->GetC());
			}

			if(in->FindId(0x0401))
			{
				in->CurrentChunk()->ReadSeek(1, SEEK_SET);

				WriteValue(ULAVALUE_ICONTROL, in->CurrentChunk()->GetC());
				WriteValue(ULAVALUE_ISTATUS, in->CurrentChunk()->GetC());
				for(c = 0; c < 4; c++)
				{
					WriteValue(ULAVALUE_VBASE+2+c, in->CurrentChunk()->GetC());
				}
				WriteValue(ULAVALUE_PAGED, in->CurrentChunk()->GetC());
				for(c = 0; c < 10; c++)
				{
					WriteValue(ULAVALUE_VBASE+6+c, in->CurrentChunk()->GetC());
				}
				Fix();
			}

			if(in->FindId(0x0410))
			{
				in->CurrentChunk()->ReadSeek(1, SEEK_SET);
				addr.a = 0;

				temp = in->CurrentChunk()->GetC();
				if(!temp)
				{
					in->CurrentChunk()->Read(ccpu->GetByteReadPtr(addr), 32768);
				}
			}

			/* Multiplexing, ROM preferences, etc */
			SetGlobals(in);

			if(Multiplexable(in))
			{
				if(in->FindId(0x0411))
				{
					in->CurrentChunk()->ReadSeek(1, SEEK_SET);
					addr.a = 0;

					temp = in->CurrentChunk()->GetC();
					if(!temp)
					{
						c = (in->CurrentChunk()->GetLength() - 2) >> 2;
						while(c--)
						{
							quad.b.b0 = in->CurrentChunk()->GetC();
							quad.b.b1 = in->CurrentChunk()->GetC();
							quad.b.b2 = in->CurrentChunk()->GetC();
							quad.b.b3 = in->CurrentChunk()->GetC();
							((C_6502Base *)cpuBig)->WriteDWord(addr.a, quad.a);
							addr.a++;
						}
					}
				}


				if(in->FindId(0x0412))
				{
					in->CurrentChunk()->ReadSeek(1, SEEK_SET);

					quad.b.b0 = in->CurrentChunk()->GetC();
					quad.b.b1 = in->CurrentChunk()->GetC();
					quad.b.b2 = in->CurrentChunk()->GetC();
					quad.b.b3 = in->CurrentChunk()->GetC();
					((C_6502Base *)cpuBig)->WriteDWord(NMOS6502ADDR_A, quad.a);

					quad.b.b0 = in->CurrentChunk()->GetC();
					quad.b.b1 = in->CurrentChunk()->GetC();
					quad.b.b2 = in->CurrentChunk()->GetC();
					quad.b.b3 = in->CurrentChunk()->GetC();
					((C_6502Base *)cpuBig)->WriteDWord(NMOS6502ADDR_P, quad.a);

					quad.b.b0 = in->CurrentChunk()->GetC();
					quad.b.b1 = in->CurrentChunk()->GetC();
					quad.b.b2 = in->CurrentChunk()->GetC();
					quad.b.b3 = in->CurrentChunk()->GetC();
					((C_6502Base *)cpuBig)->WriteDWord(NMOS6502ADDR_X, quad.a);

					quad.b.b0 = in->CurrentChunk()->GetC();
					quad.b.b1 = in->CurrentChunk()->GetC();
					quad.b.b2 = in->CurrentChunk()->GetC();
					quad.b.b3 = in->CurrentChunk()->GetC();
					((C_6502Base *)cpuBig)->WriteDWord(NMOS6502ADDR_Y, quad.a);
				}
			}
		}

		fs->ReleaseUEFChunkSelector(in);

		return loaded;
	}

	return false;
}

void UEFWriteDWord(C_UEFChunkSelector *out, unsigned __int32 value)
{
	BrokenQuad v;

	v.a = value;
	out->CurrentChunk()->PutC(v.b.b0);
	out->CurrentChunk()->PutC(v.b.b1);
	out->CurrentChunk()->PutC(v.b.b2);
	out->CurrentChunk()->PutC(v.b.b3);
}

bool C_ULA::SaveSnapshot(char *path, char *name, unsigned __int8 id)
{
	C_UEFChunkSelector *out;
	int c;
	BrokenWord addr;
	MultiState state;

	if(out = fs->RequestUEFChunkSelector(path, NULL, name, "w"))
	{
		out->CurrentChunk()->SetId(0x0005);
		out->CurrentChunk()->PutC(17);

		//6502 state
		out->EstablishChunk();
		out->CurrentChunk()->SetId(0x0400);
		out->CurrentChunk()->PutC(id);

		out->CurrentChunk()->PutC(ccpu->ReadByte(NMOS6502ADDR_A));
		out->CurrentChunk()->PutC(ccpu->ReadByte(NMOS6502ADDR_P));
		out->CurrentChunk()->PutC(ccpu->ReadByte(NMOS6502ADDR_X));
		out->CurrentChunk()->PutC(ccpu->ReadByte(NMOS6502ADDR_Y));
		out->CurrentChunk()->PutC(ccpu->ReadByte(NMOS6502ADDR_S));
		out->CurrentChunk()->PutC(ccpu->ReadByte(NMOS6502ADDR_PCL));
		out->CurrentChunk()->PutC(ccpu->ReadByte(NMOS6502ADDR_PCH));

		//ULA state
		out->EstablishChunk();
		out->CurrentChunk()->SetId(0x0401);
		out->CurrentChunk()->PutC(id);

		out->CurrentChunk()->PutC(ReadValue(ULAVALUE_ICONTROL));
		out->CurrentChunk()->PutC(ReadValue(ULAVALUE_ISTATUS));
		out->CurrentChunk()->PutC(ReadValue(ULAVALUE_VBASE + 2));
		out->CurrentChunk()->PutC(ReadValue(ULAVALUE_VBASE + 3));
		out->CurrentChunk()->PutC(ReadValue(ULAVALUE_VBASE + 4));
		out->CurrentChunk()->PutC(ReadValue(ULAVALUE_VBASE + 5));
		out->CurrentChunk()->PutC(ReadValue(ULAVALUE_PAGED));
		for(c = 0; c < 10; c++)
		{
			out->CurrentChunk()->PutC(ReadValue(ULAVALUE_VBASE + c + 6));
		}
		out->CurrentChunk()->PutC(0);
		out->CurrentChunk()->PutC(0);
		out->CurrentChunk()->PutC(0);
		out->CurrentChunk()->PutC(0);

		//memory state
		out->EstablishChunk();
		out->CurrentChunk()->SetId(0x0410);
		out->CurrentChunk()->PutC(id);
		out->CurrentChunk()->PutC(0);

		addr.a = 0;
		out->CurrentChunk()->Write(ccpu->GetByteReadPtr(addr), 32768);

		state = GetAvailableState();

		//save multiplex data if applicable
		if(state.multicpu)
		{
			//multiplexing state
			out->EstablishChunk();
			out->CurrentChunk()->SetId(0x0006);
			out->CurrentChunk()->PutC(1);
			out->CurrentChunk()->PutC(9);

			//6502 state
			out->EstablishChunk();
			out->CurrentChunk()->SetId(0x0412);
			out->CurrentChunk()->PutC(id);
			UEFWriteDWord(out, ((C_6502Base *)cpuBig)->ReadDWord(NMOS6502ADDR_A));
			UEFWriteDWord(out, ((C_6502Base *)cpuBig)->ReadDWord(NMOS6502ADDR_P));
			UEFWriteDWord(out, ((C_6502Base *)cpuBig)->ReadDWord(NMOS6502ADDR_X));
			UEFWriteDWord(out, ((C_6502Base *)cpuBig)->ReadDWord(NMOS6502ADDR_Y));

			//memory state
			out->EstablishChunk();
			out->CurrentChunk()->SetId(0x0411);
			out->CurrentChunk()->PutC(id);
			out->CurrentChunk()->PutC(0);

			addr.a = 0;
			c = 32768;
			while(c--)
			{
				UEFWriteDWord(out, ((C_6502Base *)cpuBig)->ReadDWord(addr.a));
				addr.a++;
			}

			//palette
			out->EstablishChunk();
			out->CurrentChunk()->SetId(0x0007);
			c = 256;
			while(c--)
			{
				out->CurrentChunk()->PutC((unsigned __int8)(255.0f*gfx->GetPaletteR(255-c, PAL_ELK)));
				out->CurrentChunk()->PutC((unsigned __int8)(255.0f*gfx->GetPaletteG(255-c, PAL_ELK)));
				out->CurrentChunk()->PutC((unsigned __int8)(255.0f*gfx->GetPaletteB(255-c, PAL_ELK)));
			}
		}

		fs->ReleaseUEFChunkSelector(out);

		return true;
	}

	return false;
}