#ifndef ULA_H
#define ULA_H

#include "../6502/6502.h"

class C_screen;
class C_input;
class C_gfx_api;
class C_tape;
class C_SoundProvider;
class C_Speaker;
class C_config;
class C_timer;
class C_6502Regular;
class C_6502Wide;
class C_UEFChunkSelector;
class C_FSAccess;
class C_GUI;

class C_Plus1;

class C_WD1770;
class C_drive;

#define ULA_NUMROMS			13
const int cind[]= { 0, 1, 2, 3, 4, 5, 6, 7, 10, 12, 13, 14, 15 };
const int revcind[]= { 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 8, 8, 9, 10, 11, 12 };

struct MultiState
{
	bool multicpu, multiscr;
	bool usestate, palonly;
	unsigned __int8 gstyle;
};

enum MRB_MODES
{
	MRBMODE_SHADOW, MRBMODE_TURBO, MRBMODE_OFF, MRBMODE_UNSET
};

class C_ULA : C_6502CalloutDevice
{
	public:
		C_ULA(void);
		~C_ULA(void);

		/* for the 6502 to relate to */
		bool Keepalive(int counter);
		void MinimiseTimer(int subvalue);

		int ReadFrom(BrokenWord addr, unsigned __int8 *value, unsigned __int32 *bvalue, int time);
		int WriteTo(BrokenWord addr, unsigned __int8 value, unsigned __int32 bvalue, int time);

		bool TrapOpcode(unsigned __int8 opcode);
		void UnknownOpcode(unsigned __int8 opcode);
		void VirtualOpcode(unsigned __int8 opcode, unsigned __int8 argument, int time);
		void KillOpcode(unsigned __int8 opcode);

		/* for the screen, moved here to fix 'shadow' mode emulation */
		unsigned __int8 *GetRAMBasePtr(void);
		unsigned __int32 *GetWideRAMBasePtr(void);

		/* actual ULA functionality */
		bool Setup(void);
		char *GetError(void);
		void SetQuit(void);

		void CloseDown(void);
		void Go(void);
		void Reset(void);
		void SuperReset(void);

		/* stuff for manipulating ROMs */
		unsigned __int8 *GetROMPtr(int rnum);
		void SetROMPtr(int rnum, unsigned __int8 *ptr);
		void FreeROM(int rnum);
		int GetHardwareROMSlot(int rnum);

		void SetROMPresent(int rnum, int mode);
		int GetROMPresent(int rnum);

		char *GetROMString(int rnum);

		void SetROMLoaded(int rnum, bool truth);
		bool GetROMLoaded(int rnum);

		bool LoadROM(char *name, int rnum, char *newname);

		void DoPage(void);

		/* thing for invalid opcodes */
		void SetToIgnore(unsigned __int8 id);

		/* IRQ thing */
		void AddIRQ(unsigned __int8 val);
		void RemoveIRQ(unsigned __int8 val);

		/* Read / Writes to ULA variables, principly for snapshot support */
		unsigned __int8 ReadValue(int id);
		void WriteValue(int id, unsigned __int8 value);
		void Fix(void);

		/* FAST tape bits */
		void InstallTapeHack(void);
		void RemoveTapeHack(void);
		bool TapeHackPresent(void);
		bool PageTape(void);
		void UnPageTape(void);
		void WriteTapeString(char *text);

		/* Slogger MRB bits */
		void SetSloggerMRBMode(MRB_MODES mode);
		MRB_MODES GetSloggerMRBMode(void);

		/* multiplexing handling bits */
		MultiState GetMultiState(void);
		void SetMultiState(MultiState newstate);
		bool ProposeMultiState(MultiState newstate);
		MultiState GetAvailableState(void);

		bool Multiplexable(C_UEFChunkSelector *cnk);
		void SetGlobals(C_UEFChunkSelector *cnk);

		/* JIM bits */
		bool GetJimEnabled(void);
		void SetJimEnabled(bool);

		/* autoload */
		void SetAutoloadID(unsigned __int8 v);

		/* for loading and saving state snapshots */
		bool LoadSnapshot(char *path, char *name, unsigned __int8 id);
		bool SaveSnapshot(char *path, char *name, unsigned __int8 id);
		bool SnapshotUseful(C_UEFChunkSelector *in);
		bool SnapshotAvailable(C_UEFChunkSelector *in, unsigned __int8 id);

		/* First Byte joystick */
		int GetJoystickId(void);
		void SetJoystickId(int id);

		/* this provides a way to make the Electron aware it has been asleep */
		void WakeUp(void);

		/* referrals */
		C_screen *GetScreen(void);
		C_6502Base *GetCPU(void);
		C_6502Base *GetRCPU(void);
		C_6502Base *GetWCPU(void);
		C_input *GetInput(void);
		C_gfx_api *GetGFX(void);
		C_tape *GetTape(void);
		C_Speaker *GetSpeaker(void);
		C_SoundProvider *GetSound(void);
		C_config *GetConfig(void);
		C_WD1770 *GetDisc(void);
		C_timer *GetTimer(void);
		C_FSAccess *GetFS(void);
		C_Plus1 *GetPlus1(void);

		/* two disc drives are supported nowadays */
		C_drive *GetCurrentDrive(void);
		void SetCurrentDrive(int num);
		C_drive *GetParticularDrive(int num);

	private :
		/* private JIM bits */
		void SetupJIM(void);
		void CloseJIM(void);
		void PageJIM(int page);
		unsigned __int8 **JimRAM;
		unsigned __int32 **WideJimRAM;
		int JimStack[2];	//a 2 element stack
		bool JimOn;

		/* autoload bits */
		unsigned __int8 AutoID;

		/* First Byte joystick */
		int joyid;
		bool joyon;

		/* ROM stuff */
		void FixWROM(void);
		void SelectWROM(unsigned __int8 mode);
		unsigned __int8 *TapeRom;

		/* timing */
		int tcounter;

		/* fast tape hack */
		bool ConsiderTapeHack(BrokenWord addr);
		void ConstructTapeRom(void);
		void DestroyTapeRom(void);
		bool TapeHacked, TapePrinting;
		char *textptr;

		/* misc. */
		bool SeedExternal(void);
		void EndExternal(void);

		MRB_MODES MRBMode;
		C_6502Base *ccpu, *newcpu;

		unsigned __int8 *RamPtr;
		unsigned __int8 *RomPtrs[ULA_NUMROMS+1];
		unsigned __int32 *WideRamPtr;

		unsigned __int32 *WideRomPtrs[ULA_NUMROMS];
		unsigned __int32 *WideOSRoms[3], *CWideOSRom;

		unsigned __int8 *ScratchPad;

		bool InstIgnore[256];

		bool switchcpu;
		MultiState set, available;

		bool RomLoaded[ULA_NUMROMS+1];
		int RomMode[ULA_NUMROMS+1];
		char *RomNames[ULA_NUMROMS+1];

		unsigned __int8 IntStatus, IntControl, PagedRom, LastPage, MemPage;
		unsigned __int8 Values[16];

		unsigned int KeepaliveCounter1, KeepaliveCounter2, KeepaliveCounter3;
		char *errstring;

		unsigned __int8 ErrorOpcode;
		bool forcegui;

		C_gfx_api *gfx;
		C_screen *scr;
		C_config *settings;
		C_Speaker *speaker;
		C_SoundProvider *sound;
		C_tape *tape;
		C_timer *slowdown;
		C_6502Regular *cpu;
		C_6502Wide *cpuBig;
		C_input *input;
		C_FSAccess *fs;
		C_GUI *gui;

		/* the Plus 1 */
		C_Plus1 *plus1;

		/* the Plus 3 */
		C_drive *drives[2], *cdrive;
		C_WD1770 *disc;
};

#define ULAJIM_POP			256
#define ULAJIM_ROM			257
#define ULAJIM_TAPE			258

#define ULAIRQ_MASTER		0x01
#define ULAIRQ_POWER		0x02
#define ULAIRQ_DISPLAY		0x04
#define ULAIRQ_RTC			0x08
#define ULAIRQ_SEND			0x10
#define ULAIRQ_RECEIVE		0x20
#define ULAIRQ_HTONE		0x40

#define ULACLEAR_NMI		0x80
#define ULACLEAR_HTONE		0x40
#define ULACLEAR_RTC		0x20
#define ULACLEAR_DISPLAY	0x10

#define ULAVALUE_VBASE		0x00
#define ULAVALUE_ICONTROL	0x10
#define ULAVALUE_ISTATUS	0x11
#define ULAVALUE_PAGED		0x12
#define ULAVALUE_OPCODE		0x13

#define ULAROMMODE_OFF		0
#define ULAROMMODE_ON		1
#define ULAROMMODE_RAM		2

#define VOP_TAPE			0x8b
#define VOP_AUTO			0x93

#define ULAROM_OS			ULA_NUMROMS

#define TimeToSample(x, y) (((int)( (float)(x) / ((20000.0f/441.0f)*y) ))&SOUND_MASK)

#endif
