/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	ula/ulamisc.cpp
	===============

	Features some more ULA stuff

*/

#include "../screen/screen.h"
#include "ula.h"
#include "../tape/tape.h"
#include "../gui/gui.h"
#include "../gui/romclasses.h"
#include "../wd1770/wd1770.h"
#include "../misc/timing.h"
#include "../sound/speaker.h"
#include <malloc.h>

/* the snapshot related things */
unsigned __int8 C_ULA::ReadValue(int id)
{
	if(id < 16)
		return Values[id];

	switch(id)
	{
		case ULAVALUE_ICONTROL :	return IntControl;
		case ULAVALUE_ISTATUS :		return IntStatus;
		case ULAVALUE_PAGED :		return PagedRom;
		case ULAVALUE_OPCODE :		return ErrorOpcode;
	}

	return 0;
}

void C_ULA::WriteValue(int id, unsigned __int8 value)
{
	if(id < 16)
		Values[id] = value;

	switch(id)
	{
		case ULAVALUE_ICONTROL :	IntControl = value;	break;
		case ULAVALUE_ISTATUS :		IntStatus = value;	break;
		case ULAVALUE_PAGED :		PagedRom = value;	break;
	}
}

void C_ULA::Fix(void)
{
	int c;
	BrokenWord addr;
	unsigned __int8 paged;

	paged = PagedRom;

	addr.a = 0xfe02;
	WriteTo(addr, Values[2], 0, 0);
	addr.a = 0xfe03;
	WriteTo(addr, Values[3], 0, 0);
	addr.a = 0xfe04;
	WriteTo(addr, Values[4], 0, 0);

	for(c = 6; c < 16; c++)
	{
		addr.a = 0xfe00 + c;
		WriteTo(addr, Values[c], 0, 0);
	}

	if(IntStatus&ULAIRQ_MASTER)
		ccpu->SetIRQ();

	PagedRom = paged;
	DoPage();
}

void C_ULA::DoPage(void)
{
	if(PagedRom != LastPage)
	{
		int trom = revcind[PagedRom];
		LastPage = PagedRom;

		if(!(PagedRom == 8 || PagedRom == 9) && RomMode[trom])
		{
			ccpu->FastSetAddrPtrs(128, 64, RomPtrs[trom], (RomMode[trom] == ULAROMMODE_RAM) ? RomPtrs[trom] : ScratchPad);
		}
		else
			ccpu->FastSetAddrFlags(128, 64, true);
	}
}

char *C_ULA::GetError(void)
{
	return errstring;
}

void C_ULA::Go(void)
{
	speaker->Unpause(0);

	/*

	a CPU switch may already be required - e.g. if a multiplexed file was
	requested on the command line or via the 'load = ' config file option

	*/
	if(switchcpu)
		newcpu->LeechState(ccpu);

	do
	{
		switchcpu = false;
		ccpu = newcpu;

		scr->ResetCPU();
		slowdown->Resync();
		ccpu->SetQuit(false);
		ccpu->Go();

		newcpu->LeechState(ccpu);

	}while(switchcpu);
}

void C_ULA::AddIRQ(unsigned __int8 val)
{
	if(IntControl & val)
	{
		IntStatus |= val | ULAIRQ_MASTER;
		ccpu->SetIRQ();
	}

	scr->AlertToInterrupt(tcounter);
}

void C_ULA::SuperReset(void)
{
	int c;
	c = 256;
	while(c--)
		InstIgnore[c] = false;

	IntStatus = ULAIRQ_POWER;
	IntControl = 0;

	available.multicpu = available.multiscr = false;
	ProposeMultiState(available);

	PagedRom = 10;
	DoPage();

	scr->Reset();
	Reset();
}

void C_ULA::Reset(void)
{
	AutoID = 0xff;
	gui->PassROMHint(false, ROMCLASS_RESET, NULL);

	joyon = (joyid != -1) ? true : false;

	/* establish MRB mode */
	BrokenWord addr;
	addr.a = 0xfc7f;
	WriteTo(addr, 0x00, 0, 0);

/*	if(MRBMode == MRBMODE_SHADOW)
		ccpu->SetAddrPtrs(0, 128, &RamPtr[32768], NULL, true, true);
	else
		ccpu->SetAddrPtrs(0, 128, RamPtr, NULL, true, true);*/

	ccpu->Reset();
	disc->Reset();
	tape->SignalReset();
	SetMultiState(set);
}

void C_ULA::SetToIgnore(unsigned __int8 opcode)
{
	InstIgnore[opcode] = true;
}

bool C_ULA::TrapOpcode(unsigned __int8 opcode)
{
	switch(opcode)
	{
		default :
		return InstIgnore[opcode];

		case VOP_TAPE :
		case VOP_AUTO :
		return true;
	}
	return false;
}

void C_ULA::UnknownOpcode(unsigned __int8 opcode)
{
	ErrorOpcode = opcode;
	gui->Go(GUISCR_OPCODEERROR);
	slowdown->Resync();
}

void C_ULA::KillOpcode(unsigned __int8 opcode)
{
	ErrorOpcode = opcode;
	gui->Go(GUISCR_HALT);
	slowdown->Resync();
}

#include "../6502/6502.h"
#include "../6502/regular/regular.h"

void C_ULA::VirtualOpcode(unsigned __int8 opcode, unsigned __int8 argument, int time)
{
	switch(opcode)
	{
		case VOP_TAPE :
			tape->HandleVOP(argument, TimeToSample(time, slowdown->GetMultiplier()), time);
		break;

		case VOP_AUTO :

			/* for now : no autoload */
			cpu->WriteByte(NMOS6502ADDR_X, AutoID);

			BrokenWord pc;

			pc.b.l = cpu->ReadByte(NMOS6502ADDR_PCL);
			pc.b.h = cpu->ReadByte(NMOS6502ADDR_PCH);

			cpu->Push(pc.a-2);
		break;
	}
}

MultiState C_ULA::GetMultiState(void)
{
	return set;
}

MultiState C_ULA::GetAvailableState(void)
{
	return available;
}

void C_ULA::SetMultiState(MultiState newstate)
{
	set = newstate;

	ProposeMultiState(available);
}

bool C_ULA::ProposeMultiState(MultiState newstate)
{
	bool gfxmulti;

	available = newstate;

	if(!set.usestate)
	{
		newcpu = set.multicpu ? (C_6502Base *)cpuBig : (C_6502Base *)cpu;
		gfxmulti = set.multiscr;
	}
	else
	{
		newcpu = (available.multicpu && set.multicpu) ? (C_6502Base *)cpuBig : (C_6502Base *)cpu;
		gfxmulti = (available.multiscr && set.multiscr) ? true : false;
	}

	if(newcpu != ccpu)
	{
		switchcpu = true;
		ccpu->SetQuit(true);
	}

	if(scr) scr->SetMultiMode(gfxmulti, available.gstyle);

	return (newcpu == (C_6502Base *)cpuBig) ? true : false;
}

void C_ULA::SetQuit(void)
{
	switchcpu = false;
	ccpu->SetQuit(true);
}

void C_ULA::ConstructTapeRom(void)
{
	unsigned __int8 TapeInstr[] =
	{
		0x48,
		0x08,
		0x8a,
		0x48,

		0x4c, 0x0a, 0xfd,

		0x20, 0xee, 0xff,

		0xad, 0xff, 0xfd,
		0xd0, 0xf8,

		0x68,
		0xaa,
		0x28,
		0x68,

		0x8b, 0x01
	};

	TapeRom = (unsigned __int8 *)malloc(256);
	memcpy(TapeRom, TapeInstr, 0x22);
}

void C_ULA::DestroyTapeRom(void)
{
	free(TapeRom);
}

void C_ULA::WriteTapeString(char *text)
{
	textptr = text;
}

#include <stdio.h>

bool C_ULA::PageTape(void)
{
	if(!TapeRom) return false;

	PageJIM(ULAJIM_TAPE);
	ccpu->Push(0xfcff);

	TapePrinting = true;

	return true;
}

void C_ULA::UnPageTape(void)
{
	if(!TapeRom) return;

	TapePrinting = false;
	PageJIM(ULAJIM_POP);
}
/* FastTape ROM (really just a 'print string' loop, using a 'next char' reg at fdff) :
.entry										[0xFD00]
	PHA					0x48
	PHP					0x08
	TXA					0x8a
	PHA					0x48

	JMP loadpart		0x4c 0x0a 0xfd

.wrloop										[0xFD07]
	JSR &FFEE			0x20 0xee 0xff

.loadpart									[0xFD0a]
	LDA &FDFF			0xad 0xff 0xfd
	BNE wrloop			0xd0 0xf8

	PLA					0x68
	TAX					0xaa
	PLP					0x28
	PLA					0x68

	VOP finish			0x8b 0x01
*/

C_ULA::C_ULA(void)
{
	int c;

	/* NULLify all connected devices */
	ccpu = newcpu = NULL;
	gfx = NULL;
	scr = NULL;
	settings = NULL;
	speaker = NULL;
	tape = NULL;
	drives[0] = drives[1] = NULL;
	disc = NULL;
	slowdown = NULL;
	cpu = NULL;
	cpuBig = NULL;
	input = NULL;
	fs = NULL;

	MRBMode = MRBMODE_UNSET;

	/* NULLify all ROM bits */
	c = ULA_NUMROMS+1;
	while(c--)
	{
		RomNames[c] = NULL;
		RomPtrs[c] = NULL;
		RomLoaded[c] = false;
		RomMode[c] = ULAROMMODE_OFF;
	}

	c = ULA_NUMROMS;
	while(c--)
	{
		WideRomPtrs[c] = NULL;
	}
	WideOSRoms[0] = WideOSRoms[1] = WideOSRoms[2] = NULL;

	/* NULLify RAM bits */
	RamPtr = ScratchPad = TapeRom = NULL;
	WideRamPtr = NULL;
	errstring = NULL;
}

C_ULA::~C_ULA(void)
{
}

void C_ULA::SetSloggerMRBMode(MRB_MODES mode)
{
	switch(mode)
	{
		case MRBMODE_OFF :

			MRBMode = MRBMODE_OFF;
			ccpu->SetAddrFlags(0, 128, false, RAM);

		break;

		case MRBMODE_SHADOW :

			MRBMode = MRBMODE_SHADOW;
			ccpu->SetAddrFlags(0, 128, false, TWOMhz_BUS);

		break;

		case MRBMODE_TURBO :

			MRBMode = MRBMODE_TURBO;
			ccpu->SetAddrFlags(0, 32, false, TWOMhz_BUS);

		break;
	}

	ccpu->ReconsiderTiming();
}

MRB_MODES C_ULA::GetSloggerMRBMode(void)
{
	return MRBMode;
}

void C_ULA::WakeUp(void)
{
	slowdown->Resync();
}

void C_ULA::SetAutoloadID(unsigned __int8 v)
{
	AutoID = v;
	gui->SetAutoId(v);
}

/* JIM */

bool C_ULA::GetJimEnabled(void)
{
	return JimOn;
}

void C_ULA::SetJimEnabled(bool t)
{
	if(!t)
		PageJIM(ULAJIM_ROM);

	JimOn = t;

	if(t)
		PageJIM(0);
}

void C_ULA::SetupJIM(void)
{
	JimRAM = (unsigned __int8 **)malloc(sizeof(unsigned __int8 *)*256);
	WideJimRAM = (unsigned __int32 **)malloc(sizeof(unsigned __int32 *)*256);
	int c = 256;
	while(c--)
	{
		JimRAM[c] = (unsigned __int8 *)calloc(256, sizeof(unsigned __int8));
		WideJimRAM[c] = (unsigned __int32 *)calloc(256, sizeof(unsigned __int32));
	}

	PageJIM(ULAJIM_ROM);
}

void C_ULA::CloseJIM(void)
{
	if(JimRAM)
	{
		int c = 256;
		while(c--)
		{
			free(JimRAM[c]);
			free(WideJimRAM[c]);
		}

		free(JimRAM);
		free(WideJimRAM);

		JimRAM = NULL;
	}
}

void C_ULA::PageJIM(int page)
{
	if((page < 256) && !JimOn) return;

	if(page != ULAJIM_POP)
	{
		JimStack[1] = JimStack[0];
		JimStack[0] = page;

		if(JimStack[1] < 256)
		{
			ccpu->SetAddrPtrs(0xfd, 1, ScratchPad, (unsigned __int32 *)ScratchPad, false, true);
		}
	}

	switch(page)
	{
		default :

			cpu->SetAddrFlags(0xfd, 1, false, ONEMhz_BUS);
			ccpu->SetAddrPtrs(0xfd, 1, JimRAM[page], WideJimRAM[page], true, true);

		break;

		case ULAJIM_POP :

			JimStack[0] = JimStack[1];
			PageJIM(JimStack[0]);

		break;

		case ULAJIM_ROM :

			cpu->SetAddrFlags(0xfd, 1, false, ONEMhz_BUS);
			ccpu->SetAddrPtrs(0xfd, 1, &RomPtrs[ULAROM_OS][0xfd00 - 0xc000], &CWideOSRom[0xfd00 - 0xc000], true, false);

		break;

		case ULAJIM_TAPE :

			cpu->SetAddrFlags(0xfd, 1, true, ONEMhz_BUS);
			ccpu->SetAddrPtrs(0xfd, 1, TapeRom, NULL, true, false);

		break;
	}
}

