/*

	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This file at some time additionally worked upon by Paul Boddie

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	ula/ulasetup.cpp
	===============

	Features the ULA setup routines, the unknown opcode callouts,
	the ULA level IRQ bits, the install and deinstall of the tape hack,
	and the class pointer claim things.

*/
#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <string.h>

#include "../6502/regular/regular.h"
#include "../6502/wide/wide.h"
#include "ula.h"

#include "../gui/gui.h"

#include "../screen/screen.h"
#include "../wd1770/wd1770.h"
#include "../wd1770/drive.h"
#include "../tape/tape.h"
#include "../sound/speaker.h"
#include "../misc/timing.h"
#include "../misc/input.h"
#include "../plus1/plus1.h"

#include "../misc/config.h"

#include "../fs/fs.h"

#define GENERAL_DEFINES
#include "../targets.h"
#undef GENERAL_DEFINES

#define CheckInterrupts()	\
	if(!(IntStatus & (ULAIRQ_DISPLAY | ULAIRQ_RTC | ULAIRQ_HTONE | ULAIRQ_SEND | ULAIRQ_RECEIVE)))\
	{\
		ccpu->ClearIRQ();\
		IntStatus &= ~ULAIRQ_MASTER;\
	}

void C_ULA::RemoveIRQ(unsigned __int8 val)
{
	IntStatus &= ~val;
	CheckInterrupts();
}

bool C_ULA::Keepalive(int counter)
{
	int SampleTime = TimeToSample(counter, slowdown->GetMultiplier());

	tcounter = counter;

	if(!scr->Keepalive(counter))
	{
		unsigned __int8 s;

		slowdown->RegisterBeat();
		input->UpdateKeyboard();

		s = input->GetShifts();

		if(s&IPT_SHFT_QUIT)
			SetQuit();

		if(s&IPT_SHFT_MENU || forcegui)
		{
			bool internalgui = true;
			forcegui = false;

			if(settings->FindSetting("extgui", BOOLEAN_RESULT))
			{
				if(settings->ReadSetting().result_bool)
				{
					internalgui = false;
					SaveSnapshot("~/", "cntnuity.uef", 0);
					SetQuit();
				}
			}

			if(internalgui)
			{
				speaker->Pause(SampleTime);
				gui->Go(GUISCR_DEFAULT);
				speaker->Unpause(SampleTime);
				slowdown->Resync();
			}
		}

		if(s&IPT_SHFT_BREAK)
		{
			if(s&IPT_SHFT_SPRBRK)
				SuperReset();
			else
				Reset();
		}

		check_quit(SetQuit());
	}

	sound->Keepalive(SampleTime);
	tape->Keepalive(SampleTime, counter);
	speaker->Keepalive(SampleTime);

	KeepaliveCounter3--;
	if(!KeepaliveCounter3)
	{
		KeepaliveCounter3 = 312*50;
		return true;
	}

	return false;
}

void C_ULA::MinimiseTimer(int sub)
{
	int SampleSub = TimeToSample(sub, slowdown->GetMultiplier());

	speaker->MinimiseTimer(SampleSub);
	tape->MinimiseTimer(SampleSub, sub);
//	sound->MinimiseTimer(SampleSub);

	scr->MinimiseTimer(sub);
	plus1->MinimiseTimer(sub);
}

bool C_ULA::Setup(void)
{
	KeepaliveCounter3 = 1;
	KeepaliveCounter2 = 13;
	KeepaliveCounter1 = 312*50;

	errstring = NULL;
	JimRAM = NULL;
	forcegui = TapePrinting = false;

	/* create instances of FS related classes */
	settings = new C_config();
	fs = new C_FSAccess(settings, "~/.electrem.cfg");

	/* determine multiplex options */
	MultiState str;

	str.multicpu = str.multiscr = str.usestate = str.palonly = false;

	if(settings->FindSetting("multicpu", BOOLEAN_RESULT))
	{
		str.multicpu = settings->ReadSetting().result_bool;
	}

	if(settings->FindSetting("multiscr", BOOLEAN_RESULT) && str.multicpu)
	{
		str.multiscr = settings->ReadSetting().result_bool;
	}

	if(settings->FindSetting("readuefmulti", BOOLEAN_RESULT))
	{
		str.usestate = settings->ReadSetting().result_bool;
	}

	if(settings->FindSetting("acceptpal", BOOLEAN_RESULT))
	{
		str.palonly = settings->ReadSetting().result_bool;
	}

	SetMultiState(str);

	/*

	create new instances of all the relevant classes

	*/

	gfx = new C_gfx_api();
	scr = new C_screen(gfx);

	speaker = new C_Speaker();
	sound = new C_SoundProvider();

	tape = new C_tape();
	cdrive = drives[0] = new C_drive(this);
	drives[1] = new C_drive(this);
	disc = new C_WD1770();

	slowdown = new C_timer();
	input = new C_input(this);

	cpu = new C_6502Regular();
	cpuBig = new C_6502Wide();

	gui = new C_GUI();
	plus1 = new C_Plus1();

	/* establish default paths */
	fs->SetPath("/", "startdir");
	fs->SetPath("roms/", "rompath");
	fs->SetPath("~/", "datapath");
	fs->SetPath("~/", "keybpath");

	/* setup CPU stuff */
	ccpu = cpu;

	if(!cpu->Setup(true))
	{
		errstring = "Error allocating CPU work tables";
		return false;
	}

	cpu->SetCalloutDevice(this);
	cpu->SetCalloutTime(128);
	cpuBig->Setup(false);
	cpuBig->LeechTables(cpu);
	cpuBig->LeechState(cpu);

	/* Create some RAM, with 64kb sideways RAM and 16kb overspill for ROM writes */
	if(!(RamPtr = (unsigned __int8 *)malloc(64*1024)))
	{
		errstring = "Could not allocate main RAM";
		return false;
	}

	memset(RamPtr, 255, 64*1024);

	if(!(WideRamPtr = (unsigned __int32 *)malloc(64*1024*4)))
	{
		errstring = "Could not allocate wide main RAM";
		return false;
	}

	if(!(ScratchPad = (unsigned __int8 *)malloc(16*1024*4)))
	{
		errstring = "Could not allocate scratch pad work area";
		return false;
	}

	cpu->SetAddrFlags(0, 128, false, RAM);
	cpu->SetAddrPtrs(0, 128, RamPtr, WideRamPtr, true, true);

	/* set flags for ROM region */
	cpu->SetAddrFlags(128, 128, false, TWOMhz_BUS);

	/* Put 16kb overspill (mapped twice) over ROM writes */
	cpu->SetAddrPtrs(128, 64, ScratchPad, (unsigned __int32 *)ScratchPad, true, true);
	cpu->SetAddrPtrs(192, 64, ScratchPad, (unsigned __int32 *)ScratchPad, true, true);

	/* Load OS */
	if(!LoadROM("os", ULAROM_OS, "os"))
	{
		errstring = "Could not load OS rom";
		return false;
	}

	/*

	make a wide copy of the OS at least good enough for the font expansion/
	implosion bits to work correctly

	*/

	if(	!(WideOSRoms[0] = (unsigned __int32 *)malloc(16*1024*4)) ||
		!(WideOSRoms[1] = (unsigned __int32 *)malloc(16*1024*4)) ||
		!(WideOSRoms[2] = (unsigned __int32 *)malloc(16*1024*4))
	)
	{
		errstring = "Could not allocate wide OS ROM area";
		return false;
	}

	unsigned short addr;
	unsigned char *ptr = GetROMPtr(ULAROM_OS);

	/* map up relevant OS colour tables */

	for(addr = 0x0000; addr < 0x4000; addr++)
	{
		WideOSRoms[0][addr] =	((ptr[addr] & 0x80) ? 0x10000000 : 0) |
								((ptr[addr] & 0x40) ? 0x01000000 : 0) |
								((ptr[addr] & 0x20) ? 0x00100000 : 0) |
								((ptr[addr] & 0x10) ? 0x00010000 : 0) |
								((ptr[addr] & 0x08) ? 0x00001000 : 0) |
								((ptr[addr] & 0x04) ? 0x00000100 : 0) |
								((ptr[addr] & 0x02) ? 0x00000010 : 0) |
								((ptr[addr] & 0x01) ? 0x00000001 : 0);

		WideOSRoms[1][addr] =	((ptr[addr]<<9)&0x1000) | ((ptr[addr]<<6)&0x2000) |
								((ptr[addr]<<6)&0x0100) | ((ptr[addr]<<3)&0x0200) |
								((ptr[addr]<<3)&0x0010) | ((ptr[addr]<<0)&0x0020) |
								((ptr[addr]<<0)&0x0001) | ((ptr[addr]>>3)&0x0002);

		WideOSRoms[2][addr] =	((ptr[addr]>>0)&0x01) | ((ptr[addr]>>1)&0x02) | ((ptr[addr]>>2)&0x04) | ((ptr[addr]>>3)&0x08) |
								((ptr[addr]<<3)&0x10) | ((ptr[addr]<<2)&0x20) | ((ptr[addr]<<1)&0x40) | ((ptr[addr]<<0)&0x80);
	}
	cpu->SetAddrPtrs(192, 64, RomPtrs[ULAROM_OS], WideOSRoms[0], true, false);

	/* set io locations & speeds */
	cpu->SetAddrFlags(0xfe, 1, true, ONEMhz_BUS);
	cpu->SetAddrFlags(0xfd, 1, true, ONEMhz_BUS); //0xfd must be a device page for the fast-tape to work, since it transiently has a register there
	cpu->SetAddrFlags(0xfc, 1, true, TWOMhz_BUS);

	/* if regular CPU is not the current CPU, allow that to catch up on state */

	if(ccpu != cpu)
	{
		ccpu->LeechState(cpu);
	}

	/* page in BASIC */
	MemPage = LastPage = 0;
	PagedRom = 10;
	DoPage();

	/* CPU INIT FINISHED */

	//initialise all parts connected to ULA which aren't the CPU
	if(!SeedExternal()) return false;

	tape->Setup(this);

	//initialise gui
	if(!gui->Initialise(this))
	{
		errstring = "problem initialising GUI - is 'newdata.uef' present?";
		return false;
	}

	/* Slogger MRB check */
	MRB_MODES mmode = MRBMODE_OFF;
	if(settings->FindSetting("sloggermrb", INTEGER_RESULT))
	{
		switch(settings->ReadSetting().result_int)
		{
			case 0 : mmode = MRBMODE_OFF; break;
			case 1 : mmode = MRBMODE_TURBO; break;
			case 2 : mmode = MRBMODE_SHADOW; break;
		}
	}
	SetSloggerMRBMode(mmode);

	/* autoload */
	AutoID = 0xff;

	/* tape hack things */
	TapeHacked = false;
	ConstructTapeRom();

	if(settings->FindSetting("tapespeed", INTEGER_RESULT))
		if(settings->ReadSetting().result_int == 3)
			InstallTapeHack();

	/* JIM */
	SetupJIM();
	if(settings->FindSetting("jimon", BOOLEAN_RESULT))
		SetJimEnabled(settings->ReadSetting().result_bool);
	else
		SetJimEnabled(false);

	/* get everything into a necessarily valid state */
	Fix();
	/* get everything into the correct state */
	SuperReset();

	/* external gui? */
	if(settings->FindSetting("extgui", BOOLEAN_RESULT))
	{
		if(settings->ReadSetting().result_bool)
		{
			LoadSnapshot("~/", "cntnuity.uef", 0);
			fs->Erase(NULL, NULL, "cntnuity.uef");
		}
	}

	/* force gui? */
	if(!forcegui)
	{
		if(settings->FindSetting("guifirst", BOOLEAN_RESULT))
		{
			forcegui = settings->ReadSetting().result_bool;
		}
		else
			forcegui = false;
	}

	gui->EvaluatePreload();

	return true;
}

void C_ULA::CloseDown(void)
{
	config_setting s;
	int c;

	EndExternal();

	/* put tape speed into config file */
	s.result_type = INTEGER_RESULT;
	s.result_int = TapeHackPresent() ? 3 : 0;
	settings->FindSetting("tapespeed", ANY_RESULT);
	settings->WriteSetting(s);
	tape->Closedown();

	/* put multiplexing state into config file */
	MultiState str = GetMultiState();

	s.result_type = BOOLEAN_RESULT;

	s.result_bool = str.multicpu;
	settings->FindSetting("multicpu", ANY_RESULT);
	settings->WriteSetting(s);

	s.result_bool = str.multiscr;
	settings->FindSetting("multiscr", ANY_RESULT);
	settings->WriteSetting(s);

	s.result_bool = str.usestate;
	settings->FindSetting("readuefmulti", ANY_RESULT);
	settings->WriteSetting(s);

	s.result_bool = str.palonly;
	settings->FindSetting("acceptpal", ANY_RESULT);
	settings->WriteSetting(s);

	/* Slogger MRB setting -> config file */
	s.result_type = INTEGER_RESULT;
	settings->FindSetting("sloggermrb", ANY_RESULT);
	switch(GetSloggerMRBMode())
	{
		case MRBMODE_OFF : s.result_int = 0; break;
		case MRBMODE_TURBO : s.result_int = 1; break;
		case MRBMODE_SHADOW : s.result_int = 2; break;
		case MRBMODE_UNSET : s = settings->ReadSetting(); break;
	}
	settings->WriteSetting(s);

	/* JIM! */
	settings->FindSetting("jimon", ANY_RESULT);
	s.result_type = BOOLEAN_RESULT;
	s.result_bool = GetJimEnabled();
	settings->WriteSetting(s);

	/*

	tell GUI that the end is nigh

	*/

	gui->DeInitialise();

	/* free some memory */

	c = ULA_NUMROMS+1;
	while(c--)
	{
		if(RomNames[c]) free(RomNames[c]);
		if(RomPtrs[c]) free(RomPtrs[c]);

		RomNames[c] = NULL;
		RomPtrs[c] = NULL;
	}

	c = ULA_NUMROMS;
	while(c--)
	{
		if(WideRomPtrs[c]) free(WideRomPtrs[c]);
		WideRomPtrs[c] = NULL;
	}

	c = 3;
	while(c--)
	{
		if(WideOSRoms[c]) free(WideOSRoms[c]);
		WideOSRoms[c] = NULL;
	}

	if(RamPtr)
	{
		free(RamPtr);
		RamPtr = NULL;
	}

	if(WideRamPtr)
	{
		free(WideRamPtr);
		WideRamPtr = NULL;
	}

	cpu->CloseDown();
	cpuBig->CloseDown();

	DestroyTapeRom();

	/* JIM */
	CloseJIM();

	/*

	delete all instances of the relevant classes, being sure to get fs last

	*/

	delete gui;

	delete scr;
	delete gfx;

	delete speaker;
	delete tape;
	delete sound;
	delete drives[0];
	delete drives[1];
	delete disc;
	delete slowdown;
	delete input;

	delete cpu;
	delete cpuBig;

	delete fs;
	delete settings;

	delete plus1;
}

/* the various 'get's */

C_6502Base *C_ULA::GetCPU(void)		{ return ccpu;		}
C_6502Base *C_ULA::GetRCPU(void)	{ return cpu;		}
C_6502Base *C_ULA::GetWCPU(void)	{ return cpuBig;	}
C_screen *C_ULA::GetScreen(void)	{ return scr;		}
C_input *C_ULA::GetInput(void)		{ return input;		}
C_gfx_api *C_ULA::GetGFX(void)		{ return gfx;		}
C_tape *C_ULA::GetTape(void)		{ return tape;		}
C_Speaker *C_ULA::GetSpeaker(void)	{ return speaker;	}
C_SoundProvider *C_ULA::GetSound(void)	{ return sound;	}
C_config *C_ULA::GetConfig(void)	{ return settings;	}
C_WD1770 *C_ULA::GetDisc(void)		{ return disc;		}
C_timer *C_ULA::GetTimer(void)		{ return slowdown;	}
C_FSAccess *C_ULA::GetFS(void)		{ return fs;		}
C_Plus1 *C_ULA::GetPlus1(void)		{ return plus1;		}

/* disc drives */
C_drive *C_ULA::GetCurrentDrive(void)
{
	return cdrive;
}

void C_ULA::SetCurrentDrive(int num)
{
	cdrive = drives[num];
}

C_drive *C_ULA::GetParticularDrive(int num)
{
	return drives[num];
}

/* fast tape hack stuff */

/*

	addresses :

		OSFILE	: f1d6
		OSARGS	: f0cc
		OSBGET	: f420
		OSBPUT	: f480
		OSGBPB	: ffa2
		OSFIND	: f320
		OSFSC	: f0e8

*/

void C_ULA::InstallTapeHack(void)
{
	ccpu->SetAddrFlags(0xf1, 1, true, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xf0, 1, true, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xf4, 1, true, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xf3, 1, true, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xff, 1, true, TWOMhz_BUS);
	TapeHacked = true;
}

void C_ULA::RemoveTapeHack(void)
{
	ccpu->SetAddrFlags(0xf1, 1, false, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xf0, 1, false, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xf4, 1, false, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xf3, 1, false, TWOMhz_BUS);
	ccpu->SetAddrFlags(0xff, 1, false, TWOMhz_BUS);
	TapeHacked = false;
}

bool C_ULA::TapeHackPresent(void)
{
	return TapeHacked;
}

bool C_ULA::ConsiderTapeHack(BrokenWord addr)
{
	unsigned __int8 pcl;

	if(!tape->GetOperationMode(TAPE_INPUTMODE_SPEEDUP))
		return false;

	pcl = ccpu->ReadByte(NMOS6502ADDR_PCL);

	switch(addr.b.h)
	{
		default : return false;

		case 0xf0 :
			if(pcl == addr.b.l && addr.b.l == 0xe8) //OSFSC
				return tape->EnterService(6);

			if(pcl == addr.b.l && addr.b.l == 0xcc) //OSARGS
				return tape->EnterService(1);
		break;

		case 0xf1 :
			if(pcl == addr.b.l && addr.b.l == 0xd6) //OSFILE
				return tape->EnterService(0);
		break;

		case 0xf3 :
			if(pcl == addr.b.l && addr.b.l == 0x20) //OSFIND
				return tape->EnterService(5);
		break;

		case 0xf4 :
			if(pcl == addr.b.l && addr.b.l == 0x20) //OSBGET
				return tape->EnterService(2);

			if(pcl == addr.b.l && addr.b.l == 0x40) //OSBPUT
				return tape->EnterService(3);
		break;

		case 0xff :
			if(pcl == addr.b.l && addr.b.l == 0xa2) //OSGBPB
				return tape->EnterService(4);
		break;

	}

	return false;
}

int C_ULA::GetJoystickId(void)
{
	return joyid;
}

void C_ULA::SetJoystickId(int id)
{
	joyid = (id < 0 || id >= input->GetNumJoysticks()) ? -1 : id;
}

unsigned __int8 *C_ULA::GetRAMBasePtr(void)
{
	return RamPtr;
}

unsigned __int32 *C_ULA::GetWideRAMBasePtr(void)
{
	return WideRamPtr;
}

