/*
	ElectrEm (c) 2000/1 Thomas Harte - an Acorn Electron Emulator

	This is open software, distributed under the GPL 2, see 'Copying' for
	details

	wd1770/drive.cpp
	================

	The disk drive emulation - lets an external device open or create a new
	disc image, step the track head, read particular sectors and select disc
	side.

	This class is used by the WD1770, and eventually the 8271, if indeed one
	is ever implemented. It was previously used by the vDFS also.

	UEF, SSD, DSD, ADF and ADL disc images (+ differently named variations)
	are supported.

*/
#include "drive.h"
#include "../misc/config.h"
#include "../types.h"
#include "../fs/fs.h"
#include "../ula/ula.h"

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

/* Constructor & Destructor - average stuff */

C_drive::C_drive(C_ULA *elk)
{
	Data1 = Data2 = NULL;
	Len1 = Len2 = 0;
	disc = NULL;
	electron = elk;
	name = path = NULL;
	gz = false;
	fileprot =
	prot = false;
}

C_drive::~C_drive(void)
{
	Close();
}

bool C_drive::Useful(C_UEFChunkSelector *selector)
{
	selector->Reset();
	return selector->FindIdMajor(0x02);
}

bool C_drive::FindSide(bool side1)
{
	unsigned char lowtemp;

	disc->ResetOverRan();
	disc->Reset();

	do
	{
		if(!disc->FindId(0x0201))
			return false;

		lowtemp = disc->CurrentChunk()->GetC();

		if((lowtemp&127) == id)
		{
			if((lowtemp&128) == (side1 ? 0 : 128))
				return true;
		}
	}
	while(!disc->OverRan());

	return false;
}

/* Open / Close - lets the drive read & write UEFs */

void C_drive::SaveTraditional(void)
{
	unsigned __int8 *savedata;
	int savelength;

	Data = NULL;

	if(Data2)
	{
		if(!FirstSide()) Flip();

		Data = (unsigned __int8 *)malloc(Tracks*Sectors*SectorLength*2);
		ReInterleave(Sectors*SectorLength);

		savedata = Data;
		savelength = DataLen;
	}
	else
	{
		savedata = Data1;
		savelength = Len1;
	}

	if(gz)
	{
		//save a GZip'd file
		gzFile outfile;

		if(outfile = electron->GetFS()->GetGZ(path, NULL, name, "wb"))
		{
			gzwrite(outfile, savedata, savelength);
			electron->GetFS()->ReleaseGZ(outfile);
		}
	}
	else
	{
		//save a normal file
		FILE *outfile;

		if(outfile = electron->GetFS()->GetFile(path, NULL, name, "wb"))
		{
			fwrite(savedata, 1, savelength, outfile);
			electron->GetFS()->ReleaseFile(outfile);
		}
	}

	if(Data)
	{
		free(Data);
	}
}

void C_drive::ReInterleave(int dlen)
{
	int dptr, sptr;

	dptr = sptr = 0;
	DataLen = 0;

	while(sptr < Len1 || sptr < Len2)
	{
		if(sptr < Len1)
		{
			memcpy(&Data[dptr], &Data1[sptr], dlen);
			DataLen = dptr+dlen;
		}
		dptr += dlen;

		if(sptr < Len2)
		{
			memcpy(&Data[dptr], &Data2[sptr], dlen);
			DataLen = dptr+dlen;
		}
		dptr += dlen;
		sptr += dlen;
	}
}

void C_drive::DeInterleave(int dlen)
{
	int dptr, sptr;

	Len1 = Len2 = 0;
	dptr = sptr = 0;

	Data1 = (unsigned __int8 *)malloc(Tracks*Sectors*SectorLength);
	Data2 = (unsigned __int8 *)malloc(Tracks*Sectors*SectorLength);

	while(dptr < DataLen)
	{
		memcpy(&Data1[sptr], &Data[dptr], dlen);
		dptr += dlen;
		Len1 += dlen;

		if(dptr < DataLen)
		{
			memcpy(&Data2[sptr], &Data[dptr], dlen);
			dptr += dlen;
			Len2 += dlen;
		}

		sptr += dlen;
	}
}

unsigned __int8 CheckSum(unsigned __int8 *ptr)
{
	int c = 256;
	unsigned __int8 ret = 0;

	while(c--)
	{
		ret += *ptr;
		ptr++;
	}

	return ret;
}

bool C_drive::NewSide(FSTypes type)
{
	unsigned __int8 *SectPtr;

	if(!(Data1 = (unsigned __int8 *)calloc(Sectors*SectorLength*Tracks, 1))) return false;

	if(type == FS_ADFS || type == FS_DBLADFS)
	{
		/* write an empty ADFS directory tree */

		/* two sided disc needs extra work memory for emulation */
		if(type == FS_DBLADFS)
			if(!(Data2 = (unsigned __int8 *)calloc(Sectors*SectorLength*Tracks, 1)))
			{
				free(Data1);
				Data1 = NULL;
				return false;
			}


		/* SECTOR 0 : free space map */
		SectPtr = Data1;

		/* free space : starts at sector 7 */
		SectPtr[0] = 7;
		/* and there are either 2*80*16 = 2560 = 10*256 or half that sectors
		on this disc */
		SectPtr[253] = (type == FS_DBLADFS) ? 10 : 5;
		/* checksum */
		SectPtr[255] = CheckSum(SectPtr);

		/* SECTOR 1 : free space map */
		SectPtr += 256;

		/* first free space is either 0x9f9 or 0x4f9 sectors long */
		SectPtr[0] = 0xf9;
		SectPtr[1] = (type == FS_DBLADFS) ? 0x9 : 0x4;
		/* random disc identifier */
		SectPtr[251] = rand()&255;
		SectPtr[252] = rand()&255;
		/* boot option */
		SectPtr[253] = 0;
		/* end of free space map pointer */
		SectPtr[254] = 3;
		/* checksum */
		SectPtr[255] = CheckSum(SectPtr);

		/* sectors encompasing root directory */
		SectPtr += 256;

		/* fixed strings */
		SectPtr[1] = 'H'; SectPtr[2] = 'u'; SectPtr[3] = 'g'; SectPtr[4] = 'o';
		SectPtr[1275] = 'H'; SectPtr[1276] = 'u'; SectPtr[1277] = 'g'; SectPtr[1278] = 'o';
		/* is root directory, etc */
		SectPtr[1228] = '$';
		SectPtr[1238] = 2;
		SectPtr[1241] = 0x0d;

		/* the quantity of space already occupied */
		Len1 = 1280;
		Len2 = 0;
	}
	else
	{
		/* write an empty (single sided) DFS directory tree */
		SectPtr = Data1 + 256;

		/* 80*10 sectors on this disc = (3 * 256) + 32 */
		SectPtr[6] = 3; SectPtr[7] = 32;
	}

	return true;
}

bool C_drive::GenerateNew(FSTypes type)
{
	SectorLength = 256;
	Tracks = 80;

	switch(type)
	{
		case FS_DBLDFS :
		case FS_DFS :		Sectors = 10; break;

		case FS_DBLADFS :
		case FS_ADFS :		Sectors = 16; break;
	}

	/* Get first side */
	if(!NewSide(type)) return false;

	/* Maybe get second side - for DFS discs */
	if(type == FS_DBLDFS)
	{
		Len2 = Len1;
		Data2 = Data1;

		if(!NewSide(type))
		{
			Close();
			return false;
		}
	}

	return true;
}

#define CheckExtension(t)	\
	if(!electron->GetFS()->ExtensionIncluded(name, t))	\
	{	\
		name = electron->GetFS()->GetCleanFilename(name, t[0]);	\
		free(name2);	\
	}

bool C_drive::LoadTraditional(bool neu, FSTypes type)
{
	bool result = false;
	gzFile stream;
	FILE *nstream;

	unsigned __int8 gz_magic[2] = {0x1f, 0x8b}; //for detecting GZip'd files
	unsigned __int8 tspace[2];

	char *linear[] =
	{
		"adf\a", "adm\a", "ssd\a", "img\a", NULL
	};
	char *ssdextension[] =
	{
		"ssd\a", "img\a", NULL
	};
	char *adfextension[] =
	{
		"adf\a", "adm\a", NULL
	};

	char *dsdextension[] =
	{
		"dsd\a", NULL
	};
	char *adlextension[] =
	{
		"adl\a", NULL
	};

	/*

	ZLib can transparently handle GZip and non-compressed files, but does so
	as far as not being willing to tell you what type is open! Since ElectrEm
	opens a file, then rewrites later it if contents have changed, it needs
	to know and hence checks the two 'GZip magic bytes' itself

	*/

	if(!(nstream = electron->GetFS()->GetFile(path, NULL, name, "rb")))
	{
		if(!neu) return false;

		/* generate new disc */

		/* check extension */
		char *name2 = name;

		switch(type)
		{
			case FS_DFS :		CheckExtension(ssdextension);	break;
			case FS_DBLDFS :	CheckExtension(dsdextension);	break;
			case FS_ADFS :		CheckExtension(adfextension);	break;
			case FS_DBLADFS :	CheckExtension(adlextension);	break;
		}

		/* setup memory and that */
		if(!GenerateNew(type)) return false;

		dirty = true;
		gz = false;
		return true;
	}
	fread(tspace, 1, 2, nstream);
	electron->GetFS()->ReleaseFile(nstream);

	gz = ((tspace[0] == gz_magic[0]) && (tspace[1] == gz_magic[1])) ? true : false;

	/* now, the actual load code */

	if(!(stream = electron->GetFS()->GetGZ(path, NULL, name, "rb"))) return false;

	if(electron->GetFS()->ExtensionIncluded(name, linear))
	{
		//prepare for reading of an SSD
		if(electron->GetFS()->ExtensionIncluded(name, ssdextension))
		{
			SectorLength = 256; Sectors = 10; Tracks = 80;
		}

		//prepare for reading of an ADF
		if(electron->GetFS()->ExtensionIncluded(name, adfextension))
		{
			SectorLength = 256; Sectors = 16; Tracks = 80;
		}

		//load ADF or SSD
		Data1 = (unsigned __int8 *)malloc(Tracks*Sectors*SectorLength);
		Len1 = gzread(stream, Data1, Tracks*Sectors*SectorLength);
		gzclose(stream);

		return true;
	}
	else
	{
		if(electron->GetFS()->ExtensionIncluded(name, dsdextension))
		{
			SectorLength = 256; Sectors = 10; Tracks = 80;
		}

		if(electron->GetFS()->ExtensionIncluded(name, adlextension))
		{
			SectorLength = 256; Sectors = 16; Tracks = 80;
		}

		/* allocate temporary work table, load all data in file */
		Data = (unsigned __int8 *)malloc(Tracks*Sectors*SectorLength*2);
		DataLen = gzread(stream, Data, Tracks*Sectors*SectorLength*2);
		gzclose(stream);

		/* these files are track interleaved, so descramble */
		DeInterleave(Sectors*SectorLength);

		/* free temporary work table */
		free(Data);

		return true;
	}

	gzclose(stream);
	return false;
}

#undef CheckExtension

bool C_drive::New(char *pth, char *nme, unsigned __int8 discid, FSTypes type)
{
	return OpenFile(pth, nme, discid, true, type);
}

bool C_drive::Open(char *pth, char *nme, unsigned __int8 discid)
{
	return OpenFile(pth, nme, discid, false, FS_NONE);
}

bool C_drive::OpenFile(char *pth, char *nme, unsigned __int8 discid, bool neu, FSTypes type)
{
	BrokenWord value;
	bool result = false;
	char *smplexts[] =
	{
		"ssd\a", "dsd\a", "img\a", "adf\a", "adl\a", NULL
	};

	fileprot = false;

	Close();
	Len1 = Len2 = flips = 0;
	path = (pth != NULL) ? strdup(pth) : NULL;
	name = (nme != NULL) ? strdup(nme) : NULL;

	dirty = false;

	if(electron->GetFS()->ExtensionIncluded(name, smplexts))
	{
		return LoadTraditional(neu, type);
	}
	else
	{
		//try as a UEF
		id = discid;

		if(neu)
		{
			char *name2 = name;
			name = electron->GetFS()->GetCleanFilename(name, "uef");
			free(name2);

			if(!(disc = electron->GetFS()->RequestUEFChunkSelector(path, NULL, name, "w")))
				return false;

			/* create new disc in memory */
			GenerateNew(type);
			dirty = true;

			/* write some UEF data */
			disc->EstablishChunk();
			disc->CurrentChunk()->SetId(0x0200);

			/* disc id */
			disc->CurrentChunk()->PutC(0);
			/* sector length */
			disc->CurrentChunk()->PutC(0); disc->CurrentChunk()->PutC(1);
			/* sector per track */
			disc->CurrentChunk()->PutC(((type == FS_DFS) || (type == FS_DBLDFS)) ? 10 : 16);
			/* tracks */
			disc->CurrentChunk()->PutC(80);
			/* fs type */
			disc->CurrentChunk()->PutC(((type == FS_DFS) || (type == FS_DBLDFS)) ? 1 : 3);

			/* establish a first side */
			disc->EstablishChunk();
			disc->CurrentChunk()->SetId(0x0201);
			disc->CurrentChunk()->PutC(0);

			if((type == FS_DBLADFS) || (type == FS_DBLDFS))
			{
				disc->EstablishChunk();
				disc->CurrentChunk()->SetId(0x0201);
				disc->CurrentChunk()->PutC(0x80);
			}

			return true;
		}

		if(!(disc = electron->GetFS()->RequestUEFChunkSelector(path, NULL, name, "rw")))
		{
			fileprot = true;

			if(!(disc = electron->GetFS()->RequestUEFChunkSelector(path, NULL, name, "r")))
				return false;
		}

		if(!disc->FindId(0x0200))
			return false;

		electron->SetGlobals(disc);

		disc->CurrentChunk()->ReadSeek(1, SEEK_SET);

		value.b.l = disc->CurrentChunk()->GetC();
		value.b.h = disc->CurrentChunk()->GetC();
		SectorLength = value.a;

		Sectors = disc->CurrentChunk()->GetC();
		Tracks = disc->CurrentChunk()->GetC();

		Track = 0; Trackadd = 1;

		if(FindSide(true))
		{
			Len1 = disc->CurrentChunk()->GetLength()-1;
			Data1 = (unsigned __int8 *)malloc(Tracks*Sectors*SectorLength);
			disc->CurrentChunk()->Read(Data1, Len1);
		}

		if(FindSide(false))
		{
			Len2 = disc->CurrentChunk()->GetLength()-1;
			Data2 = (unsigned __int8 *)malloc(Tracks*Sectors*SectorLength);
			disc->CurrentChunk()->Read(Data2, Len2);
		}

		if(!Data1 && !Data2) //no disc data was found, you know?
			Close();

		FILE *temp = fopen("adfs.dmp", "wb");
		fwrite(Data1, 1, 7*1024, temp);
		fclose(temp);

		return true;
	}

	return false;
}

void C_drive::Close(void)
{
	if(disc)
	{
		if(dirty && !GetWriteProtect())
		{
			if(Data1 && FindSide(true))
			{
				disc->CurrentChunk()->SetLength(1);
				disc->CurrentChunk()->WriteSeek(1, SEEK_SET);

				disc->CurrentChunk()->Write(Data1, Len1);
			}

			if(Data2 && FindSide(false))
			{
				disc->CurrentChunk()->SetLength(1);
				disc->CurrentChunk()->WriteSeek(1, SEEK_SET);

				disc->CurrentChunk()->Write(Data2, Len2);
			}
		}

		electron->GetFS()->ReleaseUEFChunkSelector(disc);
		disc = NULL;
	}
	else
	{
		if(dirty && Data1)
			SaveTraditional();
	}

	if(name)
	{
		free(name);
		name = NULL;
	}

	if(path)
	{
		free(path);
		path = NULL;
	}

	if(Data1)
	{
		free(Data1);
		Data1 = NULL;
	}

	if(Data2)
	{
		free(Data2);
		Data2 = NULL;
	}
}

/* The track stepping / setting bits */

void C_drive::SetDir(bool inwards)
{
	Trackadd = inwards ? -1 : 1;
}

bool C_drive::GetDir(void)
{
	return Trackadd == -1;
}

bool C_drive::Step(void)
{
	Track += Trackadd;

	if(Track < 0 || Track > Tracks)
	{
		Track -= Trackadd;
		return false;
	}

	return true;
}

bool C_drive::SetTrack(int ntrack)
{
	if(ntrack < 0 || ntrack > Tracks)
		return false;

	Track = ntrack;
	return true;
}

int C_drive::GetTrack(void)
{
	return Track;
}

/* The sector read or write bit */

unsigned __int8 *C_drive::GetSectorPtr(int snum, bool write)
{
	int addr = (Track*Sectors + snum) * SectorLength;

	if(snum < 0 || snum >= Sectors || !Data1)
		return NULL;

	if(write)
	{
		dirty = true;

		/*

		important note : loads of disc images (most) are 'chopped' such that
		they do not have the blank space included. Therefore it is necessary
		to extend their length if a write occurs beyond what is stored.

		*/
		if((addr+SectorLength) > Len1)
		{
			Len1 = addr+SectorLength;
		}
	}

	return &Data1[addr];
}

void C_drive::ReleaseSectorPtr(int snum)
{
}

/* Disc flipping / side selection */

void C_drive::Flip(void)
{
	unsigned __int8 *tmp;

	if(Data2)
	{
		tmp = Data2; Data2 = Data1; Data1 = tmp;
		Len1 ^= Len2; Len2 ^= Len1; Len1 ^= Len2;
		flips++;
	}
}

bool C_drive::FirstSide(void)
{
	return (flips&1) ? false : true;
}

/* Just a disk present check kind of thing */

bool C_drive::HasFile(void)
{
	return Data1 ? true : false;
}

bool C_drive::GetWriteProtect(void)
{
	return fileprot ? false : prot;
}

bool C_drive::SetWriteProtect(bool protect)
{
	prot = protect;

	return GetWriteProtect();
}

/* Autoload ID */
unsigned __int8 C_drive::GetAutoloadID(void)
{
	switch(Sectors)
	{
		case 16 : //assume 16 sectors = ADFS format
			return 12+Data1[SectorLength+253];
		break;

		case 10 : //and 10 sectors = DFS format
			return 4+((Data1[SectorLength+6] >> 4)&3);
		break;

		default :
		return 0xff;
	}
}

